!---------------------------------------------------------------------!
!                                PHAML                                !
!                                                                     !
! The Parallel Hierarchical Adaptive MultiLevel code for solving      !
! linear elliptic partial differential equations of the form          !
! (PUx)x + (QUy)y + RU = F on 2D polygonal domains with mixed         !
! boundary conditions, and eigenvalue problems where F is lambda*U.   !
!                                                                     !
! PHAML is public domain software.  It was produced as part of work   !
! done by the U.S. Government, and is not subject to copyright in     !
! the United States.                                                  !
!                                                                     !
!     William F. Mitchell                                             !
!     Applied and Computational Mathematics Division                  !
!     National Institute of Standards and Technology                  !
!     william.mitchell@nist.gov                                       !
!     http://math.nist.gov/phaml                                      !
!                                                                     !
!---------------------------------------------------------------------!

module refine_adapt_mod

!----------------------------------------------------
! This module contains routines for the basic control of adaptive refinement.
!
! communication tags in this module are of the form 33xx
!----------------------------------------------------

!----------------------------------------------------
! Other modules used:

use global
use gridtype_mod
use grid_util
use linsystype_mod
use hp_strategies
use refine_elements
use message_passing
use hash_mod
use linsys_io
use load_balance
use error_estimators
use omp_lib
use zoltan_interf
use stopwatch

!----------------------------------------------------

implicit none
private
public refine_adaptive, refedge_based_refine_grid, &
       reconcile

!----------------------------------------------------
! Non-module procedures used are:

!----------------------------------------------------
! The following defined constants are defined:

!----------------------------------------------------
! The following variables are defined:

!----------------------------------------------------

contains

!          ---------------
subroutine refine_adaptive(linsys,grid,procs,refine_control,solver_control, &
                           lb,still_sequential,init_nvert,init_nelem,init_dof, &
                           loop,partition_method,balance_what,predictive, &
                           stalled,no_time)
!          ---------------

!----------------------------------------------------
! This routine performs adaptive refinement of the grid.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linsys_type), intent(inout) :: linsys
type(grid_type), intent(inout) :: grid
type (proc_info), target, intent(inout) :: procs
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
type(Zoltan_Struct), pointer :: lb
integer, intent(in) :: init_nvert, init_nelem, init_dof, loop, &
                       partition_method, balance_what
logical, intent(in) :: still_sequential, predictive
logical, intent(out) :: stalled
logical, intent(in), optional :: no_time
!----------------------------------------------------
! Local variables:

logical :: any_changes, any_p_coarsened, any_h_coarsened, any_p_refined, &
           any_h_refined, timeit
real(my_real) :: target
real(my_real), save :: last_nelem_change = -1.0_my_real
integer :: starting_nelem, ending_nelem
integer :: nelem, loop_count, astat, itemp
integer, pointer :: desired_level(:),desired_degree(:,:), &
                    leaf_elements(:)
logical(small_logical), pointer, save :: was_hcoarsened(:) => NULL()
logical(small_logical), pointer :: temp_was_coarsened(:)
!----------------------------------------------------
! Begin executable code

any_h_coarsened = .false.
any_p_coarsened = .false.
any_h_refined   = .false.
any_p_refined   = .false.

if (present(no_time)) then
   timeit = .not. no_time
else
   timeit = .true.
endif

! Allocate memory.

allocate(desired_level(size(grid%element)), &
         desired_degree(2,size(grid%element)),leaf_elements(grid%biggest_elem),&
         stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in refine_adaptive")
   stop
endif

! If the refinement termination is MATCH_TOL and the number of times through the
! refine-solve loop meets recoarsen_limit, get the starting number of elements

if (refine_control%refterm == MATCH_TOL .and. &
    loop >= refine_control%recoarsen_limit) then
   call get_grid_info(grid,procs,still_sequential,3309, &
                      total_nelem_leaf=starting_nelem,no_master=.true.)
endif

! Indicate no elements came from coarsening, unless the refinement termination
! is MATCH_TOL, the number of times through the refine-solve loop exceeds
! recoarsen_limit, and the change in the number of elements last time was less
! than recoarsen_tol.

if (refine_control%refterm /= MATCH_TOL .or. &
    loop <= refine_control%recoarsen_limit .or. &
    last_nelem_change > refine_control%recoarsen_tol) then

   if (associated(was_hcoarsened)) then
      if (size(was_hcoarsened) /= size(grid%element)) then
         deallocate(was_hcoarsened)
      endif
   endif

   allocate(was_hcoarsened(size(grid%element)),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in refine_adaptive")
      stop
   endif

   was_hcoarsened = .false.

! If refinement termination is MATCH_TOL, etc., and was_hcoarsened was kept, and
! it is not big enough, reallocate and copy

else

   if (associated(was_hcoarsened)) then
      if (size(was_hcoarsened) /= size(grid%element)) then
         temp_was_coarsened => was_hcoarsened
         nullify(was_hcoarsened)
         allocate(was_hcoarsened(size(grid%element)),stat=astat)
         if (astat /= 0) then
            ierr = ALLOC_FAILED
            call fatal("memory allocation failed in refine_adaptive")
            stop
         endif
         was_hcoarsened(1:size(temp_was_coarsened)) = temp_was_coarsened
         was_hcoarsened(size(temp_was_coarsened)+1:) = .false.
         deallocate(temp_was_coarsened)
      endif
   endif

endif

! Recompute error indicators, if they're stale

if (.not. grid%errind_up2date) then
   call all_error_indicators(grid,refine_control%error_estimator,solver_control)
endif

! Set the target that determines when enough refinement has occured.

target = set_termination_target(grid,procs,refine_control,solver_control, &
                                init_nvert,init_nelem,init_dof,loop, &
                                still_sequential)

! Repeat until enough refinement has occured.

loop_count = 0
do
   loop_count = loop_count + 1
   if (loop_count > refine_control%refmaxloop) exit

! Make a list of leaf elements.

   if (size(leaf_elements) /= grid%biggest_elem) then
      deallocate(leaf_elements)
      allocate(leaf_elements(grid%biggest_elem),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in refine_adaptive")
         stop
      endif
   endif

   call list_elements(grid,leaf_elements,nelem,leaf=.true.)

! Mark elements for refinement.  This sets, for each leaf element, the new
! h-level and p in the conforming adapted grid.

   call mark_for_refinement(linsys,grid,procs,refine_control,solver_control, &
                            still_sequential,leaf_elements,nelem,desired_level,&
                            desired_degree,was_hcoarsened,any_changes, &
                            loop_count)

! If there are no changes on any processor, quit.

   if (.not. any_changes) exit

! Perform predictive load balancing on the first loop.

   if (loop_count==1 .and. predictive .and. .not. still_sequential) then
      call pred_load_balance(linsys,grid,procs,refine_control,solver_control, &
                             lb,partition_method,still_sequential,balance_what,&
                             desired_level,desired_degree,was_hcoarsened, &
                             any_changes)

! Reallocate if predictive load balance reallocated grid%element

      if (size(leaf_elements) /= grid%biggest_elem) then
         deallocate(leaf_elements)
         allocate(leaf_elements(grid%biggest_elem),stat=astat)
         if (astat /= 0) then
            ierr = ALLOC_FAILED
            call fatal("memory allocation failed in refine_adaptive")
            stop
         endif
      endif

! redo mark for refinement because load balance changed grid

      if (any_changes) then
         call list_elements(grid,leaf_elements,nelem,leaf=.true.)
         call mark_for_refinement(linsys,grid,procs,refine_control, &
                                  solver_control,still_sequential, &
                                  leaf_elements,nelem,desired_level, &
                                  desired_degree,was_hcoarsened,any_changes, &
                                  loop_count)
      endif
   endif

   if (refine_control%derefine) then

! Perform coarsenings, first p then h.

      if (refine_control%reftype == P_UNIFORM .or. &
          refine_control%reftype == P_ADAPTIVE .or. &
          refine_control%reftype == HP_ADAPTIVE) then
         call p_coarsen_grid(grid,refine_control,solver_control, &
                             desired_degree,leaf_elements,nelem, &
                             any_p_coarsened)
      else
         any_p_coarsened = .false.
      endif
      if (refine_control%reftype == H_UNIFORM .or. &
          refine_control%reftype == H_ADAPTIVE .or. &
          refine_control%reftype == HP_ADAPTIVE) then
         call h_coarsen_grid(grid,leaf_elements,nelem,desired_level, &
                             desired_degree,refine_control,solver_control, &
                             was_hcoarsened,any_h_coarsened)
      else
         any_h_coarsened = .false.
      endif

! If not increasing the size of the grid, check to see if we're done after
! the coarsening phase.

      if (refine_control%refterm == KEEP_NVERT        .or. &
          refine_control%refterm == KEEP_NELEM        .or. &
          refine_control%refterm == KEEP_NEQ) then
         if (adapt_done(grid,procs,refine_control,solver_control, &
                        still_sequential,target,.true.)) then
            any_p_refined = .false.
            any_h_refined = .false.

! If it is done and there was any p coarsening, need to enforce the edge rule
! and compute error indicators.

            if (any_p_coarsened) then
               call enforce_and_errind(grid,refine_control,solver_control, &
                                       leaf_elements,nelem)
            endif
            exit
         endif
      endif

   else

      any_p_coarsened = .false.
      any_h_coarsened = .false.

   endif

! Remake the list of leaf elements because h coarsening invalidates it.

   call list_elements(grid,leaf_elements,nelem,leaf=.true.)

! Perform refinements, first p then h.

   if (refine_control%reftype == P_UNIFORM .or. &
       refine_control%reftype == P_ADAPTIVE .or. &
       refine_control%reftype == HP_ADAPTIVE) then
      call p_refine_grid(grid,refine_control,solver_control,desired_degree, &
                         leaf_elements,nelem,any_p_refined)
   else
      any_p_refined = .false.
   endif
   if (refine_control%reftype == H_UNIFORM .or. &
       refine_control%reftype == H_ADAPTIVE .or. &
       refine_control%reftype == HP_ADAPTIVE) then
      call h_refine_grid(grid,refine_control,solver_control,desired_level, &
                         desired_degree,leaf_elements,nelem,was_hcoarsened, &
                         any_h_refined)
   else
      any_h_refined = .false.
   endif

! See if enough refinement has occured.

   if (adapt_done(grid,procs,refine_control,solver_control,still_sequential, &
                  target, any_p_coarsened .or. any_h_coarsened .or. &
                  any_p_refined .or. any_h_refined)) exit

end do

! check for stalling

if (loop_count == 1) then
   if (any_p_coarsened .or. any_h_coarsened .or. any_p_refined .or. &
       any_h_refined) then
      itemp = 1
   else
      itemp = 0
   endif
   itemp = phaml_global_max(procs,itemp,3307)
   if (itemp == 0) then
      stalled = .true.
   else
      stalled = .false.
   endif
else
   stalled = .false.
endif

! If refterm is MATCH_TOL and the number of times through the refine-solve loop
! meets the recoarsen limit, keep was_hcoarsened to use next time.

if (refine_control%refterm == MATCH_TOL .and. &
    loop >= refine_control%recoarsen_limit) then
   call get_grid_info(grid,procs,still_sequential,3308, &
                      total_nelem_leaf=ending_nelem,no_master=.true.)
   last_nelem_change = abs(starting_nelem - ending_nelem) / &
                       real(starting_nelem,my_real)
else
   deallocate(was_hcoarsened)
endif

! Free memory.

deallocate(desired_level,desired_degree,leaf_elements)

end subroutine refine_adaptive

!        ----------------------
function set_termination_target(grid,procs,refine_control,solver_control, &
                                init_nvert,init_nelem,init_dof,loop, &
                                still_sequential) result(target)
!        ----------------------

!----------------------------------------------------
! This function determines the value to use as the target for terminating
! refinement.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type (proc_info), intent(in) :: procs
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
integer, intent(in) :: init_nvert, init_nelem, init_dof, loop
logical, intent(in) :: still_sequential
real(my_real) :: target
!----------------------------------------------------
! Local variables:

integer :: itarget
!----------------------------------------------------
! Begin executable code

! Set the target value depending on what type of termination is used.

select case (refine_control%refterm)

case (DOUBLE_NVERT)

   target = init_nvert*refine_control%inc_factor**loop
   if (target > refine_control%max_vert) target = refine_control%max_vert

case (DOUBLE_NELEM)

   target = init_nelem*refine_control%inc_factor**loop
   if (target > refine_control%max_elem) target = refine_control%max_elem

case (DOUBLE_NEQ)

   target = init_dof*refine_control%inc_factor**loop
   if (target > refine_control%max_dof) target = refine_control%max_dof

case (HALVE_ERREST)

   if (.not. grid%errind_up2date) then
      call all_error_indicators(grid,refine_control%error_estimator, &
                                solver_control,max_errind=target, &
                                refine_control=refine_control)
      target = phaml_global_max(procs,target,3321)
   else
      target = compute_global_max_errind(grid,refine_control,procs, &
                                      still_sequential,no_notrefinable=.true.)/2
   endif

case (KEEP_NVERT)

   if (refine_control%max_vert /= huge(0)) then
      target = refine_control%max_vert
   else
      call get_grid_info(grid,procs,still_sequential,3301,total_nvert=itarget, &
                         no_master=.true.)
      target = itarget
      if (target > refine_control%max_vert) target = refine_control%max_vert
   endif

case (KEEP_NELEM)

   if (refine_control%max_elem /= huge(0)) then
      target = refine_control%max_elem
   else
      call get_grid_info(grid,procs,still_sequential,3301, &
                         total_nelem_leaf=itarget,no_master=.true.)
      target = itarget
      if (target > refine_control%max_elem) target = refine_control%max_elem
   endif

case (KEEP_NEQ)

   if (refine_control%max_dof /= huge(0)) then
      target = refine_control%max_dof
   else
      call get_grid_info(grid,procs,still_sequential,3301,total_dof=itarget, &
                         no_master=.true.)
      target = itarget
      if (target > refine_control%max_dof) target = refine_control%max_dof
   endif

case (ONE_REF, ONE_REF_HALF_ERRIND, MATCH_TOL)

   target = 0

case default

   call fatal("illegal value for refterm",procs=procs)
   stop

end select

end function set_termination_target

!        ----------
function adapt_done(grid,procs,refine_control,solver_control,still_sequential, &
                    target,any_changes)
!        ----------

!----------------------------------------------------
! This routine checks to see if the target for terminating refinement is met.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type (proc_info), intent(in) :: procs
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
logical, intent(in) :: still_sequential
real(my_real), intent(in) :: target
logical, intent(in) :: any_changes
logical :: adapt_done
!----------------------------------------------------
! Local variables:

integer :: compare, nlev, deg, nvert, nelem, dof
real(my_real) :: max_errind
!----------------------------------------------------
! Begin executable code

! Assume we're not done.

adapt_done = .false.

! Compare the refterm-dependent value to the target.

select case (refine_control%refterm)

case (DOUBLE_NVERT, KEEP_NVERT)

   call get_grid_info(grid,procs,still_sequential,3302,total_nvert=compare, &
                      no_master=.true.)
   if (compare >= target) adapt_done = .true.

case (DOUBLE_NELEM, KEEP_NELEM)

   call get_grid_info(grid,procs,still_sequential,3302, &
                      total_nelem_leaf=compare,no_master=.true.)
   if (compare >= target) adapt_done = .true.

case (DOUBLE_NEQ, KEEP_NEQ)

   call get_grid_info(grid,procs,still_sequential,3302,total_dof=compare, &
                      no_master=.true.)
   if (compare >= target) adapt_done = .true.

case (HALVE_ERREST)

   if (.not. grid%errind_up2date) then
      call all_error_indicators(grid,refine_control%error_estimator, &
                                solver_control,max_errind=max_errind, &
                                refine_control=refine_control)
      if (phaml_global_max(procs,max_errind,3322) <= target) then
         adapt_done = .true.
      endif
   else
      if (compute_global_max_errind(grid,refine_control,procs,still_sequential,&
                                    no_notrefinable=.true.) <= target) then
         adapt_done = .true.
      endif
   endif

case (ONE_REF, ONE_REF_HALF_ERRIND)

   adapt_done = .true.

case (MATCH_TOL)

   if (any_changes) then
      compare = 1
   else
      compare = 0
   endif
   compare = phaml_global_max(procs,compare,3305)
   adapt_done = compare==0

case default

   call fatal("illegal value for refterm")
   stop

end select

! check for reaching maximum levels, degree, vertices, elements or degrees of
! freedom

if (.not. adapt_done .and. &
    (refine_control%stop_on_maxlev .or. &
     refine_control%stop_on_maxdeg .or. &
     refine_control%max_vert /= huge(0) .or. &
     refine_control%max_elem /= huge(0) .or. &
     refine_control%max_dof  /= huge(0))) then

   call get_grid_info(grid,procs,still_sequential,3304,max_nlev=nlev, &
                      maxdeg=deg,total_nvert=nvert,total_nelem_leaf=nelem, &
                      total_dof=dof,no_master=.true.)

   if (refine_control%stop_on_maxlev) then
      if (nlev >= refine_control%max_lev) adapt_done = .true.
   endif

   if (refine_control%stop_on_maxdeg) then
      if (deg >= refine_control%max_deg) adapt_done = .true.
   endif

   adapt_done = adapt_done .or. &
                nvert >= refine_control%max_vert .or. &
                nelem >= refine_control%max_elem .or. &
                dof   >= refine_control%max_dof
endif

if (.not. adapt_done .and. refine_control%refterm /= MATCH_TOL) then
   if (any_changes) then
      compare = 1
   else
      compare = 0
   endif
   compare = phaml_global_max(procs,compare,3306)
   if (compare == 0) adapt_done = .true.
endif

end function adapt_done

!          -------------------
subroutine mark_for_refinement(linsys,grid,procs,refine_control,solver_control,&
                               still_sequential,leaf_elements,nleaf, &
                               desired_level,desired_degree,was_hcoarsened, &
                               any_changes,loop_count)
!          -------------------

!----------------------------------------------------
! This routine determines the new h-level and p for each element in a
! conforming adapted grid.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linsys_type), intent(inout) :: linsys
type(grid_type), intent(inout) :: grid
type(proc_info), intent(in) :: procs
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
logical, intent(in) :: still_sequential
integer, intent(in) :: leaf_elements(:)
integer, intent(in) :: nleaf
integer, intent(out) :: desired_level(:), desired_degree(:,:)
logical(small_logical), intent(in) :: was_hcoarsened(:)
logical, intent(out) :: any_changes
integer, intent(in) :: loop_count
!----------------------------------------------------
! Local variables:

real(my_real), save :: refine_cutoff, coarsen_cutoff
integer :: i, elem, new_p(2)
character(len=1) :: reftype(grid%biggest_elem)
logical :: h_coarsen, p_coarsen
!----------------------------------------------------
! Begin executable code

! Initialize so there is something in the unused entries.

desired_level = 0
desired_degree = 0
any_changes = .false.

! Update the error indicators.

if (.not. grid%errind_up2date) then
   call all_error_indicators(grid,refine_control%error_estimator, &
                             solver_control,delayed=.true.)
endif

! Determine the error indicator cutoffs for whether or not an element
! should be refined or coarsened.

if (refine_control%refterm /= MATCH_TOL .or. loop_count == 1) then
   call determine_cutoffs(linsys,grid,procs,refine_control,still_sequential, &
                          leaf_elements,nleaf,refine_cutoff,coarsen_cutoff)
endif

! In 3D, if the refine cutoff is essentially 0, we have probably refined all
! elements, because the initial guess is setting the hierarchical coefficients
! to 0, so quit because we don't know what to refine.

if (global_element_kind == TETRAHEDRAL_ELEMENT .and. &
    refine_cutoff < 1.0e-100_my_real) return

! For each leaf element ...

!$omp parallel do &
!$omp  default(shared) &
!$omp  private(i,elem,new_p,h_coarsen,p_coarsen) &
!$omp  reduction(.or.: any_changes)

do i=1,nleaf
   elem = leaf_elements(i)

! If I don't own it, set the level to 1 to be set by the owner or
! compatibility, and leave the degree the same.

   if (.not. grid%element(elem)%iown) then

      desired_level(elem) = 1
      desired_degree(:,elem) = grid%element(elem)%degree

! If the error indicator for any eigenvector is sufficiently large, refine.

   elseif (any(grid%element_errind(elem,:) >= refine_cutoff)) then

! Determine if refinement is by h or p.

      new_p = 0
      call mark_reftype_one(elem,grid,refine_control,solver_control, &
                            -1.0_my_real,reftype,.false.,new_p)

! Set the result accordingly.

      select case (reftype(elem))

      case ("h")

! h refinement means a single h refinement.  If new_p(1:2) were set, they
! give the desired degrees of the two children, with new_p(1) being the child
! with vertex 1 of elem.  Otherwise, the desired degree is the current degree.

         desired_level(elem) = min(refine_control%max_lev, &
                                   grid%element(elem)%level+1)
         if (new_p(1) == 0) then
            desired_degree(:,elem) = grid%element(elem)%degree
         else
            desired_degree(:,elem) = new_p
         endif

      case ("p")

! p refinement means leave the level the same.  If new_p(1) is set, it is
! the desired degree of the element.  Otherwise increment it by one.

         desired_level(elem) = grid%element(elem)%level
         if (new_p(1) == 0) then
            desired_degree(:,elem) = min(refine_control%max_deg, &
                                         grid%element(elem)%degree+1)
         else
            desired_degree(:,elem) = min(refine_control%max_deg,new_p(1))
         endif

      case ("n")

! no refinement means keep level and degree the same

         desired_level(elem) = grid%element(elem)%level
         desired_degree(:,elem) = grid%element(elem)%degree

      case default
         ierr = PHAML_INTERNAL_ERROR
         call fatal("mark_reftype_one did not return h, p or n in mark_for_refinement")
         stop

      end select

! If not refining, determine if it should be h or p coarsened and set result.

   else

      call determine_coarsening(grid,refine_control,elem,coarsen_cutoff, &
                                was_hcoarsened,h_coarsen,p_coarsen)

      if (h_coarsen) then
         desired_level(elem) = max(1,grid%element(elem)%level-1)
         desired_degree(:,elem) = grid%element(elem)%degree
      elseif (p_coarsen) then
         desired_level(elem) = grid%element(elem)%level
         desired_degree(:,elem) = max(1,grid%element(elem)%degree-1)
      else
         desired_level(elem) = grid%element(elem)%level
         desired_degree(:,elem) = grid%element(elem)%degree
      endif
   endif

   if (desired_level(elem) /= grid%element(elem)%level .or. &
       desired_degree(1,elem) /= grid%element(elem)%degree) then
      any_changes = .true.
   endif

end do
!$omp end parallel do

! Reconcile the conflicting desired refinements/coarsenings to get a
! compatible grid.

call reconcile_requests(grid,procs,refine_control,still_sequential, &
                        leaf_elements,nleaf,desired_level,desired_degree, &
                        any_changes)

end subroutine mark_for_refinement

!          -----------------
subroutine determine_cutoffs(linsys,grid,procs,refine_control,still_sequential,&
                             leaf_elements,nleaf,refine_cutoff,coarsen_cutoff)
!          -----------------

!----------------------------------------------------
! This routine determines the error indicator cutoff values for determining
! whether or not an element should be refined or coarsened.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linsys_type), intent(inout) :: linsys
type(grid_type), intent(inout) :: grid
type(proc_info), intent(in) :: procs
type(refine_options), intent(in) :: refine_control
logical, intent(in) :: still_sequential
integer, intent(in) :: leaf_elements(:)
integer, intent(in) :: nleaf
real(my_real), intent(out) :: refine_cutoff, coarsen_cutoff
!----------------------------------------------------
! Local variables:

real(my_real) :: normsoln
integer :: nelem
!----------------------------------------------------
! Begin executable code

if (abs(refine_control%reftol-MAX_ERRIND_OVER_FACTOR) < .01_my_real) then
   refine_cutoff = compute_global_max_errind(grid,refine_control,procs, &
                                             still_sequential,leaf_elements, &
                                             nleaf,no_notrefinable=.true.)
else
   refine_cutoff = refine_control%reftol
endif

refine_cutoff = refine_cutoff / refine_control%reftol_factor

if (refine_control%reftol_scale_energy) then
   call get_grid_info(grid,procs,still_sequential,3303,total_nelem_leaf=nelem, &
                      no_master=.true.)
   if (special_errest == JOHN_H1 .or. special_errest == JOHN_L1 .or. &
       special_errest == JOHN_L1_NOABS) then
      refine_cutoff = refine_control%reftol/nelem
   else
      refine_cutoff = refine_control%reftol/sqrt(real(nelem,my_real))
   endif
   if (grid%errtype == RELATIVE_ERROR) then
      if (special_errest == USE_H1_ERREST) then
         call norm_solution(linsys,grid,procs,still_sequential,1,1,H1=normsoln)
      elseif (special_errest == USE_L2_ERREST) then
         call norm_solution(linsys,grid,procs,still_sequential,1,1,l2=normsoln)
      else
         call norm_solution(linsys,grid,procs,still_sequential,1,1, &
                            energy=normsoln)
      endif
      if (normsoln /= 0.0_my_real) refine_cutoff = refine_cutoff*normsoln
   endif
endif

if (abs(refine_control%coarsentol-MAX_ERRIND_OVER_FACTOR) < .01_my_real) then
   coarsen_cutoff = compute_global_max_errind(grid,refine_control,procs, &
                                              still_sequential,leaf_elements, &
                                              nleaf,no_notrefinable=.true.)
elseif (abs(refine_control%coarsentol-REFCUTOFF_OVER_FACTOR) < .01_my_real .or.&
        abs(refine_control%coarsentol-REFCUTOFF_TIMES_CONVERGENCE) < .01_my_real) then
   coarsen_cutoff = refine_cutoff
else
   coarsen_cutoff = refine_control%coarsentol
endif

coarsen_cutoff = coarsen_cutoff / refine_control%coarsentol_factor
coarsen_cutoff = min(coarsen_cutoff,refine_cutoff)

end subroutine determine_cutoffs

!          --------------------
subroutine determine_coarsening(grid,refine_control,elem,cutoff, &
                                was_hcoarsened,h_coarsen,p_coarsen)
!          --------------------

!----------------------------------------------------
! This routine determines if an element should be h or p coarsened.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(refine_options), intent(in) :: refine_control
integer, intent(in) :: elem
real(my_real), intent(in) :: cutoff
logical(small_logical), intent(in) :: was_hcoarsened(:)
logical, intent(out) :: h_coarsen, p_coarsen
!----------------------------------------------------
! Local variables:

integer :: parent, siblings(4)
logical :: hcoarsen_possible, pcoarsen_possible
real(my_real) :: hcoarsen_errind, pcoarsen_errind, factor
!----------------------------------------------------
! Begin executable code

! relatives

if (grid%element(elem)%level /= 1) then
   parent = hash_decode_key(grid%element(elem)%gid/MAX_CHILD,grid%elem_hash)
   siblings(1:2) = get_child_lid(grid%element(parent)%gid,ALL_CHILDREN, &
                                 grid%elem_hash)
   if (grid%element(parent)%mate == BOUNDARY) then
      siblings(3:4) = NO_CHILD
   else
      siblings(3:4) = get_child_lid( &
                     grid%element(parent)%mate,ALL_CHILDREN,grid%elem_hash)
   endif
endif

! Check conditions that allow h coarsening.
! derefine must be true.
! Cannot h coarsen level 1 elements.
! Only h coarsen elements that I own.
! Only do h coarsening with h or hp adaptive refinement.
! Do not coarsen elements that have previously been coarsened and then refined.

hcoarsen_possible = refine_control%derefine .and. grid%element(elem)%iown &
                    .and. grid%element(elem)%level > 1
if (refine_control%reftype /= H_ADAPTIVE .and. &
    refine_control%reftype /= HP_ADAPTIVE) hcoarsen_possible = .false.
if (hcoarsen_possible) then
   if (was_hcoarsened(parent)) hcoarsen_possible = .false.
endif

! Check conditions that allow p coarsening.
! derefine must be true.
! Cannot p coarsen an element that has p==1.
! Only p coarsen elements that I own.
! Only do p coarsening with p or hp adaptive.

pcoarsen_possible = refine_control%derefine .and. grid%element(elem)%iown &
                    .and. grid%element(elem)%degree > 1
if (refine_control%reftype /= P_ADAPTIVE .and. &
    refine_control%reftype /= HP_ADAPTIVE) pcoarsen_possible = .false.

! Compute coarsening error indicators

if (hcoarsen_possible) then
   hcoarsen_errind = hcoarsen_indicator(grid,parent,siblings,refine_control)
else
   hcoarsen_errind = huge(0.0_my_real)
endif
if (pcoarsen_possible) then
   pcoarsen_errind = pcoarsen_indicator(grid,elem)
else
   pcoarsen_errind = huge(0.0_my_real)
endif

! If either coarsening indicator is small enough, perform the type of
! coarsening that has the least impact on the solution

if (abs(refine_control%coarsentol-REFCUTOFF_TIMES_CONVERGENCE) < .01_my_real) then
   select case (global_element_kind)
   case (TRIANGULAR_ELEMENT)
      factor = 0.5_my_real**(grid%element(elem)%degree/2.0_my_real)
   case (TETRAHEDRAL_ELEMENT)
      factor = 0.5_my_real**(grid%element(elem)%degree/3.0_my_real)
   end select
else
   factor = 1.0_my_real
endif

h_coarsen = .false.
p_coarsen = .false.
if (hcoarsen_errind < pcoarsen_errind .and. &
    hcoarsen_errind < cutoff*factor) then
   h_coarsen = .true.
elseif (pcoarsen_errind < cutoff*factor) then
   p_coarsen = .true.
endif

end subroutine determine_coarsening

!          ------------------
subroutine reconcile_requests(grid,procs,refine_control,still_sequential, &
                              leaf_elements,nleaf,desired_level, &
                              desired_degree,any_changes)
!          ------------------

!----------------------------------------------------
! This routine reconciles the desired new h-levels and p's so that the new
! desired levels and p's will result in a compatible grid across all processors
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(proc_info), intent(in) :: procs
type(refine_options), intent(in) :: refine_control
logical, intent(in) :: still_sequential
integer, intent(in) :: leaf_elements(:), nleaf
integer, intent(inout) :: desired_level(:),desired_degree(:,:)
logical, intent(inout) :: any_changes
!----------------------------------------------------
! Local variables:

integer :: part_bound_elements(grid%biggest_elem), npartbound, &
           new_desired(grid%biggest_elem), i, elem, j, neigh, &
           neighbors(NEIGHBORS_PER_ELEMENT,grid%biggest_elem), &
           neighneigh3, neigh_change, loop, neighbor_degree
logical :: changed, changed_this_pass
!----------------------------------------------------
! Begin executable code

! 3D does not support MPI parallel or p compatibility, so there is nothing
! in here for it

if (global_element_kind == TETRAHEDRAL_ELEMENT) return

! Make a list of owned leaf elements that have an unowned neighbor.

if (num_proc(procs) > 1 .and. .not. still_sequential) then
   call list_elements(grid,part_bound_elements,npartbound,own=.true., &
                      leaf=.true.,bound_vert=.true.)
else
   npartbound = 0
endif

! Exchange the desired p and h levels of partition-boundary elements with
! the neighbors.  The p's are needed so that the edges on the partition
! boundary get the right degree, but only have to be passed in this first
! message.

changed = .true.
loop = 0
if (num_proc(procs) > 1 .and. .not. still_sequential) then
   call exchange_desired(loop,grid,procs,changed,part_bound_elements, &
                         npartbound,any_changes,desired_level,desired_degree)
endif

! Identify the neighbors of each leaf element

!$omp parallel do &
!$omp  default(shared) &
!$omp  private(i,elem)

do i=1,nleaf
   elem = leaf_elements(i)
   neighbors(:,elem) = get_neighbors(elem,grid)
end do

!$omp end parallel do

! Repeat until no processor made a change in its desireds.

new_desired = desired_level(1:grid%biggest_elem)

do while (changed)
   loop = loop + 1

! Repeat until there are no more changes to my desireds.

   changed = .false.

   do

      changed_this_pass = .false.

! For each leaf element ...

!$omp parallel do &
!$omp  default(shared) &
!$omp  private(i, elem, j, neigh, neighneigh3, neigh_change) &
!$omp  reduction(.or.: changed_this_pass)

      do i=1,nleaf
         elem = leaf_elements(i)

! For each neighbor of elem ...

         do j=1,NEIGHBORS_PER_ELEMENT
            neigh = neighbors(j,elem)
            if (neigh == BOUNDARY) cycle

! Get the 3rd neighbor of the neighbor, to see if elem is across neigh's base.

            neighneigh3 = neighbors(3,neigh)

! See how much the desired level of the neighbor differs from its current level.

            neigh_change = desired_level(neigh)-grid%element(neigh)%level

! Increase the desired level of elem, if necessary to make a compatible grid.
! By always increasing, this gives a preference to refining over coarsening.

            if (neigh_change < 0) then ! coarsening of neighbor

               if (neighneigh3 == elem) then ! shares base
                  if (desired_level(neigh) > new_desired(elem) .and. &
                      desired_level(neigh) > desired_level(elem)) then
                     new_desired(elem) = desired_level(neigh)
                     changed_this_pass = .true.
                  endif
               else ! does not share base
                  if (desired_level(neigh) > new_desired(elem) .and. &
                      desired_level(neigh) > desired_level(elem)) then
                     new_desired(elem) = desired_level(neigh)
                     changed_this_pass = .true.
                  endif
               endif

            elseif (neigh_change == 0) then ! neighbor stays the same

               if (neighneigh3 == elem) then ! shares base
                  if (desired_level(neigh) > new_desired(elem)-1 .and. &
                      desired_level(neigh)-1 > desired_level(elem)) then
                     new_desired(elem) = desired_level(neigh)-1
                     changed_this_pass = .true.
                  endif
               else ! does not share base
                  if (desired_level(neigh) > new_desired(elem) .and. &
                      desired_level(neigh) > desired_level(elem)) then
                     new_desired(elem) = desired_level(neigh)
                     changed_this_pass = .true.
                  endif
               endif

            elseif (neigh_change == 1) then ! neighbor refined once

               if (neighneigh3 == elem) then ! shares base
                  if (desired_level(neigh) > new_desired(elem)-1 .and. &
                      desired_level(neigh)-1 > desired_level(elem)) then
                     new_desired(elem) = desired_level(neigh)-1
                     changed_this_pass = .true.
                  endif
               else ! does not share base
                  if (desired_level(neigh) > new_desired(elem)-1 .and. &
                      desired_level(neigh)-1 > desired_level(elem)) then
                     new_desired(elem) = desired_level(neigh)-1
                     changed_this_pass = .true.
                  endif
               endif

            elseif (neigh_change == 2*(neigh_change/2)) then
                                      ! neighbor refined an even number of times

               if (neighneigh3 == elem) then ! shares base
                  if (desired_level(neigh) > new_desired(elem)-2 .and. &
                      desired_level(neigh)-2 > desired_level(elem)) then
                     new_desired(elem) = desired_level(neigh)-2
                     changed_this_pass = .true.
                  endif
               else ! does not share base
                  if (desired_level(neigh) > new_desired(elem)-1 .and. &
                      desired_level(neigh)-1 > desired_level(elem)) then
                     new_desired(elem) = desired_level(neigh)-1
                     changed_this_pass = .true.
                  endif
               endif

            else ! neighbor refined an odd number of times but more than once

               if (neighneigh3 == elem) then ! shares base
                  if (desired_level(neigh) > new_desired(elem)-1 .and. &
                      desired_level(neigh)-1 > desired_level(elem)) then
                     new_desired(elem) = desired_level(neigh)-1
                     changed_this_pass = .true.
                  endif
               else ! does not share base
                  if (desired_level(neigh) > new_desired(elem)-2 .and. &
                      desired_level(neigh)-2 > desired_level(elem)) then
                     new_desired(elem) = desired_level(neigh)-2
                     changed_this_pass = .true.
                  endif
               endif

            endif

         end do ! neighbors
      end do ! leaf elements
!$omp end parallel do

! Check for any changes this time through the loop, and repeat if there are.

      if (.not. changed_this_pass) exit
      changed = .true.
      desired_level(1:grid%biggest_elem) = new_desired
   end do

! Exchange the desired h levels of partition-boundary elements with
! the neighbors.

   if (num_proc(procs) > 1 .and. .not. still_sequential) then
      call exchange_desired(loop,grid,procs,changed,part_bound_elements, &
                            npartbound,any_changes,desired_level)
   endif

end do ! until no processor made a change in its desireds

! Enforce p compatibility

if (refine_control%p_compatibility /= PCOMPAT_NONE) then

! Repeat until there are no more changes to my desireds.

   changed = .true.
   do while (changed)
      changed = .false.

! For each leaf element ...

      do i=1,nleaf
         elem = leaf_elements(i)

! If p compatibility is enforced by increasing p and elem's first child's p is
! too small relative to the neighbors, increase it.
! This is only for triangles, so p compatibility is not supported in 3D.

         if (refine_control%p_compatibility == PCOMPAT_INCREASE) then

            neighbor_degree = desired_degree(2,elem)
            if (neighbors(2,elem) /= BOUNDARY) then
               neighbor_degree = max(neighbor_degree, &
                                    maxval(desired_degree(:,neighbors(2,elem))))
            endif
            if (neighbors(3,elem) /= BOUNDARY) then
               neighbor_degree = max(neighbor_degree, &
                                     desired_degree(1,neighbors(3,elem)))
            endif
            if (desired_degree(1,elem) < neighbor_degree - &
                                         refine_control%max_p_diff) then
               desired_degree(1,elem) = neighbor_degree - &
                                        refine_control%max_p_diff
               changed = .true.
            endif

! Likewise for the second child

            neighbor_degree = desired_degree(1,elem)
            if (neighbors(1,elem) /= BOUNDARY) then
               neighbor_degree = max(neighbor_degree, &
                                    maxval(desired_degree(:,neighbors(1,elem))))
            endif
            if (neighbors(3,elem) /= BOUNDARY) then
               neighbor_degree = max(neighbor_degree, &
                                     desired_degree(2,neighbors(3,elem)))
            endif
            if (desired_degree(2,elem) < neighbor_degree - &
                                         refine_control%max_p_diff) then
               desired_degree(2,elem) = neighbor_degree - &
                                        refine_control%max_p_diff
               changed = .true.
            endif

! Except that for the second child, if elem is not going to be h-refined then
! the new degree should go in desired_degree(1,elem)

            if (desired_level(elem) == grid%element(elem)%level) then
               if (desired_degree(1,elem) < neighbor_degree - &
                                            refine_control%max_p_diff) then
                  desired_degree(1,elem) = neighbor_degree - &
                                           refine_control%max_p_diff
                  changed = .true.
               endif
            endif

! If p compatibility is enforced by decreaing p and elem's first child's p is
! too big relative to the neighbors, decrease it

         else ! p_compatibility == PCOMPAT_DECREASE

            neighbor_degree = desired_degree(2,elem)
            if (neighbors(2,elem) /= BOUNDARY) then
               neighbor_degree = min(neighbor_degree, &
                                    minval(desired_degree(:,neighbors(2,elem))))
            endif
            if (neighbors(3,elem) /= BOUNDARY) then
               neighbor_degree = min(neighbor_degree, &
                                     desired_degree(1,neighbors(3,elem)))
            endif
            if (desired_degree(1,elem) > neighbor_degree + &
                                         refine_control%max_p_diff) then
               desired_degree(1,elem) = neighbor_degree + &
                                        refine_control%max_p_diff
               changed = .true.
            endif

! Likewise for the second child

            neighbor_degree = desired_degree(1,elem)
            if (neighbors(1,elem) /= BOUNDARY) then
               neighbor_degree = min(neighbor_degree, &
                                    minval(desired_degree(:,neighbors(1,elem))))
            endif
            if (neighbors(3,elem) /= BOUNDARY) then
               neighbor_degree = min(neighbor_degree, &
                                     desired_degree(2,neighbors(3,elem)))
            endif
            if (desired_degree(2,elem) > neighbor_degree + &
                                         refine_control%max_p_diff) then
               desired_degree(2,elem) = neighbor_degree + &
                                        refine_control%max_p_diff
               changed = .true.
            endif

! Except that for the second child, if elem is not going to be h-refined then
! the new degree should go in desired_degree(1,elem)

            if (desired_level(elem) /= grid%element(elem)%level) then
               if (desired_degree(1,elem) > neighbor_degree + &
                                            refine_control%max_p_diff) then
                  desired_degree(1,elem) = neighbor_degree + &
                                           refine_control%max_p_diff
                  changed = .true.
               endif
            endif

         endif ! increase or decrease
      end do ! next leaf element
   end do ! until no changes
endif ! p compatibility

end subroutine reconcile_requests

!          ----------------
subroutine exchange_desired(loop,grid,procs,changed,part_bound_elements, &
                            npartbound,any_changes,desired_level,desired_degree)
!          ----------------

!----------------------------------------------------
! This routine exchanges the desired p and h levels of partition-boundary
! elements with the neighbors.  On input, if changed is false then there
! were no changes to desired_level on this processor so the data does not need
! to be passed.  On output, if changed is false then it is false on all
! processors, so we are done reconciling requests.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

integer, intent(in) :: loop
type(grid_type), intent(inout) :: grid
type(proc_info), intent(in) :: procs
logical, intent(inout) :: changed
integer, intent(in) :: part_bound_elements(:), npartbound
logical, intent(inout) :: any_changes
integer, intent(inout) :: desired_level(:)
integer, intent(inout), optional :: desired_degree(:,:)
!----------------------------------------------------
! Local variables:

integer :: my_processor, nproc, nisend, count, i, ip, lid,astat, istart, &
           ivert, nneigh, global_change(2)
integer, pointer :: neigh(:), neighproc(:)
logical :: on_boundary, need_alltoall
integer, allocatable :: isend(:), nrecv(:)
integer, pointer :: irecv(:)
!----------------------------------------------------
! Begin executable code

! Convenience variables.

my_processor = my_proc(procs)
nproc = num_proc(procs)

! Get the number of neighbor processors.  If the neighborhood has not been set
! up yet, use all processors.

call get_neighbor_procs(procs,neighproc)
if (associated(neighproc)) then
   nneigh = size(neighproc)
   need_alltoall = .false.
   deallocate(neighproc)
else
   nneigh = nproc
   need_alltoall = .true.
endif

! Allocate message.

nisend = 3 + (1+KEY_SIZE)*npartbound
if (present(desired_degree)) then
   nisend = nisend + 2*npartbound
endif
allocate(isend(nisend),nrecv(nproc),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in exchange_desired")
   stop
endif

! Pack the changed flag

if (changed) then
   isend(1) = 1
else
   isend(1) = 0
endif
if (any_changes) then
   isend(2) = 1
else
   isend(2) = 0
endif

! Reduce the changed flag

global_change = isend(1:2)
global_change = phaml_global_max(procs,global_change,3330+loop)
changed = global_change(1)==1
any_changes = global_change(2)==1

! Pack the number of elements

isend(3) = npartbound

! For each element in the list of partition-boundary elements, pack the
! element ID and desired values.

if (changed) then
   count = 4
   do i=1,npartbound
      call hash_pack_key(grid%element(part_bound_elements(i))%gid,isend,count)
      count = count + KEY_SIZE
      isend(count) = desired_level(part_bound_elements(i))
      count = count + 1
      if (present(desired_degree)) then
         isend(count:count+1) = desired_degree(:,part_bound_elements(i))
         count = count + 2
      endif
   end do
else
   isend(4:nisend) = 0
endif

! Exchange the message with the other processors.

if (my_processor /= MASTER) then

   if (need_alltoall) then
      call phaml_alltoall(procs,isend,nisend,irecv,nrecv,3310+loop)
   else
      call phaml_neighbor_alltoall(procs,isend,nisend,irecv,nrecv,3310+loop)
   endif

! Receive and process the messages.

   istart = 0
   do ip=1,nneigh
      if (ip == my_processor) then
         istart = istart + nrecv(ip)
         cycle
      endif

! If there are changes in this message, desired should be the maximum of
! what the owner of each element thinks and what I have, if the element is
! on the partition boundary.

      if (irecv(1+istart) == 1) then
         count = 4+istart
         do i=1,irecv(3+istart) ! irecv(3) is the number of elements sent
            lid = hash_decode_key(hash_unpack_key(irecv,count),grid%elem_hash)
            if (lid == HASH_NOT_FOUND) then
               count = count + KEY_SIZE+1
               if (present(desired_degree)) count = count + 2
            else
               on_boundary = .false.
               do ivert=1,VERTICES_PER_ELEMENT
                  call get_vertex_elements(grid, &
                                           grid%element(lid)%vertex(ivert),&
                                          neigh,include_periodic=.true.)
                  if (any(grid%element(neigh)%iown)) then
                     on_boundary = .true.
                  endif
                  deallocate(neigh)
                  if (on_boundary) exit
               end do
               count = count + KEY_SIZE
               if (on_boundary) then
                  desired_level(lid) = max(desired_level(lid),irecv(count))
                  count = count + 1
                  if (present(desired_degree)) then
                     desired_degree(1,lid) = max(desired_degree(1,lid), &
                                                 irecv(count))
                     desired_degree(2,lid) = max(desired_degree(2,lid), &
                                                 irecv(count+1))
                     count = count + 2
                  endif
               else
                  count = count + 1
                  if (present(desired_degree)) count = count + 2
               endif
            endif
         end do
      end if
      istart = istart + nrecv(ip)
   end do
   deallocate(irecv)
endif

end subroutine exchange_desired

!          --------------
subroutine p_coarsen_grid(grid,refine_control,solver_control,desired_degree, &
                          elem_list,nelem,any_changes)
!          --------------

!----------------------------------------------------
! This routine p-coarsens elements in elem_list whose current degree is larger
! than the desired degree.  NOTE: When this routine returns, the edge degree
! rule is not satisfied.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
integer, intent(in) :: desired_degree(:,:), elem_list(:), nelem
logical, intent(out) :: any_changes
!----------------------------------------------------
! Local variables:

integer :: i, errcode, delta_dof, delta_dof_own, total_delta_dof, &
           total_delta_dof_own
!----------------------------------------------------
! Begin executable code

any_changes = .false.

! p-coarsen the interior of each element that wants to be p coarsened.

total_delta_dof = 0
total_delta_dof_own = 0

!$omp parallel do schedule(dynamic) &
!$omp  default(shared) &
!$omp  private(i,errcode,delta_dof,delta_dof_own) &
!$omp  reduction(+ : total_delta_dof, total_delta_dof_own) &
!$omp  reduction(.or.: any_changes)

do i=1,nelem
   do while (grid%element(elem_list(i))%degree > desired_degree(1,elem_list(i)))
      call p_coarsen_elem_interior(grid,elem_list(i),errcode,refine_control, &
                                   solver_control,delta_dof,delta_dof_own)
      if (refine_control%refterm == MATCH_TOL) then
         if (grid%element(elem_list(i))%iown) then
            if (errcode == 0) any_changes = .true.
         endif
      else
         if (errcode == 0) any_changes = .true.
      endif
      total_delta_dof = total_delta_dof + delta_dof
      total_delta_dof_own = total_delta_dof_own + delta_dof_own
   end do
end do
!$omp end parallel do

! do things that must be done OpenMP-sequentially after the parallel loop

grid%dof = grid%dof + total_delta_dof
grid%dof_own = grid%dof_own + total_delta_dof_own

end subroutine p_coarsen_grid

!          --------------
subroutine h_coarsen_grid(grid,leaf_list,nleaf,desired_level,desired_degree, &
                          refine_control,solver_control,was_hcoarsened, &
                          any_changes)
!          --------------

!----------------------------------------------------
! This routine h-coarsens elements in elem_list whose current level is larger
! than the desired level.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
integer, intent(inout) :: leaf_list(:), nleaf, desired_level(:), &
                          desired_degree(:,:)
type(refine_options) :: refine_control
type(solver_options) :: solver_control
logical(small_logical), intent(inout) :: was_hcoarsened(:)
logical, intent(out) :: any_changes
!----------------------------------------------------
! Local variables:

integer :: i, lev, coarsen_list(grid%biggest_elem), ncoarsen, errcode, &
           delta_nelem, delta_nelem_leaf, delta_nelem_leaf_own, delta_nedge, &
           delta_nedge_leaf, delta_nedge_leaf_own, delta_nvert, &
           delta_nvert_own, delta_dof, delta_dof_own, &
           total_delta_nelem, total_delta_nelem_leaf, &
           total_delta_nelem_leaf_own, total_delta_nedge, &
           total_delta_nedge_leaf, total_delta_nedge_leaf_own, &
           total_delta_nvert, total_delta_nvert_own, total_delta_dof, &
           total_delta_dof_own
!----------------------------------------------------
! Begin executable code

any_changes = .false.

! For each level, finest to coarsest ...

do lev=grid%nlev,2,-1

! Make a list of the parents of elements of this level that want to be
! coarsened, but only include one of an element and its mate.

   call make_h_coarsen_list(grid,leaf_list,nleaf,lev,desired_level, &
                            desired_degree,coarsen_list,ncoarsen)

! Coarsen each element on the list.

   total_delta_nelem = 0
   total_delta_nelem_leaf = 0
   total_delta_nelem_leaf_own = 0
   total_delta_nedge = 0
   total_delta_nedge_leaf = 0
   total_delta_nedge_leaf_own = 0
   total_delta_nvert = 0
   total_delta_nvert_own = 0
   total_delta_dof = 0
   total_delta_dof_own = 0

!$omp parallel do &
!$omp  default(shared) &
!$omp  private(i,errcode,delta_nelem, delta_nelem_leaf, delta_nelem_leaf_own, &
!$omp          delta_nedge, delta_nedge_leaf, delta_nedge_leaf_own, &
!$omp          delta_nvert, delta_nvert_own, delta_dof, delta_dof_own) &
!$omp reduction(+ : total_delta_nelem, total_delta_nelem_leaf, &
!$omp               total_delta_nelem_leaf_own, total_delta_nedge, &
!$omp               total_delta_nedge_leaf, total_delta_nedge_leaf_own, &
!$omp               total_delta_nvert, total_delta_nvert_own, total_delta_dof, &
!$omp               total_delta_dof_own) &
!$omp  reduction(.or.: any_changes)

   do i=1,ncoarsen

! In 3D, not supporting the delta's, and cannot use OpenMP here

      if (global_element_kind == TETRAHEDRAL_ELEMENT) then
         if (using_OpenMP) then
            ierr = PHAML_INTERNAL_ERROR
            call fatal("h_coarsen_grid: cannot use OpenMP for 3D")
            stop
         endif
         call h_coarsen_element(grid,coarsen_list(i),errcode,refine_control, &
                                solver_control,was_hcoarsened,desired_level)
         delta_nelem = 0
         delta_nelem_leaf = 0
         delta_nelem_leaf_own = 0
         delta_nedge = 0
         delta_nedge_leaf = 0
         delta_nedge_leaf_own = 0
         delta_nvert = 0
         delta_nvert_own = 0
         delta_dof = 0
         delta_dof_own = 0

      else

         call h_coarsen_element(grid, coarsen_list(i), errcode, refine_control,&
                                solver_control, was_hcoarsened, desired_level, &
                                delta_nelem, delta_nelem_leaf, &
                                delta_nelem_leaf_own, delta_nedge, &
                                delta_nedge_leaf, delta_nedge_leaf_own, &
                                delta_nvert, delta_nvert_own, delta_dof, &
                                delta_dof_own)
      endif

      if (refine_control%refterm == MATCH_TOL) then
         if (grid%element(coarsen_list(i))%iown) then
            if (errcode == 0) any_changes = .true.
         elseif (.not. grid%element(coarsen_list(i))%mate == BOUNDARY) then
            if (grid%element(hash_decode_key(grid%element(coarsen_list(i))%mate,grid%elem_hash))%iown) then
               if (errcode == 0) any_changes = .true.
            endif
         endif
      else
         if (errcode == 0) any_changes = .true.
      endif
      total_delta_nelem = total_delta_nelem + delta_nelem
      total_delta_nelem_leaf = total_delta_nelem_leaf + delta_nelem_leaf
      total_delta_nelem_leaf_own = total_delta_nelem_leaf_own + delta_nelem_leaf_own
      total_delta_nedge = total_delta_nedge + delta_nedge
      total_delta_nedge_leaf = total_delta_nedge_leaf + delta_nedge_leaf
      total_delta_nedge_leaf_own = total_delta_nedge_leaf_own + delta_nedge_leaf_own
      total_delta_nvert = total_delta_nvert + delta_nvert
      total_delta_nvert_own = total_delta_nvert_own + delta_nvert_own
      total_delta_dof = total_delta_dof + delta_dof
      total_delta_dof_own = total_delta_dof_own + delta_dof_own

   end do
!$omp end parallel do

! Do things that must be done OpenMP-sequentially after the parallel loop.

   grid%nelem = grid%nelem + total_delta_nelem
   grid%nelem_leaf = grid%nelem_leaf + total_delta_nelem_leaf
   grid%nelem_leaf_own = grid%nelem_leaf_own + total_delta_nelem_leaf_own
   grid%nedge = grid%nedge + total_delta_nedge
   grid%nedge_leaf = grid%nedge_leaf + total_delta_nedge_leaf
   grid%nedge_leaf_own = grid%nedge_leaf_own + total_delta_nedge_leaf_own
   grid%nvert = grid%nvert + total_delta_nvert
   grid%nvert_own = grid%nvert_own + total_delta_nvert_own
   grid%dof = grid%dof + total_delta_dof
   grid%dof_own = grid%dof_own + total_delta_dof_own

end do ! level

end subroutine h_coarsen_grid

!          -------------------
subroutine make_h_coarsen_list(grid,leaf_list,nleaf,lev,desired_level, &
                               desired_degree,coarsen_list,ncoarsen)
!          -------------------

!----------------------------------------------------
! This routine makes a list of the parents of elements of level lev that
! want to be coarsened, but only includes one of an element and its mate.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(in) :: grid
integer, intent(inout) :: leaf_list(:), nleaf
integer, intent(in) :: lev
integer, intent(inout) :: desired_level(:), desired_degree(:,:)
integer, intent(out) :: coarsen_list(:), ncoarsen
!----------------------------------------------------
! Local variables:

integer :: i, elem, parent, mate, new_nleaf, sibling
logical(small_logical) :: onlist(grid%biggest_elem)
!----------------------------------------------------
! Begin executable code

ncoarsen = 0

! Mark all elements as not on the list

onlist = .false.

! For each element on the list that has level lev ...

new_nleaf = nleaf
do i=1,nleaf
   elem = leaf_list(i)
   if (grid%element(elem)%level /= lev) cycle

! See if it wants to be coarsened

   if (grid%element(elem)%level <= desired_level(elem)) cycle

! Identify the parent and mate
! Note: in 3D the elements that share a refinement edge are not
!       necessarily on the same level, so it is possible for the parent to
!       not exist, hence the HASH_NOT_FOUND test.

   parent = hash_decode_key(grid%element(elem)%gid/MAX_CHILD,grid%elem_hash)
   if (parent == HASH_NOT_FOUND) cycle
   if (.not. grid%element(parent)%mate == BOUNDARY) then
      mate = hash_decode_key(grid%element(parent)%mate,grid%elem_hash)
   else
      mate = BOUNDARY
   endif

! If either the parent or the mate is already on the list, skip.

   if (onlist(parent)) cycle
   if (mate /= BOUNDARY) then
      if (onlist(mate)) cycle
   endif

! Add the parent to the list.

   ncoarsen = ncoarsen + 1
   coarsen_list(ncoarsen) = parent
   onlist(parent) = .true.

! Add the parent and mate to the leaf list, and assign desired level and degree

   leaf_list(new_nleaf+1) = parent
   desired_level(parent) = desired_level(elem)
   desired_degree(:,parent) = desired_degree(1,elem)
   if (mate == BOUNDARY) then
      new_nleaf = new_nleaf + 1
   else
      new_nleaf = new_nleaf + 2
      leaf_list(new_nleaf) = mate
      desired_level(mate) = desired_level(elem)
      sibling = get_child_lid(grid%element(parent)%mate,1,grid%elem_hash)
      desired_degree(:,mate) = desired_degree(1,sibling)
   endif

end do

nleaf = new_nleaf

end subroutine make_h_coarsen_list

!          -------------
subroutine p_refine_grid(grid,refine_control,solver_control,desired_degree, &
                         elem_list,nelem,any_changes)
!          -------------

!----------------------------------------------------
! This routine p-refines elements in elem_list for which the current degree is
! less than the desired degree.  The edge rule does not have to be satisfied
! on entry, but it will be satisfied on return.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
integer, intent(in) :: desired_degree(:,:), elem_list(:), nelem
logical, intent(out) :: any_changes
!----------------------------------------------------
! Local variables:

integer :: total_delta_dof, total_delta_dof_own, i, errcode, delta_dof, &
           delta_dof_own
!----------------------------------------------------
! Begin executable code

any_changes = .false.

! p-refine the interior of each element that wants to be p refined.

total_delta_dof = 0
total_delta_dof_own = 0

!$omp parallel &
!$omp  default(shared) &
!$omp  private(i,errcode,delta_dof,delta_dof_own) &
!$omp  reduction(+ : total_delta_dof, total_delta_dof_own) &
!$omp  reduction(.or.: any_changes)

!$omp do schedule(dynamic)
do i=1,nelem
   do while (grid%element(elem_list(i))%degree < desired_degree(1,elem_list(i)))
      call p_refine_element_interior(grid,refine_control,elem_list(i), &
                                     errcode,delta_dof,delta_dof_own)
      if (refine_control%refterm == MATCH_TOL) then
         if (grid%element(elem_list(i))%iown) then
            if (errcode == 0) any_changes = .true.
         elseif (.not. grid%element(elem_list(i))%mate == BOUNDARY) then
            if (grid%element(hash_decode_key(grid%element(elem_list(i))%mate,grid%elem_hash))%iown) then
               if (errcode == 0) any_changes = .true.
            endif
         endif
      else
         if (errcode == 0) any_changes = .true.
      endif
      if (errcode == 0) any_changes = .true.
      total_delta_dof = total_delta_dof + delta_dof
      total_delta_dof_own = total_delta_dof_own + delta_dof_own
   end do
end do
!$omp end do

!$omp end parallel

! do things that must be done OpenMP-sequentially after the parallel loop

grid%dof = grid%dof + total_delta_dof
grid%dof_own = grid%dof_own + total_delta_dof_own

! Enforce the edge rule and update the error indicators

call enforce_and_errind(grid,refine_control,solver_control,elem_list,nelem)

end subroutine p_refine_grid

!          ------------------
subroutine enforce_and_errind(grid,refine_control,solver_control,elem_list, &
                              nelem)
!          ------------------

!----------------------------------------------------
! This routine enforces the edge rule and updates the error indicators.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
integer, intent(in) :: elem_list(:), nelem
!----------------------------------------------------
! Local variables:

integer :: total_delta_dof, total_delta_dof_own, i, delta_dof, &
           delta_dof_own, edge_list(grid%biggest_edge), nedge, &
           face_list(grid%biggest_face), nface
!----------------------------------------------------
! Begin executable code

! make a list of all the edges and faces that no longer satisfy the edge rule

call list_edges_without_rule(grid,refine_control,elem_list,nelem,edge_list, &
                             nedge)

if (global_element_kind == TETRAHEDRAL_ELEMENT) then
   call list_faces_without_rule(grid,refine_control,elem_list,nelem,face_list, &
                                nface)
else
   nface = 0
endif

total_delta_dof = 0
total_delta_dof_own = 0

!$omp parallel &
!$omp  default(shared) &
!$omp  private(i,delta_dof,delta_dof_own) &
!$omp  reduction(+ : total_delta_dof, total_delta_dof_own)

! enforce the edge rule on the listed edges

!$omp do
do i=1,nedge
   call enforce_edge_rule(grid,refine_control,edge_list(i),delta_dof, &
                          delta_dof_own)
   total_delta_dof = total_delta_dof + delta_dof
   total_delta_dof_own = total_delta_dof_own + delta_dof_own
end do
!$omp end do

! enforce the face rule on the listed faces

!$omp do
do i=1,nface
   call enforce_face_rule(grid,refine_control,face_list(i),delta_dof, &
                          delta_dof_own)
   total_delta_dof = total_delta_dof + delta_dof
   total_delta_dof_own = total_delta_dof_own + delta_dof_own
end do
!$omp end do

!$omp end parallel

! do things that must be done OpenMP-sequentially after the parallel loop

grid%dof = grid%dof + total_delta_dof
grid%dof_own = grid%dof_own + total_delta_dof_own

! compute error indicators for entities that have changed

call all_error_indicators(grid,refine_control%error_estimator,solver_control, &
                          delayed=.true.)

end subroutine enforce_and_errind

!          -------------
subroutine h_refine_grid(grid,refine_control,solver_control,desired_level, &
                         desired_degree,elem_list,nelem,was_hcoarsened, &
                         any_changes)
!          -------------

!----------------------------------------------------
! This routine h-refines elements in elem_list for which the current level is
! less than the desired level.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
integer, pointer :: desired_level(:), desired_degree(:,:), elem_list(:)
integer, intent(in) :: nelem
logical(small_logical), pointer :: was_hcoarsened(:)
logical, intent(out) :: any_changes
!----------------------------------------------------
! Local variables:

integer :: lev, elem, errcode, mate, i, nrefine, newsize, &
           delta_nelem, delta_nelem_leaf, delta_nelem_leaf_own, &
           delta_nedge, delta_nedge_leaf, delta_nedge_leaf_own, &
           delta_nvert, delta_nvert_own, delta_dof, delta_dof_own, &
           total_delta_nelem, total_delta_nelem_leaf, &
           total_delta_nelem_leaf_own, total_delta_nedge, &
           total_delta_nedge_leaf, total_delta_nedge_leaf_own, &
           total_delta_nvert, total_delta_nvert_own, total_delta_dof, &
           total_delta_dof_own, max_nlev, max_max_nlev, astat, next_ind
integer, allocatable :: refine_list(:), vert_lid(:,:), edge_lid(:,:), &
                        elem_lid(:,:)
logical(small_logical), allocatable :: onlist(:)
logical :: reallocated
logical(small_logical), pointer :: been_there(:)

!----------------------------------------------------
! Begin executable code

! Cannot use OpenMP in 3D because the deltas are not supported

if (using_OpenMP .and. global_element_kind == TETRAHEDRAL_ELEMENT) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("h_refine_grid: Cannot use OpenMP in 3D")
   stop
endif

any_changes = .false.

! Allocate refine_list and lid lists

allocate(refine_list(grid%biggest_elem), vert_lid(2,grid%biggest_elem), &
         edge_lid(6,grid%biggest_elem), elem_lid(4,grid%biggest_elem), &
         stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in h_refine_grid")
   stop
endif

! Allocate onlist

allocate(onlist(size(grid%element)),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in h_refine_grid")
   stop
endif

! Note: grid%dof etc must not be used in before_h_refine or after_h_refine
! because the reduction does not occur until the end of the parallel section

total_delta_nelem = 0
total_delta_nelem_leaf = 0
total_delta_nelem_leaf_own = 0
total_delta_nedge = 0
total_delta_nedge_leaf = 0
total_delta_nedge_leaf_own = 0
total_delta_nvert = 0
total_delta_nvert_own = 0
total_delta_dof = 0
total_delta_dof_own = 0
max_max_nlev = 0

!$omp parallel &
!$omp  default(shared) &
!$omp  private(lev, i, errcode, delta_nelem, delta_nelem_leaf, &
!$omp    delta_nelem_leaf_own, delta_nedge, delta_nedge_leaf, &
!$omp    delta_nedge_leaf_own, delta_nvert, delta_nvert_own, delta_dof, &
!$omp    delta_dof_own, max_nlev, been_there) &
!$omp  reduction(+ : total_delta_nelem, total_delta_nelem_leaf, &
!$omp    total_delta_nelem_leaf_own, total_delta_nedge, total_delta_nedge_leaf,&
!$omp    total_delta_nedge_leaf_own, total_delta_nvert, total_delta_nvert_own, &
!$omp    total_delta_dof, total_delta_dof_own) &
!$omp  reduction(max : max_max_nlev) &
!$omp  reduction(.or.: any_changes)

if (global_element_kind == TETRAHEDRAL_ELEMENT) then
   allocate(been_there(size(grid%element)),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("memory allocation failed in h_refine_grid")
      stop
   endif
   been_there = .false.
endif

! for each level, starting with the coarsest level ...

next_ind = 1
do lev=1,grid%nlev

!$omp barrier
!$omp single

! Make sure onlist is still big enough.

   if (size(onlist) /= size(grid%element)) then
      deallocate(onlist)
      allocate(onlist(size(grid%element)),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in h_refine_grid")
         stop
      endif
   endif

! Create a list of elements on this level to be refined.
! There is a slight possibility that refine_list and the lid lists will
! come up short.  If that happens, double the allocation and start over.

! Note that elem_list was created by list_elements, which creates the list
! level by level starting with level=1, so we can find the elements of the
! current level by starting at the end of the last level.

   reallocated = .true.
   do while (reallocated)
      reallocated = .false.
      onlist(1:grid%biggest_elem) = .false.
      nrefine = 0
      do i=next_ind,nelem
         elem = elem_list(i)
         if (grid%element(elem)%level /= lev) then
            next_ind = i
            exit
         endif
         if (grid%element(elem)%level < desired_level(elem)) then
            if (onlist(elem)) cycle
            if (.not. grid%element(elem)%mate == BOUNDARY) then
               mate = hash_decode_key(grid%element(elem)%mate,grid%elem_hash)
            else
               mate = BOUNDARY
            endif
            if (mate /= BOUNDARY) then
               if (onlist(mate)) cycle
            endif
            nrefine = nrefine + 1
            if (nrefine > size(refine_list)) then
               newsize = 2*size(refine_list)
               deallocate(refine_list,vert_lid,edge_lid,elem_lid)
               allocate(refine_list(newsize), vert_lid(2,newsize), &
                        edge_lid(6,newsize), elem_lid(4,newsize), stat=astat)
               if (astat /= 0) then
                  ierr = ALLOC_FAILED
                  call fatal("memory allocation failed in h_refine_grid")
                  stop
               endif
               reallocated = .true.
               exit
            endif
            refine_list(nrefine) = elem
            onlist(elem) = .true.
         endif
         if (i == nelem) next_ind = nelem+1
      end do
   end do

! Get lids for the children, and other things that must be done
! OpenMP-sequentially before the parallel loop.

   if (nrefine /= 0) then
      call before_h_refine(grid,refine_list,nrefine,vert_lid=vert_lid, &
                           edge_lid=edge_lid,elem_lid=elem_lid, &
                           desired_level=desired_level, &
                           desired_degree=desired_degree, &
                           elem_list=elem_list,was_hcoarsened=was_hcoarsened)
   endif

!$omp end single

! If the list is empty, move on to the next level.

   if (nrefine == 0) cycle

!$omp do
   do i=1,nrefine
      if (global_element_kind == TETRAHEDRAL_ELEMENT) then
         call h_refine_element(grid,refine_list(i),errcode,refine_control,&
                   solver_control,vert_child_lid=vert_lid(:,i), &
                   edge_child_lid=edge_lid(:,i),elem_child_lid=elem_lid(:,i), &
                   max_nlev=max_nlev,was_hcoarsened=was_hcoarsened, &
                   desired_level=desired_level,desired_degree=desired_degree, &
                   elem_list=elem_list,use_compatibility_chain=.true., &
                   been_there=been_there)
         delta_nelem = 0
         delta_nelem_leaf = 0
         delta_nelem_leaf_own = 0
         delta_nedge = 0
         delta_nedge_leaf = 0
         delta_nedge_leaf_own = 0
         delta_nvert = 0
         delta_nvert_own = 0
         delta_dof = 0
         delta_dof_own = 0

      else
         call h_refine_element(grid,refine_list(i),errcode,refine_control,&
                   solver_control,vert_child_lid=vert_lid(:,i), &
                   edge_child_lid=edge_lid(:,i),elem_child_lid=elem_lid(:,i), &
                   delta_nelem=delta_nelem,delta_nelem_leaf=delta_nelem_leaf, &
                   delta_nelem_leaf_own=delta_nelem_leaf_own, &
                   delta_nedge=delta_nedge,delta_nedge_leaf=delta_nedge_leaf, &
                   delta_nedge_leaf_own=delta_nedge_leaf_own, &
                   delta_nvert=delta_nvert,delta_nvert_own=delta_nvert_own, &
                   delta_dof=delta_dof,delta_dof_own=delta_dof_own, &
                   max_nlev=max_nlev,was_hcoarsened=was_hcoarsened, &
                   desired_level=desired_level,desired_degree=desired_degree, &
                   elem_list=elem_list,use_compatibility_chain=.true., &
                   been_there=been_there)
      endif

      if (refine_control%refterm == MATCH_TOL) then
         if (grid%element(refine_list(i))%iown) then
            if (errcode == 0) any_changes = .true.
         elseif (.not. grid%element(refine_list(i))%mate == BOUNDARY) then
            if (grid%element(hash_decode_key(grid%element(refine_list(i))%mate,grid%elem_hash))%iown) then
               if (errcode == 0) any_changes = .true.
            endif
         endif
      else
         if (errcode == 0) any_changes = .true.
      endif
      total_delta_nelem = total_delta_nelem + delta_nelem
      total_delta_nelem_leaf = total_delta_nelem_leaf + delta_nelem_leaf
      total_delta_nelem_leaf_own = total_delta_nelem_leaf_own + delta_nelem_leaf_own
      total_delta_nedge = total_delta_nedge + delta_nedge
      total_delta_nedge_leaf = total_delta_nedge_leaf + delta_nedge_leaf
      total_delta_nedge_leaf_own = total_delta_nedge_leaf_own + delta_nedge_leaf_own
      total_delta_nvert = total_delta_nvert + delta_nvert
      total_delta_nvert_own = total_delta_nvert_own + delta_nvert_own
      total_delta_dof = total_delta_dof + delta_dof
      total_delta_dof_own = total_delta_dof_own + delta_dof_own
      max_max_nlev = max(max_max_nlev,max_nlev)

   end do
!$omp end do

! Do things that must be done OpenMP-sequentially after the parallel loop.

!$omp single

   call after_h_refine(grid,refine_list,nrefine,vert_lid,edge_lid,elem_lid, &
                       refine_control,solver_control)

!$omp end single

end do

if (global_element_kind == TETRAHEDRAL_ELEMENT) then
   deallocate(been_there)
endif

!$omp end parallel

! total_* are reduced at end parallel, now add them to grid

grid%nelem = grid%nelem + total_delta_nelem
grid%nelem_leaf = grid%nelem_leaf + total_delta_nelem_leaf
grid%nelem_leaf_own = grid%nelem_leaf_own + total_delta_nelem_leaf_own
grid%nedge = grid%nedge + total_delta_nedge
grid%nedge_leaf = grid%nedge_leaf + total_delta_nedge_leaf
grid%nedge_leaf_own = grid%nedge_leaf_own + total_delta_nedge_leaf_own
grid%nvert = grid%nvert + total_delta_nvert
grid%nvert_own = grid%nvert_own + total_delta_nvert_own
grid%dof = grid%dof + total_delta_dof
grid%dof_own = grid%dof_own + total_delta_dof_own
grid%nlev = max(grid%nlev,max_max_nlev)

! Free memory.

deallocate(onlist,refine_list,vert_lid,edge_lid,elem_lid)

end subroutine h_refine_grid

!          -------------------------
subroutine refedge_based_refine_grid(grid,refine_control,solver_control, &
                                     procs,still_sequential,no_time)
!          -------------------------

!----------------------------------------------------
! This routine refines the grid OpenMP parallel using a task-based approach
! with each task refining (all the elements around) a refinement edge.
! This can be used for either h-uniform or h-adaptive refinement.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
type(proc_info) :: procs
logical, intent(in) :: still_sequential
logical, intent(in), optional :: no_time

!----------------------------------------------------
! Local variables:

integer :: i, nleaf, errcode, refedge, jerr, astat
integer, allocatable :: leaf_elements(:)
real(my_real) :: max_errind, refine_cutoff
integer(omp_lock_kind), allocatable :: refedge_locks(:)
logical(small_logical), allocatable :: lock_initialized(:)
logical :: timeit, reallocated

!----------------------------------------------------
! Begin executable code

if (present(no_time)) then
   timeit = .not. no_time
else
   timeit = .true.
endif

errcode = 0

! If we run out of space in grid%vertices, edge or elements, or need to
! increase the allocation for levels, the reallocation has to occur outside
! the OpenMP parallel region.  In that case, h_refine_refinement_edge aborts
! and returns error code 4.  Here we then exit the loop over leaves, leave the
! parallel section, increase the allocation, and start over again.  We will not
! refine new leaves created in the previous outermost loop because the child
! error indicators are set to -1 in h_refine_refinement_edge.

reallocated = .true.
do while(reallocated)
   reallocated = .false.

   allocate(refedge_locks(size(grid%edge)),lock_initialized(size(grid%edge)), &
            leaf_elements(grid%biggest_elem),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("refedge_based_refine_grid: memory allocation failed")
      stop
   endif

! Get the leaf elements.

   call list_elements(grid,leaf_elements,nleaf,leaf=.true.)

! Compute error indicators and maximum error indicator.

   if (refine_control%reftype /= H_UNIFORM) then

      if (.not. grid%errind_up2date) then

         call all_error_indicators(grid,refine_control%error_estimator, &
                                   solver_control,leaf_elements=leaf_elements, &
                                   nleaf=nleaf,max_errind=max_errind, &
                                   refine_control=refine_control)
         max_errind = phaml_global_max(procs,max_errind,3323)

      else

         max_errind = compute_global_max_errind(grid,refine_control,procs, &
                                                still_sequential,leaf_elements,&
                                                nleaf,.true.)

      endif

      refine_cutoff = max_errind / refine_control%reftol_factor

   else

! For h-uniform refinement, set the refinement cutoff to -huge(0) so that all
! error indicators are larger.

      refine_cutoff = -huge(0.0_my_real)

   endif

   lock_initialized = .false.

!$omp parallel &
!$omp  default(shared) &
!$omp  private(refedge,jerr)

   jerr = 0

!$omp do
   do i=1,nleaf
      if (errcode == 2 .or. errcode == 4) cycle
      if (grid%element(leaf_elements(i))%iown .and. &
          any(grid%element_errind(leaf_elements(i),:) > refine_cutoff)) then

! Create a lock for the refinement edge to prevent multiple threads from
! refining the same edge.  The lock is set/unset/tested in
! h_refine_refinement_edge.

!$omp critical
         if (.not. lock_initialized( &
                      grid%element(leaf_elements(i))%refinement_edge)) then
            lock_initialized(grid%element(leaf_elements(i))%refinement_edge) = &
               .true.
            call omp_init_lock(refedge_locks( &
                                grid%element(leaf_elements(i))%refinement_edge))
         endif
!$omp end critical

         refedge = grid%element(leaf_elements(i))%refinement_edge
!$omp task
         call h_refine_refinement_edge(grid,refedge,refine_control, &
                                       solver_control,refedge_locks, &
                                       lock_initialized,jerr)
!$omp atomic
         errcode = max(errcode,jerr)
!$omp end task

      endif
   end do
!$omp end do
!$omp taskwait

! Destroy the locks

!$omp do
   do i=1,size(lock_initialized)
      if (lock_initialized(i)) then
         call omp_destroy_lock(refedge_locks(i))
      endif
   end do
!$omp end do

!$omp end parallel

   deallocate(refedge_locks,lock_initialized)

! Check for needing reallocation

   if (errcode == 4) then
      errcode = 0
      deallocate(leaf_elements)
      call do_reallocations(grid,errcode)
      if (errcode /= 0) then
         ierr = ALLOC_FAILED
         call fatal("refedge_based_refine_grid: memory reallocation failed")
         stop
      endif
      reallocated = .true.
   endif

end do

end subroutine refedge_based_refine_grid

!          ----------------
subroutine do_reallocations(grid,errcode)
!          ----------------

!----------------------------------------------------
! This routine increases allocation for any of vertices, edges, elements and
! levels for which the current use is close to the current allocation.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
integer, intent(out) :: errcode

!----------------------------------------------------
! Local variables:

integer :: jerr1, jerr2, jerr3

!----------------------------------------------------
! Begin executable code

jerr1 = 0; jerr2 = 0; jerr3 = 0

! Increase entities for which the current usage is more than 3/4 the allocation

if (grid%nlev > (3*size(grid%head_level_elem))/4) call extend_nlev(grid)
if (grid%biggest_vert > (3*size(grid%vertex))/4) call more_verts(grid,jerr1)
if (grid%biggest_edge > (3*size(grid%edge))/4) call more_edges(grid,jerr2)
if (grid%biggest_elem > (3*size(grid%element))/4) call more_elements(grid,jerr3)

errcode = max(max(jerr1,jerr2),jerr3)

end subroutine do_reallocations

!          -----------------
subroutine pred_load_balance(linsys,grid,procs,refine_control,solver_control, &
                             lb,partition_method,still_sequential,balance_what,&
                             desired_level,desired_degree,was_hcoarsened, &
                             any_changes)
!          -----------------

!----------------------------------------------------
! This routine performs predictive load balancing based on desired_level and
! desired_degree.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linsys_type), intent(inout) :: linsys
type(grid_type), intent(inout) :: grid
type(proc_info), intent(inout) :: procs
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
type(Zoltan_Struct), pointer :: lb
integer, intent(in) :: partition_method
logical, intent(in) :: still_sequential
integer, intent(in) :: balance_what
integer, pointer :: desired_level(:), desired_degree(:,:)
logical(small_logical), pointer :: was_hcoarsened(:)
logical, intent(out) :: any_changes
!----------------------------------------------------
! Local variables:

type(hash_key), pointer :: export_gid(:)
integer, pointer :: export_proc(:)
logical :: repart, redist
integer :: nentity, minentity, maxentity, numexp

!----------------------------------------------------
! Begin executable code

any_changes = .false.
nullify(export_gid,export_proc)

! see if repartitioning is really necessary

! some cases where partitioning is not needed:
!   master does not participate
!   number of processors is 1
!   program compiled for sequential execution
!   balancing nothing

if (my_proc(procs) == MASTER .or. num_proc(procs) == 1 .or. &
    PARALLEL==SEQUENTIAL .or. balance_what==BALANCE_NONE) then
   repart = .false.

! otherwise, partition if the load is sufficiently out of balance

else
   select case (balance_what)
   case (BALANCE_ELEMENTS)
      call get_grid_info(grid,procs,still_sequential,3360, &
                         nelem_leaf_own=nentity,no_master=.true.)
   case (BALANCE_VERTICES)
      call get_grid_info(grid,procs,still_sequential,3360, &
                         nvert_own=nentity,no_master=.true.)
   case (BALANCE_EQUATIONS)
      call get_grid_info(grid,procs,still_sequential,3360, &
                         dof_own=nentity,no_master=.true.)
   case default
      call fatal("bad value for balance what in subroutine pred_load_balance")
   end select
   minentity = phaml_global_min(procs,nentity,3370)
   maxentity = phaml_global_max(procs,nentity,3380)
   repart = ( (maxentity-minentity)/float(maxentity) > 0.05)
endif

! partition

if (repart) then
   call lightweight_set_weights(grid,balance_what,desired_level,desired_degree)
   call partition(linsys,grid,procs,lb,refine_control,solver_control,.true., &
                  partition_method,balance_what,num_proc(procs),export_gid, &
                  export_proc,weights_already_set=.true.)

! redistribute

! see if enough elements are moved to be worthwhile
   if (associated(export_gid)) then
      numexp = size(export_gid)
   else
      numexp = 0
   endif
   numexp = phaml_global_sum(procs,numexp,3390)
   call get_grid_info(grid,procs,still_sequential,3395, &
                      total_nelem_leaf=nentity,no_master=.true.)
   redist = (numexp/float(nentity) > .05)
   if (redist) then
      call redistribute(grid,procs,refine_control,solver_control,export_gid, &
                        export_proc,desired_level,desired_degree,was_hcoarsened)
      call reconcile(grid,procs,refine_control,solver_control, &
                     still_sequential,desired_level,desired_degree, &
                     was_hcoarsened)
      any_changes = .true.
   endif
   if (associated(export_gid)) then
      deallocate(export_gid,export_proc)
   endif

endif

end subroutine pred_load_balance

!        ------------------
function pcoarsen_indicator(grid,elem)
!        ------------------

!----------------------------------------------------
! This routine computes the p coarsening error indicator for element elem
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(in) :: grid
integer, intent(in) :: elem
real(my_real) :: pcoarsen_indicator
!----------------------------------------------------
! Local variables:

integer :: i, j, k, deg
real(my_real) :: temp
!----------------------------------------------------
! Begin executable code

! Compute the l2 norm of the p-hierarchic coefficients of the current
! degree of this element.  Don't worry about whether the edges are actually
! of the same degree.
! For systems or multiple eigenvectors, use the max of the individual solutions.

deg = grid%element(elem)%degree

if (deg <= 1) then
   pcoarsen_indicator = huge(0.0_my_real)
   return
endif

pcoarsen_indicator = 0

do j=1,grid%system_size
 do k=1,max(1,grid%num_eval)
   temp = 0
   do i = element_dof(deg-1)+1, element_dof(deg)
      temp = temp + get_elem_soln(grid,elem,i,j,k)**2
   end do
   do i=1,EDGES_PER_ELEMENT
      if (grid%edge(grid%element(elem)%edge(i))%degree >= deg) then
         temp = temp + get_edge_soln(grid,grid%element(elem)%edge(i), &
                                     deg-1,j,k)**2
      endif
   end do
   pcoarsen_indicator = max(pcoarsen_indicator,sqrt(temp))
 end do
end do

end function pcoarsen_indicator

!        ------------------
function hcoarsen_indicator(grid,parent,children,refine_control)
!        ------------------

!----------------------------------------------------
! This routine computes the h coarsening error indicator for the elements in
! children, which should be siblings that are children of parent and its mate.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
integer, intent(in) :: parent, children(:)
type(refine_options), intent(in) :: refine_control
real(my_real) :: hcoarsen_indicator
!----------------------------------------------------
! Local variables:

real(my_real), allocatable :: solution(:,:)
real(my_real) :: xvert(4),yvert(4),temp
integer :: i,j,k,degree,astat,isub
!----------------------------------------------------
! Begin executable code

! Separate code for 3D

if (global_element_kind == TETRAHEDRAL_ELEMENT) then
   hcoarsen_indicator = hcoarsen_indicator_3D(grid,parent,children, &
                                              refine_control)
   return
endif

! For MATCH_TOL termination, use the maximum of the parent's children's error
! indicators, and maximum over all eigenvalues

if (refine_control%refterm == MATCH_TOL) then
   hcoarsen_indicator = max(maxval(grid%element_errind(children(1),:)), &
                            maxval(grid%element_errind(children(2),:)))
   return
endif

! set degree to be the maximum degree, and allocate space for the local
! copy of the solution coefficients

degree = 0
do i=1,4
  if (i==3 .and. children(3)==NO_CHILD) exit
  do j=1,EDGES_PER_ELEMENT
     degree = max(degree,grid%edge(grid%element(children(i))%edge(j))%degree)
  end do
end do

allocate(solution((degree+1)**2+degree**2,grid%nsoln),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in hcoarsen_indicator")
   return
endif

! copy solution coefficients to the local variable.  see subroutine
! phier2nodal for the order of the solution components

solution = 0
solution(1,:) = reshape( &
                get_vert_soln(grid,grid%element(children(1))%vertex(1),colon, &
                              colon),(/grid%nsoln/))
solution(2,:) = reshape( &
                get_vert_soln(grid,grid%element(children(2))%vertex(1),colon, &
                              colon),(/grid%nsoln/))
solution(3,:) = reshape( &
                get_vert_soln(grid,grid%element(children(1))%vertex(2),colon, &
                              colon),(/grid%nsoln/))
if (children(3) /= NO_CHILD) then
   solution(4,:) = reshape( &
                get_vert_soln(grid,grid%element(children(3))%vertex(2),colon, &
                              colon),(/grid%nsoln/))
endif
isub = 5
do i=1,degree-1
   if (grid%edge(grid%element(children(1))%edge(1))%degree >= i+1) then
      solution(isub,:) = reshape(get_edge_soln(grid, &
                grid%element(children(1))%edge(1),i,colon,colon),(/grid%nsoln/))
   endif
   isub = isub+1
end do
solution(isub,:) = reshape( &
                get_vert_soln(grid,grid%element(children(1))%vertex(3),colon, &
                              colon),(/grid%nsoln/))
isub = isub+1
do i=1,degree-1
   if (grid%edge(grid%element(children(1))%edge(2))%degree >= i+1) then
      solution(isub,:) = reshape(get_edge_soln(grid, &
                grid%element(children(1))%edge(2),i,colon,colon),(/grid%nsoln/))
   endif
   isub = isub + 1
end do
do i=1,degree-1
   if (grid%edge(grid%element(children(1))%edge(3))%degree >= i+1) then
      solution(isub,:) = reshape(get_edge_soln(grid, &
                grid%element(children(1))%edge(3),i,colon,colon),(/grid%nsoln/))
   endif
   isub = isub + 1
end do
k = 1
do j=1,degree-2
   do i=1,j
      if (grid%element(children(1))%degree >= j+2) then
         solution(isub,:) = reshape(get_elem_soln(grid,children(1),k,colon, &
                                    colon),(/grid%nsoln/))
      endif
      isub = isub + 1
      k = k+1
   end do
end do
do i=1,degree-1
   if (grid%edge(grid%element(children(2))%edge(2))%degree >= i+1) then
      solution(isub,:) = reshape(get_edge_soln(grid, &
                grid%element(children(2))%edge(2),i,colon,colon),(/grid%nsoln/))
   endif
   isub = isub + 1
end do
do i=1,degree-1
   if (grid%edge(grid%element(children(2))%edge(3))%degree >= i+1) then
      solution(isub,:) = reshape(get_edge_soln(grid, &
                grid%element(children(2))%edge(3),i,colon,colon),(/grid%nsoln/))
   endif
   isub = isub + 1
end do
k = 1
do j=1,degree-2
   do i=1,j
      if (grid%element(children(2))%degree >= j+2) then
         solution(isub,:) = reshape(get_elem_soln(grid,children(2),k,colon, &
                                    colon),(/grid%nsoln/))
      endif
      isub = isub + 1
      k = k+1
   end do
end do
if (children(3) /= NO_CHILD) then
   do i=1,degree-1
      if (grid%edge(grid%element(children(3))%edge(1))%degree >= i+1) then
         solution(isub,:) = reshape(get_edge_soln(grid, &
                grid%element(children(3))%edge(1),i,colon,colon),(/grid%nsoln/))
      endif
      isub = isub + 1
   end do
   do i=1,degree-1
      if (grid%edge(grid%element(children(3))%edge(3))%degree >= i+1) then
         solution(isub,:) = reshape(get_edge_soln(grid, &
                grid%element(children(3))%edge(3),i,colon,colon),(/grid%nsoln/))
      endif
      isub = isub + 1
   end do
   k = 1
   do j=1,degree-2
      do i=1,j
         if (grid%element(children(3))%degree >= j+2) then
            solution(isub,:) = reshape(get_elem_soln(grid,children(3),k,colon, &
                                       colon),(/grid%nsoln/))
         endif
         isub = isub + 1
         k = k+1
      end do
   end do
   do i=1,degree-1
      if (grid%edge(grid%element(children(4))%edge(3))%degree >= i+1) then
         solution(isub,:) = reshape(get_edge_soln(grid, &
                grid%element(children(4))%edge(3),i,colon,colon),(/grid%nsoln/))
      endif
      isub = isub + 1
   end do
   k = 1
   do j=1,degree-2
      do i=1,j
         if (grid%element(children(4))%degree >= j+2) then
            solution(isub,:) = reshape(get_elem_soln(grid,children(4),k,colon, &
                                       colon),(/grid%nsoln/))
         endif
         isub = isub + 1
         k = k+1
      end do
   end do
endif

! set the outer vertices of the parents

xvert(1) = grid%vertex(grid%element(parent)%vertex(1))%coord%x
yvert(1) = grid%vertex(grid%element(parent)%vertex(1))%coord%y
xvert(2) = grid%vertex(grid%element(parent)%vertex(2))%coord%x
yvert(2) = grid%vertex(grid%element(parent)%vertex(2))%coord%y
xvert(3) = grid%vertex(grid%element(parent)%vertex(3))%coord%x
yvert(3) = grid%vertex(grid%element(parent)%vertex(3))%coord%y
if (children(3) == NO_CHILD) then
   xvert(4) = huge(0.0_my_real)
   yvert(4) = huge(0.0_my_real)
else
   xvert(4) = grid%vertex(grid%element(children(3))%vertex(2))%coord%x
   yvert(4) = grid%vertex(grid%element(children(3))%vertex(2))%coord%y
endif

! convert solution to nodal basis coefficients, and then to h-hierarchical
! basis coefficients

call phier2nodal(solution,xvert,yvert,degree)
call nodal2hhier(solution,xvert,yvert,degree)

! compute the l2 norm of the red node h-hierarchic coefficients; for systems
! and multiple eigenvectors use the max over all solutions
! See subroutine phier2nodal for the order of the nodes

hcoarsen_indicator = 0

do j=1,grid%nsoln
   temp = 0
! along edge between children(1) and children(2), including central node
   isub = 5
   do i=isub,isub+2*((degree-1)/2),2
      temp = temp + solution(i,j)**2
   end do
   if (degree > 1) then
! along edge between children(1) and children(3)
      isub = degree+5
      do i=isub,isub+2*((degree-2)/2),2
         temp = temp + solution(i,j)**2
      end do
! along edge between children(2) and children(4)
      isub = 3+3*degree + element_dof(degree)
      do i=isub,isub+2*((degree-2)/2),2
         temp = temp + solution(i,j)**2
      end do
! along edge between children(3) and children(4)
      if (children(3) /= NO_CHILD) then
         isub = 1 + 5*degree + 2*element_dof(degree)
         do i=isub,isub+2*((degree-2)/2),2
            temp = temp + solution(i,j)**2
         end do
      endif
      if (degree > 2) then
! interior of children(1)
         isub = 6 + 3*(degree-1)
         do k=1,(degree-1)/2
            do i=1,degree-2*k
               temp = temp + solution(isub,j)**2
               isub = isub + 1
            end do
            isub = isub + degree - 2*k - 1
         end do
! interior of children(2)
         isub = 1 + 5*degree + element_dof(degree)
         do k=1,(degree-1)/2
            do i=1,degree-2*k
               temp = temp + solution(isub,j)**2
               isub = isub + 1
            end do
            isub = isub + degree - 2*k - 1
         end do
         if (children(3) /= NO_CHILD) then
! interior of children(3)
            isub = -1 + 7*degree + 2*element_dof(degree)
            do k=1,(degree-1)/2
               do i=1,degree-2*k
                  temp = temp + solution(isub,j)**2
                  isub = isub + 1
               end do
               isub = isub + degree - 2*k - 1
            end do
! interior of children(4)
            isub = -2 + 8*degree + 3*element_dof(degree)
            do k=1,(degree-1)/2
               do i=1,degree-2*k
                  temp = temp + solution(isub,j)**2
                  isub = isub + 1
               end do
               isub = isub + degree - 2*k - 1
            end do
         endif ! children(3) /= NO_CHILD
      endif ! degree > 2
   endif ! degree > 1

   hcoarsen_indicator = max(hcoarsen_indicator,sqrt(temp))
end do

deallocate(solution)

end function hcoarsen_indicator

!        ---------------------
function hcoarsen_indicator_3D(grid,parent,children,refine_control)
!        ---------------------

!----------------------------------------------------
! This routine computes the h coarsening error indicator for the children of
! parent for 3D.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(in) :: grid
integer, intent(in) :: parent, children(:)
type(refine_options), intent(in) :: refine_control
real(my_real) :: hcoarsen_indicator_3D
!----------------------------------------------------
! Local variables:

integer :: refinement_edge, refinement_edge_child(2), vert1, vert2, vert3
!----------------------------------------------------
! Begin executable code

! For MATCH_TOL termination, use the maximum of the parent's children's error
! indicators, and maximum over all eigenvectors

if (refine_control%refterm == MATCH_TOL) then
   hcoarsen_indicator_3D = max(maxval(grid%element_errind(children(1),:)), &
                               maxval(grid%element_errind(children(2),:)))
   return
endif

! It would be better to include the higher order bases, but for simplicity
! just use the h-hierarchical coefficient of the linear basis, i.e.  the
! difference between the solution at the vertex at the middle of the refinement
! edge and the average of the solution at the endpoints of the refinement edge.

refinement_edge = grid%element(parent)%refinement_edge
refinement_edge_child = grid%edge(refinement_edge)%child

if (grid%edge(refinement_edge_child(1))%vertex(1) == &
    grid%edge(refinement_edge_child(2))%vertex(1)) then
   vert1 = grid%edge(refinement_edge_child(1))%vertex(1)
   vert2 = grid%edge(refinement_edge_child(1))%vertex(2)
   vert3 = grid%edge(refinement_edge_child(2))%vertex(2)
elseif (grid%edge(refinement_edge_child(1))%vertex(1) == &
        grid%edge(refinement_edge_child(2))%vertex(2)) then
   vert1 = grid%edge(refinement_edge_child(1))%vertex(1)
   vert2 = grid%edge(refinement_edge_child(1))%vertex(2)
   vert3 = grid%edge(refinement_edge_child(2))%vertex(1)
elseif (grid%edge(refinement_edge_child(1))%vertex(2) == &
        grid%edge(refinement_edge_child(2))%vertex(1)) then
   vert1 = grid%edge(refinement_edge_child(1))%vertex(2)
   vert2 = grid%edge(refinement_edge_child(1))%vertex(1)
   vert3 = grid%edge(refinement_edge_child(2))%vertex(2)
elseif (grid%edge(refinement_edge_child(1))%vertex(2) == &
        grid%edge(refinement_edge_child(2))%vertex(2)) then
   vert1 = grid%edge(refinement_edge_child(1))%vertex(2)
   vert2 = grid%edge(refinement_edge_child(1))%vertex(1)
   vert3 = grid%edge(refinement_edge_child(2))%vertex(1)
else
   ierr = PHAML_INTERNAL_ERROR
   call fatal("hcoarsen_indicator_3D: did not find common vertex in refinement edge children")
   stop
endif

! use the maximum over all components/eigenvectors

hcoarsen_indicator_3D = maxval(abs(get_vert_soln(grid,vert1,colon,colon) - &
                                  (get_vert_soln(grid,vert2,colon,colon) + &
                                   get_vert_soln(grid,vert3,colon,colon))/2))

end function hcoarsen_indicator_3D

!========================================================================

!          ---------
subroutine reconcile(grid,procs,refine_control,solver_control, &
                         still_sequential,desired_level,desired_degree, &
                         was_hcoarsened)
!          ---------

!----------------------------------------------------
! This routine reconciles the refinements among the partitions to make a global
! compatible grid.  Basically it extends compatibility chains across partition
! boundaries and sets the degree for elements in the halo.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(proc_info), intent(in) :: procs
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
logical, intent(in) :: still_sequential
integer, pointer, optional :: desired_level(:), desired_degree(:,:)
logical(small_logical), pointer, optional :: was_hcoarsened(:)
!----------------------------------------------------
! Local variables:

integer :: i, j, k, myproc, nproc, nneigh, any_changes, nhalo, to, isub, rsub, &
           proc, degree, vert, edge, elem, err, astat, nbasis, ncomp, neigen, &
           nsoln, olddeg, iend
integer, pointer :: neighproc(:), allneigh(:), irecv(:)
real(my_real), pointer :: rrecv(:)
integer, allocatable :: inv_neighproc(:), insend(:), inrecv(:), iind(:), &
                        isend(:), rnsend(:), rnrecv(:), rind(:)
real(my_real), allocatable :: rsend(:)
integer :: halo(grid%biggest_elem)
type(hash_key) :: gid
integer :: edges(EDGES_PER_ELEMENT)
logical :: edge_iown(EDGES_PER_ELEMENT)
!----------------------------------------------------
! Begin executable code

! Master does not participate, and nothing to do if not parallel

if (my_proc(procs) == MASTER .or. num_proc(procs) == 1 .or. still_sequential) &
   return

! Start timing the reconciliation process

call reset_watch((/precon,cprecon/))
call start_watch((/precon,trecon/))

! Convenience variables

myproc = my_proc(procs)
nproc  = num_proc(procs)
ncomp = size_vert_soln(grid,dim=COMP_DIM)
neigen = size_vert_soln(grid,dim=EIGEN_DIM)
nsoln = ncomp*neigen

! Delete elements not needed

call delete_unowned_elements(grid,refine_control,solver_control)

! Get the neighbor processors

call get_neighbor_procs(procs,neighproc)
if (associated(neighproc)) then
   nneigh = size(neighproc)
else
   nneigh = 0
endif

allocate(insend(nneigh),inrecv(nneigh),iind(nneigh), &
         rnsend(nneigh),rnrecv(nneigh),rind(nneigh), &
         inv_neighproc(nproc),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("reconcile: allocation failed",procs=procs)
   stop
endif

inv_neighproc = -1
do i=1,nneigh
   inv_neighproc(neighproc(i)) = i
end do

! Repeat this process until no processor makes any changes

do
   any_changes = 0

! Get the elements in my inner halo

   call get_halo(grid,halo,nhalo,inner=.true.)

! Determine the size of the message segments and allocate the message

   insend = 0
   rnsend = 0
   do i=1,nhalo
      edges = grid%element(halo(i))%edge
      do k=1,EDGES_PER_ELEMENT
         edge_iown(k) = grid%element(grid%edge(edges(k))%assoc_elem)%iown
      end do
      call get_all_neighbors(grid,halo(i),allneigh)
      do j=1,size(allneigh)
         if (allneigh(j)==BOUNDARY) cycle
         if (grid%element(allneigh(j))%owner /= myproc .and. &
             grid%element(allneigh(j))%owner /= UNKNOWN_OWNER) then
            to = inv_neighproc(grid%element(allneigh(j))%owner)
            insend(to) = insend(to) + KEY_SIZE+1
            do k=1,VERTICES_PER_ELEMENT
               vert = grid%element(halo(i))%vertex(k)
               if (grid%element(grid%vertex(vert)%assoc_elem)%iown) then
                  rnsend(to) = rnsend(to) + nsoln
               else
                  rnsend(to) = rnsend(to) + 1
               endif
            end do
            do k=1,EDGES_PER_ELEMENT
               if (edge_iown(k)) then
                  insend(to) = insend(to) + 1
                  if (grid%edge(edges(k))%degree > 1) then
                     nbasis = size_edge_soln(grid,edges(k),dim=BASIS_DIM)
                     insend(to) = insend(to) + 1
                     rnsend(to) = rnsend(to) + nbasis*nsoln
                  else
                     insend(to) = insend(to) + 1
                  endif
               else
                  insend(to) = insend(to) + 2
               endif
            end do
            if (grid%element(halo(i))%degree > 2) then
               nbasis = size_elem_soln(grid,halo(i),dim=BASIS_DIM)
               insend(to) = insend(to) + 1
               rnsend(to) = rnsend(to) + nbasis*nsoln
            else
               insend(to) = insend(to) + 1
            endif
         endif
      end do
      deallocate(allneigh)
   end do
   allocate(isend(sum(insend)),rsend(sum(rnsend)),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("reconcile: allocation failed",procs=procs)
      stop
   endif

! Determine all the elements in my inner halo along with the neighbor
! processors across the partition boundary.  Pack the element and its degree
! into the message segment for that neighbor.
! Also pack the solution (and edge degree) of any owned vertices and edges
! of the element, and the solution for interior bases.  For unowned vertices,
! send huge for the solution; for unowned edges send -1 for the degree

   iind(1) = 1
   do i=2,nneigh
      iind(i) = iind(i-1) + insend(i-1)
   end do

   rind(1) = 1
   do i=2,nneigh
      rind(i) = rind(i-1) + rnsend(i-1)
   end do

! For each neighbor of each halo element ...

   do i=1,nhalo
      edges = grid%element(halo(i))%edge
      do k=1,EDGES_PER_ELEMENT
         edge_iown(k) = grid%element(grid%edge(edges(k))%assoc_elem)%iown
      end do
      call get_all_neighbors(grid,halo(i),allneigh)
      do j=1,size(allneigh)
         if (allneigh(j)==BOUNDARY) cycle

! Pack the halo element's gid and degree

         if (grid%element(allneigh(j))%owner /= myproc .and. &
             grid%element(allneigh(j))%owner /= UNKNOWN_OWNER) then
            to = inv_neighproc(grid%element(allneigh(j))%owner)
            call hash_pack_key(grid%element(halo(i))%gid,isend,iind(to))
            isend(iind(to)+KEY_SIZE) = grid%element(halo(i))%degree
            iind(to) = iind(to) + KEY_SIZE+1

! Pack the vertex solution, where owned

            do k=1,VERTICES_PER_ELEMENT
               vert = grid%element(halo(i))%vertex(k)
               if (grid%element(grid%vertex(vert)%assoc_elem)%iown) then
                  rsend(rind(to):rind(to)+nsoln-1) = &
                    reshape(get_vert_soln(grid,vert,colon,colon),(/nsoln/))
                  rind(to) = rind(to) + nsoln
               else
                  rsend(rind(to)) = huge(0.0_my_real)
                  rind(to) = rind(to) + 1
               endif
            end do

! Pack the edge solution, where owned

            do k=1,EDGES_PER_ELEMENT
               if (edge_iown(k)) then
                  degree = grid%edge(edges(k))%degree
                  isend(iind(to)) = degree
                  iind(to) = iind(to) + 1
                  if (degree > 1) then
                     nbasis = size_edge_soln(grid,edges(k),dim=BASIS_DIM)
                     isend(iind(to)) = nbasis
                     iind(to) = iind(to) + 1
                     rsend(rind(to):rind(to)+nbasis*nsoln-1) = &
                       reshape(get_edge_soln(grid,edges(k),colon,colon,colon), &
                               (/nbasis*nsoln/))
                     rind(to) = rind(to) + nbasis*nsoln
                  else
                     isend(iind(to)) = 0
                     iind(to) = iind(to) + 1
                  endif
               else
                  isend(iind(to)) = -1
                  isend(iind(to)+1) = -1
                  iind(to) = iind(to) + 2
               endif
            end do

! Pack the interior solution

            if (grid%element(halo(i))%degree > 2) then
               nbasis = size_elem_soln(grid,halo(i),dim=BASIS_DIM)
               isend(iind(to)) = nbasis
               iind(to) = iind(to) + 1
               rsend(rind(to):rind(to)+nbasis*nsoln-1) = &
                 reshape(get_elem_soln(grid,halo(i),colon,colon,colon), &
                         (/nbasis*nsoln/))
               rind(to) = rind(to) + nbasis*nsoln
            else
               isend(iind(to)) = 0
               iind(to) = iind(to) + 1
            endif
         endif
      end do
      deallocate(allneigh)
   end do

! Exchange messages

   call start_watch((/cprecon,ctrecon/))
   call phaml_neighbor_alltoall(procs,isend,insend,irecv,inrecv,3351)
   call phaml_neighbor_alltoall(procs,rsend,rnsend,rrecv,rnrecv,3352)
   call stop_watch((/cprecon,ctrecon/))

! For each received element

   isub = 1
   rsub = 1
   iend = 0
   do i=1,nneigh
      proc = neighproc(i)
      iend = iend + inrecv(i)
      if (proc == myproc) cycle
      do while (isub < iend)
         gid = hash_unpack_key(irecv,isub)
         degree = irecv(isub+KEY_SIZE)
         isub = isub + KEY_SIZE+1

! Create it if I don't have it

         elem = hash_decode_key(gid,grid%elem_hash)
         if (elem == HASH_NOT_FOUND) then
            call create_element(grid,gid,refine_control,solver_control,err, &
                                desired_level,desired_degree,was_hcoarsened, &
                                disown=.true.)
            if (err /= 0) then
               ierr = PHAML_INTERNAL_ERROR
               call fatal("reconcile: create_element returned error code", &
                          intlist=(/err/),procs=procs)
               stop
            endif
            any_changes = 1
            elem = hash_decode_key(gid,grid%elem_hash)
         endif

! Set the owner to be the sender

         if (grid%element(elem)%owner /= proc) then
            grid%element(elem)%owner = proc
            any_changes = 1
         endif

! Set its degree

         do while (grid%element(elem)%degree > degree)
            call p_coarsen_elem(grid,elem,err,refine_control,solver_control)
            if (err /= 0) then
               ierr = PHAML_INTERNAL_ERROR
               call fatal("reconcile: p_coarsen_elem returned error code", &
                          intlist=(/err/),procs=procs)
               stop
            endif
            any_changes = 1
         end do
         do while (grid%element(elem)%degree < degree)
            call p_refine_elem(grid,elem,refine_control,errcode=err)
            if (err /= 0) then
               ierr = PHAML_INTERNAL_ERROR
               call fatal("reconcile: p_refine_elem returned error code", &
                          intlist=(/err/),procs=procs)
               stop
            endif
            any_changes = 1
         end do

! Set the vertex solution for vertices owned by the sender

         do k=1,VERTICES_PER_ELEMENT
            if (rrecv(rsub) == huge(0.0_my_real)) then
               rsub = rsub + 1
            else
               call set_vert_soln(grid,grid%element(elem)%vertex(k),colon, &
                                  colon, &
                             reshape(rrecv(rsub:rsub+nsoln-1),(/ncomp,neigen/)))
               rsub = rsub + nsoln
            endif
         end do

! Set the edge degree and solution for edges owned by the sender

         do k=1,EDGES_PER_ELEMENT
            if (irecv(isub) == -1) then
               isub = isub + 2
            else
               edge = grid%element(elem)%edge(k)
               olddeg = grid%edge(edge)%degree
               degree = irecv(isub)
               grid%edge(edge)%degree = degree
               grid%dof = grid%dof + grid%system_size*(degree - olddeg)
               if (grid%element(grid%edge(edge)%assoc_elem)%iown) then
                  grid%dof_own = grid%dof_own + grid%system_size*(degree-olddeg)
               endif
               nbasis = irecv(isub+1)
               if (nbasis /= size_edge_soln(grid,edge,dim=BASIS_DIM)) then
                  call reallocate_edge_soln(grid,edge,size_basis=nbasis)
                  if (grid%have_true) then
                     call reallocate_edge_exact(grid,edge,size_basis=nbasis)
                  endif
                  if (grid%oldsoln_exists) then
                     if (associated_edge_oldsoln(grid,edge)) then
                       call reallocate_edge_oldsoln(grid,edge,size_basis=nbasis)
                     endif
                  endif
               endif
               if (nbasis > 0) then
                  call set_edge_soln(grid,edge,colon,colon,colon, &
                                     reshape(rrecv(rsub:rsub+nbasis*nsoln-1), &
                                     (/nbasis,ncomp,neigen/)))
               endif
               isub = isub + 2
               rsub = rsub + nbasis*nsoln
            endif
         end do

! Set the interior solution

         nbasis = irecv(isub)
         isub = isub + 1
         if (nbasis > 0) then
            if (nbasis /= size_elem_soln(grid,elem,dim=BASIS_DIM)) then
               call reallocate_elem_soln(grid,elem,size_basis=nbasis)
               if (grid%have_true) then
                  call reallocate_elem_exact(grid,elem,size_basis=nbasis)
               endif
               if (grid%oldsoln_exists) then
                  if (associated_elem_oldsoln(grid,elem)) then
                    call reallocate_elem_oldsoln(grid,elem,size_basis=nbasis)
                  endif
               endif
            endif
            call set_elem_soln(grid,elem,colon,colon,colon, &
                               reshape(rrecv(rsub:rsub+nbasis*nsoln-1), &
                               (/nbasis,ncomp,neigen/)))
            rsub = rsub + nbasis*nsoln
         endif

      end do! next element
   end do   ! next neighbor processor

   deallocate(isend,irecv,rsend,rrecv)

! Determine if any processor made a change and exit if not

   any_changes = phaml_global_max(procs,any_changes,3353)
   if (any_changes > 0) then
      grid_changed = .true.
   else
      exit
   endif

end do ! next outer loop

if (associated(neighproc)) deallocate(neighproc)
deallocate(insend,inrecv,iind,rnsend,rnrecv,rind,inv_neighproc)

grid%errind_up2date = .false.

! Stop timing the reconciliation process

call stop_watch((/precon,trecon/))

end subroutine reconcile

end module refine_adapt_mod
