!---------------------------------------------------------------------!
!                                PHAML                                !
!                                                                     !
! The Parallel Hierarchical Adaptive MultiLevel code for solving      !
! linear elliptic partial differential equations of the form          !
! (PUx)x + (QUy)y + RU = F on 2D polygonal domains with mixed         !
! boundary conditions, and eigenvalue problems where F is lambda*U.   !
!                                                                     !
! PHAML is public domain software.  It was produced as part of work   !
! done by the U.S. Government, and is not subject to copyright in     !
! the United States.                                                  !
!                                                                     !
!     William F. Mitchell                                             !
!     Applied and Computational Mathematics Division                  !
!     National Institute of Standards and Technology                  !
!     william.mitchell@nist.gov                                       !
!     http://math.nist.gov/phaml                                      !
!                                                                     !
!---------------------------------------------------------------------!

module refine_elements

!----------------------------------------------------
! This module contains routines with the details of element refinement.
!----------------------------------------------------

!----------------------------------------------------
! Other modules used:

use global
use gridtype_mod
use linsystype_mod
use boundary_util
use hash_mod
use grid_util
use error_estimators
use make_linsys
use linked_list
use message_passing, only: fatal
use omp_lib
!----------------------------------------------------

implicit none
private
public p_coarsen_elem, &            ! not thread safe
       p_coarsen_elem_interior, &   ! conditionally thread safe
       h_coarsen_element, &         ! not thread safe
       p_refine_elem, &             ! not thread safe
       p_refine_element_interior, & ! conditionally thread safe
       enforce_edge_rule, &         ! conditionally thread safe
       enforce_face_rule, &         ! conditionally thread safe
       before_h_refine, &           ! thread safe
       h_refine_element, &          ! not thread safe
       after_h_refine, &            ! thread safe
       h_refine_refinement_edge, &  ! thread safe
       create_element, &            ! not thread safe
       remove_hanging_nodes         ! not thread safe

!----------------------------------------------------
! Non-module procedures used are:

!----------------------------------------------------
! The following defined constants are defined:

!----------------------------------------------------
! The following variables are defined:

!----------------------------------------------------

contains

!          --------------
subroutine p_coarsen_elem(grid,elem,errcode,refine_control,solver_control)
!          --------------

!----------------------------------------------------
! This routine reduces the degree of element elem by 1.
! errcode is 0 if successful
!            1 if an error occurred
!           -1 if cannot coarsen
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem
integer, intent(out) :: errcode
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
!----------------------------------------------------
! Local variables:

integer :: i, deg, diff
!----------------------------------------------------
! Begin executable code

errcode = 0

! if the degree is 1, cannot coarsen

if (grid%element(elem)%degree <= 1) then
   errcode = -1
   return
endif

! Reduce element degree.

deg = grid%element(elem)%degree - 1
grid%element(elem)%degree = deg
if (deg >= 3) then
   diff = element_dof(deg+1) - element_dof(deg)
   grid%dof = grid%dof - grid%system_size*diff
   if (grid%element(elem)%iown) grid%dof_own = grid%dof_own - &
                                               grid%system_size*diff
endif

! Enforce the edge rule on each face and edge.

do i=1,FACES_PER_ELEMENT
   call enforce_face_rule(grid,refine_control,grid%element(elem)%face(i))
end do

do i=1,EDGES_PER_ELEMENT
   call enforce_edge_rule(grid,refine_control,grid%element(elem)%edge(i))
end do

! fix exact for the element bases; other solution coefficients stay the same
! because the basis is p-hierarchical

if (grid%have_true) then
   do i=1,grid%system_size
      call elem_exact(grid,elem,i,"t")
   end do
endif

! compute error indicator

call error_indicator(grid,elem,refine_control%error_estimator, &
                     solver_control,grid%element_errind(elem,:), &
                     grid%element(elem)%work)

grid%grid_modified = .true.

end subroutine p_coarsen_elem

!          -----------------------
subroutine p_coarsen_elem_interior(grid,elem,errcode,refine_control, &
                                   solver_control,delta_dof,delta_dof_own)
!          -----------------------

!----------------------------------------------------
! This routine reduces the degree of element elem by 1, but does not change
! the edge degrees.
! errcode is 0 if successful
!            1 if an error occurred
!           -1 if cannot coarsen
!
! Thread safe if delta_dof and delta_dof_own are present, and the error
! estimator is not equilibrated residual
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem
integer, intent(out) :: errcode
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
integer, intent(out), optional :: delta_dof, delta_dof_own
!----------------------------------------------------
! Local variables:

integer :: i, deg, diff
!----------------------------------------------------
! Begin executable code

if (present(delta_dof) .neqv. present(delta_dof_own)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("p_coarsen_elem_interior: delta_* must all be present or absent")
   stop
endif

! initialize delta dof's in case we return early

if (present(delta_dof)) then
   delta_dof = 0
   delta_dof_own = 0
endif

errcode = 0

! if the degree is 1, cannot coarsen

if (grid%element(elem)%degree <= 1) then
   errcode = -1
   return
endif

! Reduce element degree.

deg = grid%element(elem)%degree - 1
grid%element(elem)%degree = deg
if (deg >= 3) then
   diff = element_dof(deg+1) - element_dof(deg)
   if (present(delta_dof)) then
      delta_dof = -grid%system_size*diff
      if (grid%element(elem)%iown) delta_dof_own = -grid%system_size*diff
   else
      grid%dof = grid%dof - grid%system_size*diff
      if (grid%element(elem)%iown) grid%dof_own = grid%dof_own - &
                                                  grid%system_size*diff
   endif
endif

! fix exact for the element bases; other solution coefficients stay the same
! because the basis is p-hierarchical

if (grid%have_true) then
   do i=1,grid%system_size
      call elem_exact(grid,elem,i,"t")
   end do
endif

! compute error indicator

call error_indicator(grid,elem,refine_control%error_estimator, &
                     solver_control,grid%element_errind(elem,:), &
                     grid%element(elem)%work)

grid%grid_modified = .true.

end subroutine p_coarsen_elem_interior

!          ----------------------
subroutine h_coarsen_element(grid, parent, errcode, refcont, solver_control, &
                             was_hcoarsened, desired_level, delta_nelem, &
                             delta_nelem_leaf, delta_nelem_leaf_own, &
                             delta_nedge, delta_nedge_leaf, &
                             delta_nedge_leaf_own, delta_nvert, &
                             delta_nvert_own, delta_dof, delta_dof_own)
!          ----------------------

!----------------------------------------------------
! This routine removes the bisection refinement of element parent and
! all other elements that share the refinement edge of parent.
! errcode is 0 if successful
!            1 if an error occurred
!           -1 if cannot unrefine because of grandchildren
!           -2 if nothing to derefine (no children)
!           -3 if cannot unrefine because children have different owners
!           -4 if a child of an element that shares the refinement edge is
!                 marked for refinement
!
! Not thread safe
!
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
integer, intent(in) :: parent
integer, intent(out) :: errcode
type(refine_options), intent(in) :: refcont
type(solver_options), intent(in) :: solver_control
logical(small_logical), intent(inout), optional :: was_hcoarsened(:)
integer, intent(inout), optional :: desired_level(:)
integer, intent(out), optional :: delta_nelem, delta_nelem_leaf, &
                                  delta_nelem_leaf_own, delta_nedge, &
                                  delta_nedge_leaf, delta_nedge_leaf_own, &
                                  delta_nvert, delta_nvert_own, delta_dof, &
                                  delta_dof_own

!----------------------------------------------------
! Local variables:

integer, parameter :: MAX_SEEN = 100
integer :: seen(MAX_SEEN)
integer :: refinement_edge(4), refinement_edge_child(2,4), rm_vertices(4), &
           nparent, n_rm_child, astat, i, n_rm_vertex, n_rm_edge, j, edge, &
           n_rm_face, face, elem, c(2), nseen, num_periodic, vert, k, comp
integer, pointer :: parents(:)
integer, allocatable :: rm_children(:), parent_edges(:), &
                        rm_edges(:), parent_faces(:), rm_faces(:), &
                        verts(:), edges(:)
!----------------------------------------------------
! Begin executable code

errcode = 0

! if this is a leaf, nothing to derefine

if (grid%element(parent)%isleaf) then
   errcode = -2
   return
endif

! Identify the refinement edge of the parent, and the children of the
! refinement edge.  If the refinement edge is periodic, also identify the
! periodic mates.

refinement_edge(1) = grid%element(parent)%refinement_edge
refinement_edge_child(:,1) = grid%edge(refinement_edge(1))%child

if (is_periodic_edge(refinement_edge(1),grid)) then
   num_periodic = num_matching_periodic_edge(grid,refinement_edge(1))
   if (num_periodic == 1) then
      refinement_edge(2) = matching_periodic_edge(grid,refinement_edge(1))
      refinement_edge_child(1,2) = &
         matching_periodic_edge(grid,refinement_edge_child(1,1))
      refinement_edge_child(2,2) = &
         matching_periodic_edge(grid,refinement_edge_child(2,1))
   else
      allocate(edges(num_periodic),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in h_coarsen_element")
         stop
      endif
      i = matching_periodic_edge(grid,refinement_edge(1),edges)
      refinement_edge(2:1+num_periodic) = edges
      i = matching_periodic_edge(grid,refinement_edge_child(1,1),edges)
      refinement_edge_child(1,2:1+num_periodic) = edges
      i = matching_periodic_edge(grid,refinement_edge_child(2,1),edges)
      refinement_edge_child(2,2:1+num_periodic) = edges
      deallocate(edges)
   endif
   num_periodic = num_periodic + 1
else
   num_periodic = 1
endif

! Get all the elements that share the refinement edge

call get_edge_elements(grid,refinement_edge(1),parents,include_periodic=.true.)
nparent = size(parents)

! Get the children of each parent.  If any child does not exist or is not a
! leaf, or is not marked for coarsening, cannot derefine.

n_rm_child = 2*nparent
allocate(rm_children(n_rm_child),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in h_coarsen_element")
   stop
endif

do i=1,nparent
   rm_children(1+2*(i-1):2+2*(i-1)) = &
      get_child_lid(grid%element(parents(i))%gid,(/1,2/),grid%elem_hash)
   do j=1,2
      if (rm_children(j+2*(i-1)) == NO_CHILD) then
         errcode = -2
         deallocate(parents,rm_children)
         return
      endif
      if (.not. grid%element(rm_children(j+2*(i-1)))%isleaf) then
         errcode = -1
         deallocate(parents,rm_children)
         return
      endif
      if (present(desired_level)) then
         if (desired_level(rm_children(j+2*(i-1))) >= &
             grid%element(rm_children(j+2*(i-1)))%level) then
            errcode = -4
            deallocate(parents,rm_children)
            return
         endif
      endif
   end do
end do

! The vertex to be removed is the common vertex of the child refinement edges.

n_rm_vertex = num_periodic
do i=1,num_periodic
   if (grid%edge(refinement_edge_child(1,i))%vertex(1) == &
       grid%edge(refinement_edge_child(2,i))%vertex(1)) then
      rm_vertices(i) = grid%edge(refinement_edge_child(1,i))%vertex(1)
   elseif (grid%edge(refinement_edge_child(1,i))%vertex(1) == &
           grid%edge(refinement_edge_child(2,i))%vertex(2)) then
      rm_vertices(i) = grid%edge(refinement_edge_child(1,i))%vertex(1)
   elseif (grid%edge(refinement_edge_child(1,i))%vertex(2) == &
           grid%edge(refinement_edge_child(2,i))%vertex(1)) then
      rm_vertices(i) = grid%edge(refinement_edge_child(1,i))%vertex(2)
   elseif (grid%edge(refinement_edge_child(1,i))%vertex(2) == &
           grid%edge(refinement_edge_child(2,i))%vertex(2)) then
      rm_vertices(i) = grid%edge(refinement_edge_child(1,i))%vertex(2)
   else
      ierr = PHAML_INTERNAL_ERROR
      call fatal("h_coarsen_element: did not find common vertex between refinement edges")
      stop
   endif
end do

! The edges to be removed are the edges of the children that are not edges
! of the parents.  This is the two refinement edge children and the edges
! that bisect faces.  Make a list of the parent edges, then go through the
! edges of the children to find those that are not on a parent and not already
! on the remove list.

allocate(parent_edges(EDGES_PER_ELEMENT*nparent),rm_edges(12+nparent),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in h_coarsen_element")
   stop
endif

do i=1,nparent
   parent_edges(1+(i-1)*EDGES_PER_ELEMENT:i*EDGES_PER_ELEMENT) = &
      grid%element(parents(i))%edge
end do

do i=1,num_periodic
   rm_edges(1+2*(i-1):2+2*(i-1)) = refinement_edge_child(:,i)
end do
n_rm_edge = 2*num_periodic

do i=1,n_rm_child
   do j=1,EDGES_PER_ELEMENT
      edge = grid%element(rm_children(i))%edge(j)
      if (any(parent_edges == edge)) cycle
      if (any(rm_edges(1:n_rm_edge) == edge)) cycle
      n_rm_edge = n_rm_edge + 1
      if (n_rm_edge > size(rm_edges)) then
         ierr = PHAML_INTERNAL_ERROR
         call fatal("h_coarsen_element: rm_edges is not big enough")
         stop
      endif
      rm_edges(n_rm_edge) = edge
   end do
end do

deallocate(parent_edges)

! Likewise for the faces to be removed

allocate(parent_faces(4*nparent),rm_faces(n_rm_child+nparent+8),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("memory allocation failed in h_coarsen_element")
   stop
endif

do i=1,nparent
   parent_faces(1+(i-1)*FACES_PER_ELEMENT:i*FACES_PER_ELEMENT) = &
      grid%element(parents(i))%face
end do

n_rm_face = 0
do i=1,n_rm_child
   do j=1,FACES_PER_ELEMENT
      face = grid%element(rm_children(i))%face(j)
      if (any(parent_faces == face)) cycle
      if (any(rm_faces(1:n_rm_face) == face)) cycle
      n_rm_face = n_rm_face + 1
      if (n_rm_face > size(rm_faces)) then
         ierr = PHAML_INTERNAL_ERROR
         call fatal("h_coarsen_element: rm_faces is not big enough")
         stop
      endif
      rm_faces(n_rm_face) = face
   end do
end do

deallocate(parent_faces)

! Reassign associated elements

do i=1,nparent
   elem = parents(i)
   c = rm_children(1+2*(i-1):2+2*(i-1))
   do j=1,VERTICES_PER_ELEMENT
      if (grid%vertex(grid%element(elem)%vertex(j))%assoc_elem == c(1) .or. &
          grid%vertex(grid%element(elem)%vertex(j))%assoc_elem == c(2)) then
         grid%vertex(grid%element(elem)%vertex(j))%assoc_elem = elem
         if (num_periodic > 1) then
            if (is_periodic_vert(grid%element(elem)%vertex(j),grid)) then
               allocate(verts(num_matching_periodic_vert(grid, &
                        grid%element(elem)%vertex(j))),stat=astat)
               if (astat /= 0) then
                  ierr = ALLOC_FAILED
                  call fatal("memory allocation failed in h_coarsen_element")
                  stop
               endif
               vert = matching_periodic_vert(grid, &
                         grid%element(elem)%vertex(j),verts)
               do k=1,size(verts)
                  if (grid%vertex(verts(k))%assoc_elem == c(1) .or. &
                      grid%vertex(verts(k))%assoc_elem == c(2)) then
                     grid%vertex(verts(k))%assoc_elem = elem
                  endif
               end do
               deallocate(verts)
            endif
         endif
      endif
   end do
   do j=1,EDGES_PER_ELEMENT
      if (grid%edge(grid%element(elem)%edge(j))%assoc_elem == c(1) .or. &
          grid%edge(grid%element(elem)%edge(j))%assoc_elem == c(2)) then
         grid%edge(grid%element(elem)%edge(j))%assoc_elem = elem
         if (num_periodic > 1) then
            if (is_periodic_edge(grid%element(elem)%edge(j),grid)) then
               allocate(edges(num_matching_periodic_edge(grid, &
                        grid%element(elem)%edge(j))),stat=astat)
               if (astat /= 0) then
                  ierr = ALLOC_FAILED
                  call fatal("memory allocation failed in h_coarsen_element")
                  stop
               endif
               edge = matching_periodic_edge(grid, &
                         grid%element(elem)%edge(j),edges)
               do k=1,size(edges)
                  if (grid%edge(edges(k))%assoc_elem == c(1) .or. &
                      grid%edge(edges(k))%assoc_elem == c(2)) then
                     grid%edge(edges(k))%assoc_elem = elem
                  endif
               end do
               deallocate(edges)
            endif
         endif
      endif
   end do
   do j=1,FACES_PER_ELEMENT
      if (grid%face(grid%element(elem)%face(j))%assoc_elem == c(1) .or. &
          grid%face(grid%element(elem)%face(j))%assoc_elem == c(2)) then
         grid%face(grid%element(elem)%face(j))%assoc_elem = elem
         if (num_periodic > 1) then
            if (is_periodic_face(grid%element(elem)%face(j),grid)) then
               face = matching_periodic_face(grid,grid%element(elem)%face(j))
               if (grid%face(face)%assoc_elem == c(1) .or. &
                   grid%face(face)%assoc_elem == c(2)) then
                  grid%face(face)%assoc_elem = elem
               endif
            endif
         endif
      endif
   end do
end do

! Set the parent degree to the maximum of the two children, and enforce the
! edge rule on each face and edge.

do i=1,nparent
   grid%element(parents(i))%degree = &
      max(grid%element(rm_children(1+2*(i-1)))%degree, &
          grid%element(rm_children(2+2*(i-1)))%degree)

   do j=1,FACES_PER_ELEMENT
      call enforce_face_rule(grid,refcont,grid%element(parents(i))%face(j))
   end do

   do j=1,EDGES_PER_ELEMENT
      call enforce_edge_rule(grid,refcont,grid%element(parents(i))%edge(j))
   end do
end do

! The parents get the tags from the children, in case they have changed since
! the parent was refined.  In particular, the charge tag is the sum of the
! children charge tags.
! But for multiple point charges, merge the linked lists of charges of the
! children.

do i=1,nparent
   grid%element(parents(i))%tags = grid%element(rm_children(1+2*(i-1)))%tags
   if (sem_simulation) then
      if (.not. sem_point_charges) then
         grid%element(parents(i))%tags(chargetag) = &
               grid%element(rm_children(1+2*(i-1)))%tags(chargetag) + &
               grid%element(rm_children(2+2*(i-1)))%tags(chargetag)
      else
         grid%element(parents(i))%tags(chargetag) = &
            grid%element(rm_children(1+2*(i-1)))%tags(chargetag)
         if (grid%element(parents(i))%tags(chargetag) == 0) then
            grid%element(parents(i))%tags(chargetag) = &
               grid%element(rm_children(2+2*(i-1)))%tags(chargetag)
         elseif (grid%element(rm_children(2+2*(i-1)))%tags(chargetag) /= 0) then
            j = grid%element(parents(i))%tags(chargetag)
            do while (sem_next_charge(j) /= -1)
               j = sem_next_charge(j)
            end do
            sem_next_charge(j) = grid%element(rm_children(2+2*(i-1)))%tags(chargetag)
         endif
      endif
   endif
end do

! The parents are now leaves and do not want to be refined or coarsened

do i=1,nparent
   grid%element(parents(i))%isleaf = .true.
   if (present(desired_level)) then
      desired_level(parents(i)) = grid%element(parents(i))%level
   endif
end do

! if the children are leaves in the old solution, the parents become old leaves

do i=1,nparent
   if (grid%element(rm_children(1+2*(i-1)))%oldleaf) then
      grid%element(parents(i))%oldleaf = .true.
   endif
end do

! the refinement edge no longer has children

grid%edge(refinement_edge(1:num_periodic))%child(1) = NO_CHILD
grid%edge(refinement_edge(1:num_periodic))%child(2) = NO_CHILD

! the faces of the parents no longer have children

do i=1,nparent
   do j=1,FACES_PER_ELEMENT
      grid%face(grid%element(parents(i))%face(j))%child = NO_CHILD
   end do
end do

! set initial solution

call hcoarsen_init_guess(grid,parents,refinement_edge(1:num_periodic))

! Remove the children elements, faces, edges and vertices

do i=1,n_rm_child
   elem = rm_children(i)
   if (associated_elem_soln(grid,elem)) then
      call deallocate_elem_soln(grid,elem)
   endif
   if (associated_elem_exact(grid,elem)) then
      call deallocate_elem_exact(grid,elem)
   endif
   if (associated_elem_oldsoln(grid,elem)) then
      call deallocate_elem_oldsoln(grid,elem)
   endif
   call linked_list_destroy(grid%element(elem)%alt_tags)
   if (grid%element(elem)%previous == END_OF_LIST) then
      grid%head_level_elem(grid%element(elem)%level) = grid%element(elem)%next
   else
      grid%element(grid%element(elem)%previous)%next = grid%element(elem)%next
   endif
   if (grid%element(elem)%next == END_OF_LIST) then
      ! nothing, unless I add end_level_elem to the grid data structure
   else
      grid%element(grid%element(elem)%next)%previous = grid%element(elem)%previous
   endif
   call put_next_free_elem(grid,elem)
end do

do i=1,n_rm_face
   face = rm_faces(i)
   if (associated_face_soln(grid,face)) then
      call deallocate_face_soln(grid,face)
   endif
   if (associated_face_exact(grid,face)) then
      call deallocate_face_exact(grid,face)
   endif
   if (associated_face_oldsoln(grid,face)) then
      call deallocate_face_oldsoln(grid,face)
   endif
   if (associated(grid%face(face)%alt_bmark)) then
      deallocate(grid%face(face)%alt_bmark)
   endif
   call put_next_free_face(grid,face)
end do

do i=1,n_rm_edge
   edge = rm_edges(i)
   if (associated_edge_soln(grid,edge)) then
      call deallocate_edge_soln(grid,edge)
   endif
   if (associated_edge_exact(grid,edge)) then
      call deallocate_edge_exact(grid,edge)
   endif
   if (associated_edge_oldsoln(grid,edge)) then
      call deallocate_edge_oldsoln(grid,edge)
   endif
   if (associated(grid%edge(edge)%alt_bmark)) then
      deallocate(grid%edge(edge)%alt_bmark)
   endif
   call put_next_free_edge(grid,edge)
end do

do i=1,n_rm_vertex
   vert = rm_vertices(i)
   if (associated(grid%vertex(vert)%alt_bmark)) then
      deallocate(grid%vertex(vert)%alt_bmark)
   endif
   call put_next_free_vert(grid,vert)
end do

! Remove the children elements, faces, edges and vertices from the hash tables

do i=1,n_rm_child
   call hash_remove(grid%element(rm_children(i))%gid,grid%elem_hash)
end do
do i=1,n_rm_face
   call hash_remove(grid%face(rm_faces(i))%gid,grid%face_hash)
end do
do i=1,n_rm_edge
   call hash_remove(grid%edge(rm_edges(i))%gid,grid%edge_hash)
end do
do i=1,n_rm_vertex
   call hash_remove(grid%vertex(rm_vertices(i))%gid,grid%vert_hash)
end do

! Update counts nelem, nvert, etc.

! change in number of elements, edges and vertices

grid%nelem = grid%nelem - n_rm_child
grid%nelem_leaf = grid%nelem_leaf - n_rm_child + nparent
grid%nface = grid%nface - n_rm_face
grid%nface_leaf = grid%nface_leaf - (n_rm_face-nparent)/2 - nparent
grid%nedge = grid%nedge - n_rm_edge
grid%nedge_leaf = grid%nedge_leaf - n_rm_edge + 1
grid%nvert = grid%nvert - n_rm_vertex
do i=1,n_rm_child
   if (grid%element(rm_children(i))%iown) then
      grid%nelem_leaf_own = grid%nelem_leaf_own - 1
   endif
end do
do i=1,nparent
   if (grid%element(parents(i))%iown) then
      grid%nelem_leaf_own = grid%nelem_leaf_own + 1
   endif
end do
! NOTE: if PHAML is extended to use message passing parallel in 3D, the change
!       in the number of owned leaf faces must be computed
grid%nface_leaf_own = grid%nface_leaf
do i=1,n_rm_edge
   if (grid%element(grid%edge(rm_edges(i))%assoc_elem)%iown) then
      grid%nedge_leaf_own = grid%nedge_leaf_own - 1
   endif
end do
if (grid%element(grid%edge(refinement_edge(1))%assoc_elem)%iown) then
   grid%nedge_leaf_own = grid%nedge_leaf_own + 1
endif
do i=1,n_rm_vertex
   if (grid%element(grid%vertex(rm_vertices(i))%assoc_elem)%iown) then
      grid%nvert_own = grid%nvert_own - 1
   endif
end do

! change in number of degrees of freedom

! ... from elements

do i=1,n_rm_child
   if (grid%element(rm_children(i))%degree >= 4) then
      grid%dof = grid%dof - grid%system_size * &
                 element_dof(grid%element(rm_children(i))%degree)
      if (grid%element(rm_children(i))%iown) then
         grid%dof_own = grid%dof_own - grid%system_size * &
                        element_dof(grid%element(rm_children(i))%degree)
      endif
   endif
end do
do i=1,nparent
   if (grid%element(parents(i))%degree >= 4) then
      grid%dof = grid%dof + grid%system_size * &
                 element_dof(grid%element(parents(i))%degree)
      if (grid%element(parents(i))%iown) then
         grid%dof_own = grid%dof_own + grid%system_size * &
                        element_dof(grid%element(parents(i))%degree)
      endif
   endif
end do

! ... from faces

do i=1,n_rm_face
   if (grid%face(rm_faces(i))%degree >= 3) then
      if (grid%any_periodic) then
         do comp=1,grid%system_size
            if (.not. is_periodic_face_slave(rm_faces(i),grid,comp)) then
               grid%dof = grid%dof - face_dof(grid%face(rm_faces(i))%degree)
               if (grid%element(grid%face(rm_faces(i))%assoc_elem)%iown) then
                  grid%dof_own = grid%dof_own - &
                                 face_dof(grid%face(rm_faces(i))%degree)
               endif
            endif
         end do
      else
         grid%dof = grid%dof - grid%system_size * &
                    face_dof(grid%face(rm_faces(i))%degree)
         if (grid%element(grid%face(rm_faces(i))%assoc_elem)%iown) then
            grid%dof_own = grid%dof_own - grid%system_size * &
                           face_dof(grid%face(rm_faces(i))%degree)
         endif
      endif
   endif
end do
nseen = 0
do i=1,nparent
   do j=1,FACES_PER_ELEMENT
      face = grid%element(parents(i))%face(j)
      do k=1,num_periodic
         if (any(grid%face(face)%edge == refinement_edge(k))) then
            if (any(seen(1:nseen) == face)) cycle
            nseen = nseen + 1
            if (nseen > MAX_SEEN) then
               ierr = PHAML_INTERNAL_ERROR
               call fatal("need to increase MAX_SEEN in h_coarsen_element")
               stop
            endif
            seen(nseen) = face
            if (grid%face(face)%degree >= 3) then
               if (grid%any_periodic) then
                  do comp=1,grid%system_size
                     if (.not. is_periodic_face_slave(face,grid,comp)) then
                        grid%dof = grid%dof + face_dof(grid%face(face)%degree)
                        if (grid%element(grid%face(face)%assoc_elem)%iown) then
                           grid%dof_own = grid%dof_own + &
                                          face_dof(grid%face(face)%degree)
                        endif
                     endif
                  end do
               else
                  grid%dof = grid%dof + grid%system_size * &
                             face_dof(grid%face(face)%degree)
                  if (grid%element(grid%face(face)%assoc_elem)%iown) then
                     grid%dof_own = grid%dof_own + grid%system_size *&
                                    face_dof(grid%face(face)%degree)
                  endif
               endif
            endif
         endif
      end do
   end do
end do

! ... from edges

do i=1,n_rm_edge
   if (grid%edge(rm_edges(i))%degree >= 2) then
      if (grid%any_periodic) then
         do comp=1,grid%system_size
            if (.not. is_periodic_edge_slave(rm_edges(i),grid,comp)) then
               grid%dof = grid%dof - (grid%edge(rm_edges(i))%degree - 1)
               if (grid%element(grid%edge(rm_edges(i))%assoc_elem)%iown) then
                  grid%dof_own = grid%dof_own - &
                                 (grid%edge(rm_edges(i))%degree - 1)
               endif
            endif
         end do
      else
         grid%dof = grid%dof - grid%system_size * &
                    (grid%edge(rm_edges(i))%degree - 1)
         if (grid%element(grid%edge(rm_edges(i))%assoc_elem)%iown) then
            grid%dof_own = grid%dof_own - grid%system_size * &
                           (grid%edge(rm_edges(i))%degree - 1)
         endif
      endif
   endif
end do
! NOTE: because I am only looking at the refinement edge here, not any
!       periodic mates of it, I don't need to check is_periodic_slave
if (grid%edge(refinement_edge(1))%degree >= 2) then
   grid%dof = grid%dof + grid%system_size * &
              (grid%edge(refinement_edge(1))%degree - 1)
   if (grid%element(grid%edge(refinement_edge(1))%assoc_elem)%iown) then
      grid%dof_own = grid%dof_own + grid%system_size * &
                     (grid%edge(refinement_edge(1))%degree - 1)
   endif
endif

! ... from vertices

! NOTE: because I am only looking at the removed vertex here, not any
!       periodic mates of it, I don't need to check is_periodic_slave
grid%dof = grid%dof - grid%system_size
if (grid%element(grid%vertex(rm_vertices(1))%assoc_elem)%iown) then
   grid%dof_own = grid%dof_own - grid%system_size
endif

! compute new error indicators, and remember that these elements were previously
! refined and then coarsened

do i=1,nparent
   call error_indicator(grid,parents(i),refcont%error_estimator, &
                        solver_control,grid%element_errind(parents(i),:), &
                        grid%element(parents(i))%work)
   if (present(was_hcoarsened)) then
      was_hcoarsened(parents(i)) = .true.
   endif
end do

deallocate(parents,rm_children,rm_edges,rm_faces)

grid%grid_modified = .true.

end subroutine h_coarsen_element

!          -------------
subroutine p_refine_elem(grid,elem,refine_control,errcode)
!          -------------

!----------------------------------------------------
! This routine performs p refinement of element elem
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem
type(refine_options), intent(in) :: refine_control
integer, optional, intent(out) :: errcode
!----------------------------------------------------
! Local variables:

integer :: i, err
!----------------------------------------------------
! Begin executable code

! if the error indicator list is provided, remove the element from it

! refine the interior of the element

call p_refine_element_interior(grid,refine_control,elem,err)
if (present(errcode)) errcode = err
if (err /= 0) return

! Enforce the edge rule on each face and edge.

do i=1,FACES_PER_ELEMENT
   call enforce_face_rule(grid,refine_control,grid%element(elem)%face(i))
end do

do i=1,EDGES_PER_ELEMENT
   call enforce_edge_rule(grid,refine_control,grid%element(elem)%edge(i))
end do

! Set initial condition for new solution components.

if (refine_control%error_estimator == INITIAL_CONDITION) then
   call init_guess_ic(grid,element=elem)
else
   call init_guess_p(grid,elem)
endif

! compute error indicators later

grid%element_errind(elem,:) = -1.0_my_real

grid%grid_modified = .true.

end subroutine p_refine_elem

!          -------------------------
subroutine p_refine_element_interior(grid,refine_control,elem,errcode, &
                                     delta_dof,delta_dof_own)
!          -------------------------

!----------------------------------------------------
! This routine performs p refinement of the interior of element elem,
! but not the faces and edges.
!
! Thread safe if delta_dof and delta_dof_own are present.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(refine_options), intent(in) :: refine_control
integer, intent(in) :: elem
integer, intent(out) :: errcode
integer, intent(out), optional :: delta_dof, delta_dof_own
!----------------------------------------------------
! Local variables:

integer :: i, new_deg, new_size, old_size
!----------------------------------------------------
! Begin executable code

if (present(delta_dof) .neqv. present(delta_dof_own)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("p_refine_element_interior: delta_* must all be present or absent")
   stop
endif

! initialize delta dof's in case we return early

if (present(delta_dof)) then
   delta_dof = 0
   delta_dof_own = 0
endif

! errcode = 0 success
!           1 did not refine

errcode = 1

! verify the element is a leaf

if (.not. grid%element(elem)%isleaf) return

! new degree for the element

new_deg = grid%element(elem)%degree + 1

! if the maximum degree would be exceeded, don't p-refine.

if (new_deg > refine_control%max_deg) then
   return
endif

errcode = 0

! make sure allocated memory is large enough.  If not, reallocate to
! degree+3 so we don't reallocate at every p refinement

if (associated_elem_soln(grid,elem)) then
   old_size = size_elem_soln(grid,elem,dim=BASIS_DIM)
else
   old_size = 0
endif
new_size = element_dof(new_deg)

if (old_size < new_size) then
   new_size = element_dof(new_deg+3)
   call reallocate_elem_soln(grid,elem,size_basis=new_size)
   if (grid%have_true) then
      call reallocate_elem_exact(grid,elem,size_basis=new_size)
   endif
   if (grid%oldsoln_exists) then
      call reallocate_elem_oldsoln(grid,elem,size_basis=new_size)
   endif
endif

! Increment the element degree by one.

grid%element(elem)%degree = new_deg

if (present(delta_dof)) then
   if (new_deg >= 4) then
      delta_dof = grid%system_size*(element_dof(new_deg)-element_dof(new_deg-1))
      if (grid%element(elem)%iown) then
         delta_dof_own = grid%system_size * &
                         (element_dof(new_deg)-element_dof(new_deg-1))
      else
         delta_dof_own = 0
      endif
   else
      delta_dof = 0
      delta_dof_own = 0
   endif
else
   if (new_deg >= 4) then
      grid%dof = grid%dof + grid%system_size * &
                         (element_dof(new_deg)-element_dof(new_deg-1))
      if (grid%element(elem)%iown) grid%dof_own = grid%dof_own + &
                  grid%system_size*(element_dof(new_deg)-element_dof(new_deg-1))
   endif
endif

! set the new solution coefficients to 0, and the exact solution

if (new_deg >= 4) then
   call set_elem_soln(grid,elem, &
                      (/1+element_dof(new_deg-1),element_dof(new_deg)/), &
                      colon,colon,0.0_my_real)
   if (grid%have_true) then
      do i=1,grid%system_size
         call elem_exact(grid,elem,i,"t")
      end do
   endif
endif

! if I don't own it, set flag for reconciliation

if (.not. grid%element(elem)%iown) then
   grid%element(elem)%prefined_unowned = .true.
endif

! predicted error for SMOOTH_PRED hp strategy

grid%element(elem)%sp_eta_pred = sqrt(refine_control%sp_gamma_p) * &
                        maxval(grid%element_errind(elem,1:max(1,grid%num_eval)))

! indicate the error indicator should be recomputed later

grid%element_errind(elem,:) = -1.0_my_real

grid%grid_modified = .true.

end subroutine p_refine_element_interior

!          -----------------
subroutine enforce_edge_rule(grid,refine_control,edge,delta_dof,delta_dof_own)
!          -----------------

!----------------------------------------------------
! This routine enforces either the minimum or maximum rule for edge
!
! Thread safe if:
!   delta_dof and delta_dof_own are present
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(refine_options), intent(in) :: refine_control
integer, intent(in) :: edge
integer, intent(out), optional :: delta_dof, delta_dof_own
!----------------------------------------------------
! Local variables:

integer :: i, old_deg, new_deg, old_size, new_size
integer, pointer :: elem(:)
!----------------------------------------------------
! Begin executable code

if (present(delta_dof) .neqv. present(delta_dof_own)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("enforce_edge_rule: delta_* must all be present or absent")
   stop
endif

! get the elements that share this edge

call get_edge_elements(grid,edge,elem)

! determine what the degree of this edge should be

if (refine_control%edge_rule == MINIMUM_RULE) then
   new_deg = huge(0)
else ! (refine_control%edge_rule == MAXIMUM_RULE)
   new_deg = 0
endif
do i=1,size(elem)
   if (refine_control%edge_rule == MINIMUM_RULE) then
      new_deg = min(grid%element(elem(i))%degree,new_deg)
   else ! (refine_control%edge_rule == MAXIMUM_RULE)
      new_deg = max(grid%element(elem(i))%degree,new_deg)
   endif
end do

! assign the new degree

old_deg = grid%edge(edge)%degree
if (new_deg /= old_deg) then
   grid%edge(edge)%degree = new_deg
   if (present(delta_dof)) then
      delta_dof = (new_deg-old_deg) * &
                  count(grid%edge_type(edge,:)/=PERIODIC_SLAVE)
      if (grid%element(grid%edge(edge)%assoc_elem)%iown) then
         delta_dof_own = (new_deg-old_deg) * &
                         count(grid%edge_type(edge,:)/=PERIODIC_SLAVE)
      else
         delta_dof_own = 0
      endif
   else
      grid%dof = grid%dof + (new_deg - old_deg) * &
                            count(grid%edge_type(edge,:)/=PERIODIC_SLAVE)
      if (grid%element(grid%edge(edge)%assoc_elem)%iown) then
        grid%dof_own = grid%dof_own + (new_deg - old_deg) * &
                       count(grid%edge_type(edge,:)/=PERIODIC_SLAVE)
      endif
   endif
else
   if (present(delta_dof)) then
      delta_dof = 0
      delta_dof_own = 0
   endif
endif

! See if the amount of memory for the solution is correct.
! To save on how often allocation/deallocation occurs,
! if more memory is needed, allocate for degree+2,
! don't deallocate until the new degree is 3 less than the current allocation

if (associated_edge_soln(grid,edge)) then
   old_size = size_edge_soln(grid,edge,dim=BASIS_DIM)
else
   old_size = 0
endif

if (new_deg-1 < old_size-2) then
   new_size = new_deg-1
elseif (new_deg-1 > old_size) then
   new_size = new_deg+1
else
   new_size = old_size
endif

! if necessary, reallocate solution memory

if (new_size /= old_size) then
   call reallocate_edge_soln(grid,edge,size_basis=new_size)
   if (grid%have_true) then
      call reallocate_edge_exact(grid,edge,size_basis=new_size)
   endif
   if (associated_edge_oldsoln(grid,edge)) then
      call reallocate_edge_oldsoln(grid,edge,size_basis=new_size)
   endif
endif

! Set Dirichlet boundary conditions on Dirichlet edges, set solution to 0 for
! other new components, and set true solution.

if (new_deg /= old_deg .and. new_size > 0) then
   if (grid%have_true) then
      call set_edge_exact(grid,edge,colon,colon,colon,0.0_my_real)
   endif
   do i=1,grid%system_size
      if (grid%edge_type(edge,i) == DIRICHLET) then
         call edge_exact(grid,edge,i,"d")
      elseif (new_deg > old_deg) then
         call set_edge_soln(grid,edge,(/old_deg,new_size/),i,colon,0.0_my_real)
      elseif (new_deg < old_deg) then
         call set_edge_soln(grid,edge,(/new_deg,new_size/),i,colon,0.0_my_real)
      endif
      if (grid%have_true) call edge_exact(grid,edge,i,"t")
   end do
endif

! indicate the error indicators should be recomputed later

!$omp critical (critical_edge_rule)
do i=1,size(elem)
   grid%element_errind(elem(i),:) = -1.0_my_real
end do
!$omp end critical (critical_edge_rule)

deallocate(elem)

end subroutine enforce_edge_rule

!          -----------------
subroutine enforce_face_rule(grid,refine_control,face,delta_dof,delta_dof_own)
!          -----------------

!----------------------------------------------------
! This routine enforces either the minimum or maximum rule for face
!
! Thread safe if:
!   delta_dof and delta_dof_own are present
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(refine_options), intent(in) :: refine_control
integer, intent(in) :: face
integer, intent(out), optional :: delta_dof, delta_dof_own
!----------------------------------------------------
! Local variables:

integer :: i, old_deg, new_deg, old_size, new_size
integer, pointer :: elem(:)
!----------------------------------------------------
! Begin executable code

if (present(delta_dof) .neqv. present(delta_dof_own)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("enforce_face_rule: delta_* must all be present or absent")
   stop
endif

! get the elements that share this edge

call get_face_elements(grid,face,elem)

! determine what the degree of this face should be

if (refine_control%edge_rule == MINIMUM_RULE) then
   new_deg = huge(0)
else ! (refine_control%edge_rule == MAXIMUM_RULE)
   new_deg = 0
endif
do i=1,size(elem)
   if (refine_control%edge_rule == MINIMUM_RULE) then
      new_deg = min(grid%element(elem(i))%degree,new_deg)
   else ! (refine_control%edge_rule == MAXIMUM_RULE)
      new_deg = max(grid%element(elem(i))%degree,new_deg)
   endif
end do

! assign the new degree

old_deg = grid%face(face)%degree
if (new_deg /= old_deg) then
   grid%face(face)%degree = new_deg
   if (present(delta_dof)) then
      delta_dof = (face_dof(new_deg)-face_dof(old_deg)) * &
                  count(grid%face_type(face,:)/=PERIODIC_SLAVE)
      if (grid%element(grid%face(face)%assoc_elem)%iown) then
         delta_dof_own = (face_dof(new_deg)-face_dof(old_deg)) * &
                         count(grid%face_type(face,:)/=PERIODIC_SLAVE)
      else
         delta_dof_own = 0
      endif
   else
      grid%dof = grid%dof + (face_dof(new_deg) - face_dof(old_deg)) * &
                            count(grid%face_type(face,:)/=PERIODIC_SLAVE)
      if (grid%element(grid%face(face)%assoc_elem)%iown) then
        grid%dof_own = grid%dof_own + (face_dof(new_deg) - face_dof(old_deg))* &
                       count(grid%face_type(face,:)/=PERIODIC_SLAVE)
      endif
   endif
else
   if (present(delta_dof)) then
      delta_dof = 0
      delta_dof_own = 0
   endif
endif

! See if the amount of memory for the solution is correct.
! To save on how often allocation/deallocation occurs,
! if more memory is needed, allocate for degree+2,
! don't deallocate until the new degree is 3 less than the current allocation

if (associated_face_soln(grid,face)) then
   old_size = size_face_soln(grid,face,dim=BASIS_DIM)
else
   old_size = 0
endif

new_size = face_dof(new_deg)
if (new_size < old_size-2) then
elseif (new_size > old_size) then
   new_size = face_dof(new_deg+2)
else
   new_size = old_size
endif

! if necessary, reallocate solution memory

if (new_size /= old_size) then
   call reallocate_face_soln(grid,face,size_basis=new_size)
   if (grid%have_true) then
      call reallocate_face_exact(grid,face,size_basis=new_size)
   endif
   if (associated_face_oldsoln(grid,face)) then
      call reallocate_face_oldsoln(grid,face,size_basis=new_size)
   endif
endif

! Set Dirichlet boundary conditions on Dirichlet faces, set solution to 0 for
! other new components, and set true solution.

if (new_deg /= old_deg .and. new_size > 0) then
   if (grid%have_true) then
      call set_face_exact(grid,face,colon,colon,colon,0.0_my_real)
   endif
   do i=1,grid%system_size
      if (grid%face_type(face,i) == DIRICHLET) then
         call face_exact(grid,face,i,"d")
      elseif (new_deg > old_deg) then
         call set_face_soln(grid,face,(/face_dof(old_deg)+1,new_size/),i, &
                            colon,0.0_my_real)
      elseif (new_deg < old_deg) then
         call set_face_soln(grid,face,(/face_dof(new_deg)+1,new_size/),i, &
                            colon,0.0_my_real)
      endif
      if (grid%have_true) call face_exact(grid,face,i,"t")
   end do
endif

! indicate the error indicators should be recomputed later

!$omp critical (critical_face_rule)
do i=1,size(elem)
   grid%element_errind(elem(i),:) = -1.0_my_real
end do
!$omp end critical (critical_face_rule)

deallocate(elem)

end subroutine enforce_face_rule

!          ---------------
subroutine before_h_refine(grid,element_list,nelem,reftype,vert_lid,edge_lid, &
                           elem_lid,desired_level,desired_degree,elem_list, &
                           was_hcoarsened)
!          ---------------

!----------------------------------------------------
! This routine performs parts of h refinement that must be done by a single
! OpenMP thread before the OpenMP-parallel refinement.
! All elements in element_list must be on the same level, and there must only
! be one from a compatibly divisible pair.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
integer, intent(in) :: element_list(:), nelem
character(len=*), optional, pointer :: reftype(:)
integer, intent(out) :: vert_lid(:,:), edge_lid(:,:), elem_lid(:,:)
integer, optional, pointer :: desired_level(:), desired_degree(:,:), &
                              elem_list(:)
logical(small_logical), optional, pointer :: was_hcoarsened(:)
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

! Not supporting OpenMP in 3D

end subroutine before_h_refine

!                    ----------------
recursive subroutine h_refine_element(grid,parent,errcode,refcont, &
                                      solver_control,reftype,reduce_p, &
                     reduce_p_max,vert_child_lid,edge_child_lid,elem_child_lid,&
                     delta_nelem, &
                     delta_nelem_leaf,delta_nelem_leaf_own,delta_nedge, &
                     delta_nedge_leaf,delta_nedge_leaf_own,delta_nvert, &
                     delta_nvert_own, delta_dof,delta_dof_own, &
                     max_nlev,desired_level,desired_degree,elem_list, &
                     was_hcoarsened,override_unrefinable, &
                     use_compatibility_chain,been_there,need_closure)
!                    ----------------

!----------------------------------------------------
! This routine refines tetrahedron parent by bisection.
! errcode = 0 for success, 1 if the grid is full, 2 if too many levels,
! 3 if the element is not a leaf, 4 if it cannot be refined for another reason
!
! Not thread safe
!
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
integer, intent(in) :: parent
integer, intent(inout) :: errcode
type(refine_options), intent(in) :: refcont
type(solver_options), intent(in) :: solver_control
character(len=*), optional, pointer :: reftype(:)
integer, optional, intent(in) :: reduce_p
integer, optional, intent(inout) :: reduce_p_max
integer, optional, intent(in) :: vert_child_lid(:), edge_child_lid(:), &
                                 elem_child_lid(:)
integer, optional, intent(out) :: max_nlev
integer, optional, pointer :: desired_level(:), desired_degree(:,:), &
                              elem_list(:)
logical(small_logical), optional, pointer :: was_hcoarsened(:)
logical, optional, intent(in) :: override_unrefinable
logical, optional, intent(in) :: use_compatibility_chain
logical(small_logical), optional, pointer :: been_there(:)
logical, optional, intent(out) :: need_closure
integer, optional, intent(out) :: delta_nelem, delta_nelem_leaf, &
                                  delta_nelem_leaf_own, delta_nedge, &
                                  delta_nedge_leaf, delta_nedge_leaf_own, &
                                  delta_nvert, delta_nvert_own, delta_dof, &
                                  delta_dof_own

!----------------------------------------------------
! Local variables:

integer :: refinement_edge, refinement_face, child1, child2, face1, face2, &
           face3, face4, face5, edge1, edge2, edge3, edge4, vert1, i, j, &
           astat, nfound, parent_face(2), vert, edge, face, child_vert(2), &
           common_vert(2), child_face(2), child_edge(4), common_edge, jerr, &
           mate_face_1, mate_face_2, mate_face_3, mate_face_4, mate_edge_1, &
           mate_edge_2, mate_edge_3, mate_edge_4, mate_vert_1, itemp, pos(4), &
           comp, inc
integer :: have_it, elem
integer :: bctype(grid%system_size)
real(my_real) :: bccoef(grid%system_size,grid%system_size), &
                 bcrhs(grid%system_size)
type(hash_key) :: parent_neighbors(NEIGHBORS_PER_ELEMENT), htemp
logical :: new_ref_edge, new_face1, new_face2, do_face1, do_face2, &
           mate_edge_1_in_cycle, edge1_in_cycle, mate_vert_1_in_cycle, &
           my_need_closure, vert1_in_cycle
integer :: refedge, nshare_refedge
integer, pointer :: share_refedge(:)
logical :: needs_refinement, any_refined, loc_use_compatibility_chain
integer :: children(2)
!----------------------------------------------------
! Begin executable code

! check for things not yet supported

if (omp_get_num_threads() /= 1) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("h_refine_element: OpenMP not yet supported in 3D")
   stop
endif

! some preliminary stuff

errcode = 0
if (present(max_nlev)) max_nlev = 0

! can't refine a non-leaf element

child1 = get_child_lid(grid%element(parent)%gid,1,grid%elem_hash)
if (child1 /= NO_CHILD) then
   errcode = 3
   return
endif

! get the refinement edge

refinement_edge = grid%element(parent)%refinement_edge

! determine the two faces of the parent that contain the refinement edge

nfound = 0
do i=1,FACES_PER_ELEMENT
   if (any(grid%face(grid%element(parent)%face(i))%edge==refinement_edge)) then
      nfound = nfound + 1
      parent_face(nfound) = grid%element(parent)%face(i)
   endif
end do
if (nfound /= 2) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("h_refine_element: didn't find two faces with refinement edge")
   stop
endif

! I don't know why it helps to put the smaller face number first, but it gave
! me better grids, and grids that are more like those I got before ordering
! the faces to be opposite the corresponding vertex

if (parent_face(2) < parent_face(1)) then
   itemp = parent_face(2)
   parent_face(2) = parent_face(1)
   parent_face(1) = itemp
endif

! if the maximum number of levels would be exceeded or the element is too
! small or a hash key would overflow, don't h-refine unless we are overriding
! this check.

if (.not. present(override_unrefinable)) then
   if (grid%element(parent)%level >= refcont%max_lev .or. &
       element_is_too_small(parent,grid,refcont) .or. &
       hash_overflow(grid%element(parent)%gid,4,3) .or. &
       hash_overflow(grid%face(parent_face(1))%gid,4,1) .or. &
       hash_overflow(grid%face(parent_face(2))%gid,4,3)) then
      errcode = 2
      return
   endif
endif

! if we have already seen this element in the recursive chain, then this would
! cause an infinite loop, so abort this compatibility chain and indicate that
! closure needs to be completed by removing hanging nodes

if (present(use_compatibility_chain)) then
   loc_use_compatibility_chain = use_compatibility_chain
else
   loc_use_compatibility_chain = .false.
endif

if (present(been_there)) then
   if (loc_use_compatibility_chain .and. been_there(parent)) then
      if (present(need_closure)) need_closure = .true.
      been_there(parent) = .false.
      return
   else
      been_there(parent) = .true.
   endif
endif

my_need_closure = .false.

! see if refinement of this element requires extending the number of levels.

if (grid%element(parent)%level >= size(grid%head_level_elem)) then
   call extend_nlev(grid)
   if (ierr /= NO_ERROR) return
endif

! If we are using the compatibility chain, refine elements that contain the
! refinement edge until their refinement edge is the refinement edge.

if (loc_use_compatibility_chain) then

! Use a different variable for the refinement edge so I can go through
! periodic mates.

   refedge = refinement_edge

! Repeat this until all elements that share the refinement edge have the
! refinement edge as their refinement edge.

   any_refined = .true.
   outer: do while (any_refined)
      any_refined = .false.

! Get all the elements that share the refinement edge

      call get_edge_elements(grid,refedge,share_refedge)
      nshare_refedge = size(share_refedge)

! Refine elements that share the refinement edge but have a different
! refinement edge.

      do i=1,nshare_refedge
         if (grid%element(share_refedge(i))%isleaf) then
            needs_refinement = refedge /= &
                               grid%element(share_refedge(i))%refinement_edge
         else
            needs_refinement = .false.
         endif
         if (needs_refinement) then
            call h_refine_element(grid,share_refedge(i),errcode,refcont, &
                                  solver_control,reftype,reduce_p,reduce_p_max,&
                                  vert_child_lid,edge_child_lid,elem_child_lid,&
                                  delta_nelem, delta_nelem_leaf, &
                                  delta_nelem_leaf_own,delta_nedge,  &
                                  delta_nedge_leaf,delta_nedge_leaf_own, &
                                  delta_nvert, delta_nvert_own, delta_dof, &
                                  delta_dof_own, max_nlev,desired_level, &
                                  desired_degree, elem_list,was_hcoarsened, &
                                  override_unrefinable=.true., &
                                  use_compatibility_chain=.true., &
                                  been_there=been_there, &
                                  need_closure=my_need_closure)
            if (errcode /= 0) then
               ierr = PHAML_INTERNAL_ERROR
               call fatal("h_refine_element: refinement for compatibility chain failed")
               stop
            endif
            if (my_need_closure) exit outer
            any_refined = .true.
         endif
      end do

      if (any_refined) then
         deallocate(share_refedge)

! If we are done with elements around refedge, see if it is periodic and do
! the periodic mate if it is.

      else
         if (is_periodic_edge(refedge,grid)) then
            refedge = grid%edge(refedge)%next
            deallocate(share_refedge)
            if (refedge == refinement_edge) then
               call get_edge_elements(grid,refedge,share_refedge)
               nshare_refedge = size(share_refedge)
            else
               any_refined = .true.
            endif
         endif
      endif

   end do outer

   if (present(been_there)) been_there(parent) = .false.

! It is possible for parent to have been refined via the compatibility chain.

   if (.not. grid%element(parent)%isleaf) return

endif ! using compatibility chain

! get the neighbors of the parent in the order of the face they share

parent_neighbors = get_ordered_neighbors(parent,jerr)

! If jerr is 1, we didn't find all matching faces; probably a neighbor has not
! been refined yet.  Those not found were assigned BOUNDARY.  parent_neighbors
! is only used to assign neighbor_hint, so use the parent's hint as the element
! that shares the faces not matched.  If the face really is BOUNDARY, then the
! parent's hint will be BOUNDARY, so that is not lost.

if (jerr == 1) then
   do i=1,FACES_PER_ELEMENT
      if (parent_neighbors(i) == BOUNDARY) then
         parent_neighbors(i) = grid%element(parent)%neighbor_hint(i)
      endif
   end do
endif

! get the lids for the first two children

child1 = get_next_free_elem(grid,errcode,reftype,desired_level,desired_degree, &
                            elem_list,was_hcoarsened,been_there)
if (errcode /= 0) return
child2 = get_next_free_elem(grid,errcode,reftype,desired_level,desired_degree, &
                            elem_list,was_hcoarsened,been_there)
if (errcode /= 0) return

! if the refinement edge has already been refined then locate the lids of
! the entities that were created by that bisection, otherwise get new lids

new_ref_edge = grid%edge(refinement_edge)%child(1) == NO_CHILD
if (new_ref_edge) then

! refinement edge is not refined, get new lids for the new vertex and edges
! that replace the refinement edge

   edge1 = get_next_free_edge(grid,errcode)
   if (errcode /= 0) return
   edge2 = get_next_free_edge(grid,errcode)
   if (errcode /= 0) return
   vert1 = get_next_free_vert(grid,errcode)
   if (errcode /= 0) return

! refinement edge is refined, identify the lids

else

   if (grid%edge(grid%edge(refinement_edge)%child(1))%vertex(1) == &
       grid%edge(refinement_edge)%vertex(1)) then
      edge1 = grid%edge(refinement_edge)%child(1)
      edge2 = grid%edge(refinement_edge)%child(2)
      vert1 = grid%edge(edge1)%vertex(2)
   elseif (grid%edge(grid%edge(refinement_edge)%child(1))%vertex(2) == &
       grid%edge(refinement_edge)%vertex(1)) then
      edge1 = grid%edge(refinement_edge)%child(1)
      edge2 = grid%edge(refinement_edge)%child(2)
      vert1 = grid%edge(edge1)%vertex(1)
   elseif (grid%edge(grid%edge(refinement_edge)%child(2))%vertex(1) == &
       grid%edge(refinement_edge)%vertex(1)) then
      edge1 = grid%edge(refinement_edge)%child(2)
      edge2 = grid%edge(refinement_edge)%child(1)
      vert1 = grid%edge(edge1)%vertex(2)
   elseif (grid%edge(grid%edge(refinement_edge)%child(2))%vertex(2) == &
       grid%edge(refinement_edge)%vertex(1)) then
      edge1 = grid%edge(refinement_edge)%child(2)
      edge2 = grid%edge(refinement_edge)%child(1)
      vert1 = grid%edge(edge1)%vertex(1)
   else
      ierr = PHAML_INTERNAL_ERROR
      call fatal("h_refine_element: couldn't identify lids in refined refinement edge")
      stop
   endif

endif

! for each of the parent faces, if the face has already been refined then
! locate the lids of the entities that were created by that bisection,
! otherwise get new lids

new_face1 = grid%face(parent_face(1))%child(1) == NO_CHILD
if (new_face1) then

   face1 = get_next_free_face(grid,errcode)
   if (errcode /= 0) return
   face2 = get_next_free_face(grid,errcode)
   if (errcode /= 0) return
   edge3 = get_next_free_edge(grid,errcode)
   if (errcode /= 0) return

else

   if (any(grid%face(grid%face(parent_face(1))%child(1))%edge==edge1)) then
      face1 = grid%face(parent_face(1))%child(1)
      face2 = grid%face(parent_face(1))%child(2)
   elseif (any(grid%face(grid%face(parent_face(1))%child(1))%edge==edge2)) then
      face1 = grid%face(parent_face(1))%child(2)
      face2 = grid%face(parent_face(1))%child(1)
   else
      ierr = PHAML_INTERNAL_ERROR
      call fatal("h_refine_element: failed to find refined edge in parent face 1")
      stop
   endif
   if (grid%face(face1)%edge(1) /= edge1 .and. &
       grid%face(face1)%edge(1) /= grid%face(parent_face(1))%edge(1) .and. &
       grid%face(face1)%edge(1) /= grid%face(parent_face(1))%edge(2) .and. &
       grid%face(face1)%edge(1) /= grid%face(parent_face(1))%edge(3)) then
      edge3 = grid%face(face1)%edge(1)
   elseif (grid%face(face1)%edge(2) /= edge1 .and. &
           grid%face(face1)%edge(2) /= grid%face(parent_face(1))%edge(1) .and. &
           grid%face(face1)%edge(2) /= grid%face(parent_face(1))%edge(2) .and. &
           grid%face(face1)%edge(2) /= grid%face(parent_face(1))%edge(3)) then
      edge3 = grid%face(face1)%edge(2)
   elseif (grid%face(face1)%edge(3) /= edge1 .and. &
           grid%face(face1)%edge(3) /= grid%face(parent_face(1))%edge(1) .and. &
           grid%face(face1)%edge(3) /= grid%face(parent_face(1))%edge(2) .and. &
           grid%face(face1)%edge(3) /= grid%face(parent_face(1))%edge(3)) then
      edge3 = grid%face(face1)%edge(3)
   else
      ierr = PHAML_INTERNAL_ERROR
      call fatal("h_refine_element: failed to find edge1 on face1")
      stop
   endif

endif

new_face2 = grid%face(parent_face(2))%child(1) == NO_CHILD
if (new_face2) then

   face3 = get_next_free_face(grid,errcode)
   if (errcode /= 0) return
   face4 = get_next_free_face(grid,errcode)
   if (errcode /= 0) return
   edge4 = get_next_free_edge(grid,errcode)
   if (errcode /= 0) return

else

   if (any(grid%face(grid%face(parent_face(2))%child(1))%edge==edge1)) then
      face3 = grid%face(parent_face(2))%child(1)
      face4 = grid%face(parent_face(2))%child(2)
   elseif (any(grid%face(grid%face(parent_face(2))%child(1))%edge==edge2)) then
      face3 = grid%face(parent_face(2))%child(2)
      face4 = grid%face(parent_face(2))%child(1)
   else
      ierr = PHAML_INTERNAL_ERROR
      call fatal("h_refine_element: failed to find refined edge in parent face 2")
      stop
   endif
   if (grid%face(face3)%edge(1) /= edge1 .and. &
       grid%face(face3)%edge(1) /= grid%face(parent_face(2))%edge(1) .and. &
       grid%face(face3)%edge(1) /= grid%face(parent_face(2))%edge(2) .and. &
       grid%face(face3)%edge(1) /= grid%face(parent_face(2))%edge(3)) then
      edge4 = grid%face(face3)%edge(1)
   elseif (grid%face(face3)%edge(2) /= edge1 .and. &
           grid%face(face3)%edge(2) /= grid%face(parent_face(2))%edge(1) .and. &
           grid%face(face3)%edge(2) /= grid%face(parent_face(2))%edge(2) .and. &
           grid%face(face3)%edge(2) /= grid%face(parent_face(2))%edge(3)) then
      edge4 = grid%face(face3)%edge(2)
   elseif (grid%face(face3)%edge(3) /= edge1 .and. &
           grid%face(face3)%edge(3) /= grid%face(parent_face(2))%edge(1) .and. &
           grid%face(face3)%edge(3) /= grid%face(parent_face(2))%edge(2) .and. &
           grid%face(face3)%edge(3) /= grid%face(parent_face(2))%edge(3)) then
      edge4 = grid%face(face3)%edge(3)
   else
      ierr = PHAML_INTERNAL_ERROR
      call fatal("h_refine_element: failed to find edge1 on face1")
      stop
   endif

endif

! identify which vertices from the parent go in each child or are common

child_vert(1) = grid%edge(refinement_edge)%vertex(1)
child_vert(2) = grid%edge(refinement_edge)%vertex(2)
nfound = 0
do i=1,VERTICES_PER_ELEMENT
   vert = grid%element(parent)%vertex(i)
   if (vert /= child_vert(1) .and. vert /= child_vert(2)) then
      nfound = nfound + 1
      common_vert(nfound) = vert
   endif
end do
if (nfound /= 2) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("h_refine_element: didn't find 2 common vertices")
   stop
endif

! make sure the first common vertex is on the first face

if (any(grid%face(parent_face(2))%vertex == common_vert(1))) then
   i = common_vert(1)
   common_vert(1) = common_vert(2)
   common_vert(2) = i
endif

! determine which faces from the parent go in each child

nfound = 0
do i=1,FACES_PER_ELEMENT
   face = grid%element(parent)%face(i)
   if (face /= parent_face(1) .and. face /= parent_face(2)) then
      if (any(grid%face(face)%vertex == child_vert(1))) then
         nfound = nfound + 1
         child_face(1) = face
      endif
      if (any(grid%face(face)%vertex == child_vert(2))) then
         nfound = nfound + 1
         child_face(2) = face
      endif
   endif
end do
if (nfound /= 2) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("h_refine_element: didn't find 2 child faces")
   stop
endif

! determine which edges from the parent go in each child and which is common

common_edge = -1
nfound = 0
do i=1,EDGES_PER_ELEMENT
   edge = grid%element(parent)%edge(i)
   if (edge == refinement_edge) cycle
   if (any(grid%edge(edge)%vertex == child_vert(1))) then
      nfound = nfound + 1
      child_edge(nfound) = edge
   elseif (any(grid%edge(edge)%vertex == common_vert(1)) .and. &
           any(grid%edge(edge)%vertex == common_vert(2))) then
      common_edge = edge
   endif
end do
do i=1,EDGES_PER_ELEMENT
   edge = grid%element(parent)%edge(i)
   if (edge == refinement_edge) cycle
   if (any(grid%edge(edge)%vertex == child_vert(2))) then
      nfound = nfound + 1
      child_edge(nfound) = edge
   endif
end do
if (nfound /= 4) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("h_refine_element: didn't identify which edges go in which child")
   stop
endif
if (common_edge == -1) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("h_refine_element: didn't identify common edge for children")
   stop
endif

! make sure child edges 1 and 3 are on the first face

if (any(grid%face(parent_face(2))%edge == child_edge(1))) then
   i = child_edge(1)
   child_edge(1) = child_edge(2)
   child_edge(2) = i
endif
if (any(grid%face(parent_face(2))%edge == child_edge(3))) then
   i = child_edge(3)
   child_edge(3) = child_edge(4)
   child_edge(4) = i
endif

! define the new vertex and edges from bisecting the refinement edge, if they
! didn't already exist

if (new_ref_edge) then

   grid%vertex(vert1)%gid = 2*grid%element(parent)%gid
   grid%vertex(vert1)%assoc_elem = child1
   nullify(grid%vertex(vert1)%boundary_vertex)
   if (grid%edge_type(refinement_edge,1) == INTERIOR .or. & ! vert1 is interior
       grid%edge_type(refinement_edge,1) == INTERIOR_DIRICHLET) then
      grid%vertex(vert1)%coord%x = &
               (grid%vertex(grid%edge(refinement_edge)%vertex(1))%coord%x + &
                grid%vertex(grid%edge(refinement_edge)%vertex(2))%coord%x)/2
      grid%vertex(vert1)%coord%y = &
               (grid%vertex(grid%edge(refinement_edge)%vertex(1))%coord%y + &
                grid%vertex(grid%edge(refinement_edge)%vertex(2))%coord%y)/2
      grid%vertex(vert1)%coord%z = &
               (grid%vertex(grid%edge(refinement_edge)%vertex(1))%coord%z + &
                grid%vertex(grid%edge(refinement_edge)%vertex(2))%coord%z)/2
   else ! vert1 is a new boundary vertex
      call define_boundary_vertex(grid,refinement_edge,vert1)
   endif
   grid%vertex(vert1)%bmark = grid%edge(refinement_edge)%bmark
   if (associated(grid%edge(refinement_edge)%alt_bmark)) then
      allocate(grid%vertex(vert1)%alt_bmark( &
                  size(grid%edge(refinement_edge)%alt_bmark)),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in h_refine_element")
         stop
      endif
      grid%vertex(vert1)%alt_bmark = grid%edge(refinement_edge)%alt_bmark
   else
      nullify(grid%vertex(vert1)%alt_bmark)
   endif
   grid%vertex(vert1)%tags = grid%edge(refinement_edge)%tags
   grid%vertex(vert1)%gmsh_element = .false.
   grid%vertex_type(vert1,:) = grid%edge_type(refinement_edge,:)
   if (any(grid%vertex_type(vert1,:) == DIRICHLET) .or. &
       any(grid%vertex_type(vert1,:) == INTERIOR_DIRICHLET) .or. &
       any(grid%vertex_type(vert1,:) == PERIODIC_MASTER_DIR) .or. &
       any(grid%vertex_type(vert1,:) == PERIODIC_SLAVE_DIR)) then
      secret_elem = parent
      call my_bconds(grid%vertex(vert1)%coord%x,grid%vertex(vert1)%coord%y, &
                     grid%vertex(vert1)%coord%z,grid%vertex(vert1)%bmark, &
                     bctype,bccoef,bcrhs)
   endif
   do i=1,grid%system_size
      if (grid%vertex_type(vert1,i) == DIRICHLET .or. &
          grid%vertex_type(vert1,i) == INTERIOR_DIRICHLET .or. &
          grid%vertex_type(vert1,i) == PERIODIC_MASTER_DIR .or. &
          grid%vertex_type(vert1,i) == PERIODIC_SLAVE_DIR) then
         call set_vert_soln(grid,vert1,i,colon,bcrhs(i))
      else
         call set_vert_soln(grid,vert1,i,colon, &
           (get_vert_soln(grid,grid%edge(refinement_edge)%vertex(1),i,colon) + &
            get_vert_soln(grid,grid%edge(refinement_edge)%vertex(2),i,colon))/2)
      endif
   end do
   if (grid%have_true) then
      do j=1,max(1,grid%num_eval)
         do i=1,grid%system_size
            call set_vert_exact(grid,vert1,i,j, &
                                my_trues(grid%vertex(vert1)%coord%x, &
                                         grid%vertex(vert1)%coord%y, &
                                         grid%vertex(vert1)%coord%z,i,j))
         end do
      end do
   endif
   grid%vertex(vert1)%next = vert1

   call hash_insert(grid%vertex(vert1)%gid,vert1,grid%vert_hash)

! edge 1

   grid%edge(edge1)%gid = 4*grid%element(parent)%gid
   grid%edge(edge1)%vertex = (/ child_vert(1),vert1 /)
   grid%edge(edge1)%bmark = grid%edge(refinement_edge)%bmark
   if (associated(grid%edge(refinement_edge)%alt_bmark)) then
      allocate(grid%edge(edge1)%alt_bmark( &
                  size(grid%edge(refinement_edge)%alt_bmark)),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in h_refine_element")
         stop
      endif
      grid%edge(edge1)%alt_bmark = grid%edge(refinement_edge)%alt_bmark
   else
      nullify(grid%edge(edge1)%alt_bmark)
   endif
   grid%edge(edge1)%tags = grid%edge(refinement_edge)%tags
   grid%edge(edge1)%gmsh_element = grid%edge(refinement_edge)%gmsh_element
   grid%edge(edge1)%degree = grid%edge(refinement_edge)%degree
   grid%edge_type(edge1,:) = grid%edge_type(refinement_edge,:)
   grid%edge(edge1)%assoc_elem = child1
   grid%edge(edge1)%child = NO_CHILD
   call nullify_edge_oldsoln(grid,edge1)
   grid%edge(edge1)%next = edge1

   call hash_insert(grid%edge(edge1)%gid,edge1,grid%edge_hash)

! edge 2

   grid%edge(edge2)%gid = 4*grid%element(parent)%gid + 1
   grid%edge(edge2)%vertex = (/ child_vert(2),vert1 /)
   grid%edge(edge2)%bmark = grid%edge(refinement_edge)%bmark
   if (associated(grid%edge(refinement_edge)%alt_bmark)) then
      allocate(grid%edge(edge2)%alt_bmark( &
                  size(grid%edge(refinement_edge)%alt_bmark)),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in h_refine_element")
         stop
      endif
      grid%edge(edge2)%alt_bmark = grid%edge(refinement_edge)%alt_bmark
   else
      nullify(grid%edge(edge2)%alt_bmark)
   endif
   grid%edge(edge2)%tags = grid%edge(refinement_edge)%tags
   grid%edge(edge2)%gmsh_element = grid%edge(refinement_edge)%gmsh_element
   grid%edge(edge2)%degree = grid%edge(refinement_edge)%degree
   grid%edge_type(edge2,:) = grid%edge_type(refinement_edge,:)
   grid%edge(edge2)%assoc_elem = child2
   grid%edge(edge2)%child = NO_CHILD
   call nullify_edge_oldsoln(grid,edge2)
   grid%edge(edge2)%next = edge2

   call hash_insert(grid%edge(edge2)%gid,edge2,grid%edge_hash)

endif ! refinement edge is new

! Find the position of child_vert(1), child_vert(2), common_vert(1) and
! common_vert(2) in the vertices of the parent

do i=1,VERTICES_PER_ELEMENT
   if (grid%element(parent)%vertex(i) == child_vert(1)) pos(1) = i
   if (grid%element(parent)%vertex(i) == child_vert(2)) pos(2) = i
   if (grid%element(parent)%vertex(i) == common_vert(1)) pos(3) = i
   if (grid%element(parent)%vertex(i) == common_vert(2)) pos(4) = i
end do

! define the new edge and faces from bisecting parent face 1, if they
! didn't already exist

if (new_face1) then

   refinement_face = parent_face(1)
   grid%edge(edge3)%gid = 4*grid%element(parent)%gid + 2
   grid%edge(edge3)%vertex = (/ common_vert(1),vert1 /)
   grid%edge(edge3)%bmark = grid%face(refinement_face)%bmark
   if (associated(grid%face(refinement_face)%alt_bmark)) then
      allocate(grid%edge(edge3)%alt_bmark( &
                  size(grid%face(refinement_face)%alt_bmark)),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in h_refine_element")
         stop
      endif
      grid%edge(edge3)%alt_bmark = grid%face(refinement_face)%alt_bmark
   else
      nullify(grid%edge(edge3)%alt_bmark)
   endif
   grid%edge(edge3)%tags = grid%face(refinement_face)%tags
   grid%edge(edge3)%gmsh_element = .false.
   grid%edge(edge3)%degree = grid%face(refinement_face)%degree
   grid%edge_type(edge3,:) = grid%face_type(refinement_face,:)
   grid%edge(edge3)%assoc_elem = child1
   grid%edge(edge3)%child = NO_CHILD
   call nullify_edge_oldsoln(grid,edge3)
   grid%edge(edge3)%next = edge3

   call hash_insert(grid%edge(edge3)%gid,edge3,grid%edge_hash)

! face 1

   grid%face(face1)%gid = 4*grid%face(parent_face(1))%gid

! vertices are common_vert(1), child_vert(1) and vert1 in the same order as
! in the parent face (vert1 replaces child_vert(2)) but cyclically shifted
! so that vert1 is last

   do i=1,3
      if (grid%face(parent_face(1))%vertex(i) == child_vert(2)) then
         select case(i)
         case (1)
            grid%face(face1)%vertex(1) = grid%face(parent_face(1))%vertex(2)
            grid%face(face1)%vertex(2) = grid%face(parent_face(1))%vertex(3)
            grid%face(face1)%vertex(3) = vert1
         case (2)
            grid%face(face1)%vertex(1) = grid%face(parent_face(1))%vertex(3)
            grid%face(face1)%vertex(2) = grid%face(parent_face(1))%vertex(1)
            grid%face(face1)%vertex(3) = vert1
         case (3)
            grid%face(face1)%vertex(1) = grid%face(parent_face(1))%vertex(1)
            grid%face(face1)%vertex(2) = grid%face(parent_face(1))%vertex(2)
            grid%face(face1)%vertex(3) = vert1
         end select
         exit
      endif
   end do
   grid%face(face1)%edge = (/ child_edge(1), edge1, edge3 /)
   grid%face(face1)%bmark = grid%face(refinement_face)%bmark
   if (associated(grid%face(refinement_face)%alt_bmark)) then
      allocate(grid%face(face1)%alt_bmark( &
                  size(grid%face(refinement_face)%alt_bmark)),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in h_refine_element")
         stop
      endif
      grid%face(face1)%alt_bmark = grid%face(refinement_face)%alt_bmark
   else
      nullify(grid%face(face1)%alt_bmark)
   endif
   grid%face(face1)%tags = grid%face(refinement_face)%tags
   grid%face(face1)%gmsh_element = grid%face(refinement_face)%gmsh_element
   grid%face(face1)%degree = grid%face(refinement_face)%degree
   grid%face_type(face1,:) = grid%face_type(refinement_face,:)
   grid%face(face1)%assoc_elem = child1
   grid%face(face1)%marked_edge = grid%face(face1)%edge(1)
   grid%face(face1)%child = NO_CHILD
   grid%face(face1)%next = 0
   call nullify_face_oldsoln(grid,face1)
   grid%face(face1)%next = face1

   call hash_insert(grid%face(face1)%gid,face1,grid%face_hash)

! face 2

   grid%face(face2)%gid = 4*grid%face(parent_face(1))%gid + 1

! vertices are common_vert(1), child_vert(2) and vert1 in the same order as
! in the parent face (vert1 replaces child_vert(1)) but cyclically shifted
! so that vert1 is last

   do i=1,3
      if (grid%face(parent_face(1))%vertex(i) == child_vert(1)) then
         select case(i)
         case (1)
            grid%face(face2)%vertex(1) = grid%face(parent_face(1))%vertex(2)
            grid%face(face2)%vertex(2) = grid%face(parent_face(1))%vertex(3)
            grid%face(face2)%vertex(3) = vert1
         case (2)
            grid%face(face2)%vertex(1) = grid%face(parent_face(1))%vertex(3)
            grid%face(face2)%vertex(2) = grid%face(parent_face(1))%vertex(1)
            grid%face(face2)%vertex(3) = vert1
         case (3)
            grid%face(face2)%vertex(1) = grid%face(parent_face(1))%vertex(1)
            grid%face(face2)%vertex(2) = grid%face(parent_face(1))%vertex(2)
            grid%face(face2)%vertex(3) = vert1
         end select
         exit
      endif
   end do
   grid%face(face2)%edge = (/ child_edge(3), edge2, edge3 /)
   grid%face(face2)%bmark = grid%face(refinement_face)%bmark
   if (associated(grid%face(refinement_face)%alt_bmark)) then
      allocate(grid%face(face2)%alt_bmark( &
                  size(grid%face(refinement_face)%alt_bmark)),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in h_refine_element")
         stop
      endif
      grid%face(face2)%alt_bmark = grid%face(refinement_face)%alt_bmark
   else
      nullify(grid%face(face2)%alt_bmark)
   endif
   grid%face(face2)%tags = grid%face(refinement_face)%tags
   grid%face(face2)%gmsh_element = grid%face(refinement_face)%gmsh_element
   grid%face(face2)%degree = grid%face(refinement_face)%degree
   grid%face_type(face2,:) = grid%face_type(refinement_face,:)
   grid%face(face2)%assoc_elem = child2
   grid%face(face2)%marked_edge = grid%face(face2)%edge(1)
   grid%face(face2)%child = NO_CHILD
   grid%face(face2)%next = 0
   call nullify_face_oldsoln(grid,face2)
   grid%face(face2)%next = face2

   call hash_insert(grid%face(face2)%gid,face2,grid%face_hash)

endif ! children of parent face 1 didn't exist

! define the new edge and faces from bisecting parent face 2, if they
! didn't already exist

if (new_face2) then

   refinement_face = parent_face(2)
   grid%edge(edge4)%gid = 4*grid%element(parent)%gid + 3
   grid%edge(edge4)%vertex = (/ common_vert(2),vert1 /)
   grid%edge(edge4)%bmark = grid%face(refinement_face)%bmark
   if (associated(grid%face(refinement_face)%alt_bmark)) then
      allocate(grid%edge(edge4)%alt_bmark( &
                  size(grid%face(refinement_face)%alt_bmark)),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in h_refine_element")
         stop
      endif
      grid%edge(edge4)%alt_bmark = grid%face(refinement_face)%alt_bmark
   else
      nullify(grid%edge(edge4)%alt_bmark)
   endif
   grid%edge(edge4)%tags = grid%face(refinement_face)%tags
   grid%edge(edge4)%gmsh_element = .false.
   grid%edge(edge4)%degree = grid%face(refinement_face)%degree
   grid%edge_type(edge4,:) = grid%face_type(refinement_face,:)
   grid%edge(edge4)%assoc_elem = child1
   grid%edge(edge4)%child = NO_CHILD
   call nullify_edge_oldsoln(grid,edge4)
   grid%edge(edge4)%next = edge4

   call hash_insert(grid%edge(edge4)%gid,edge4,grid%edge_hash)

! face 3

   grid%face(face3)%gid = 4*grid%face(parent_face(2))%gid

! vertices are common_vert(2), child_vert(1) and vert1 in the same order as
! in the parent face (vert1 replaces child_vert(2)) but cyclically shifted
! so that vert1 is last

   do i=1,3
      if (grid%face(parent_face(2))%vertex(i) == child_vert(2)) then
         select case(i)
         case (1)
            grid%face(face3)%vertex(1) = grid%face(parent_face(2))%vertex(2)
            grid%face(face3)%vertex(2) = grid%face(parent_face(2))%vertex(3)
            grid%face(face3)%vertex(3) = vert1
         case (2)
            grid%face(face3)%vertex(1) = grid%face(parent_face(2))%vertex(3)
            grid%face(face3)%vertex(2) = grid%face(parent_face(2))%vertex(1)
            grid%face(face3)%vertex(3) = vert1
         case (3)
            grid%face(face3)%vertex(1) = grid%face(parent_face(2))%vertex(1)
            grid%face(face3)%vertex(2) = grid%face(parent_face(2))%vertex(2)
            grid%face(face3)%vertex(3) = vert1
         end select
         exit
      endif
   end do
   grid%face(face3)%edge = (/ child_edge(2), edge1, edge4 /)
   grid%face(face3)%bmark = grid%face(refinement_face)%bmark
   if (associated(grid%face(refinement_face)%alt_bmark)) then
      allocate(grid%face(face3)%alt_bmark( &
                  size(grid%face(refinement_face)%alt_bmark)),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in h_refine_element")
         stop
      endif
      grid%face(face3)%alt_bmark = grid%face(refinement_face)%alt_bmark
   else
      nullify(grid%face(face3)%alt_bmark)
   endif
   grid%face(face3)%tags = grid%face(refinement_face)%tags
   grid%face(face3)%gmsh_element = grid%face(refinement_face)%gmsh_element
   grid%face(face3)%degree = grid%face(refinement_face)%degree
   grid%face_type(face3,:) = grid%face_type(refinement_face,:)
   grid%face(face3)%assoc_elem = child1
   grid%face(face3)%marked_edge = grid%face(face3)%edge(1)
   grid%face(face3)%child = NO_CHILD
   grid%face(face3)%next = 0
   call nullify_face_oldsoln(grid,face3)
   grid%face(face3)%next = face3

   call hash_insert(grid%face(face3)%gid,face3,grid%face_hash)

! face 4

   grid%face(face4)%gid = 4*grid%face(parent_face(2))%gid + 1

! vertices are common_vert(2), child_vert(2) and vert1 in the same order as
! in the parent face (vert1 replaces child_vert(1)) but cyclically shifted
! so that vert1 is last

   do i=1,3
      if (grid%face(parent_face(2))%vertex(i) == child_vert(1)) then
         select case(i)
         case (1)
            grid%face(face4)%vertex(1) = grid%face(parent_face(2))%vertex(2)
            grid%face(face4)%vertex(2) = grid%face(parent_face(2))%vertex(3)
            grid%face(face4)%vertex(3) = vert1
         case (2)
            grid%face(face4)%vertex(1) = grid%face(parent_face(2))%vertex(3)
            grid%face(face4)%vertex(2) = grid%face(parent_face(2))%vertex(1)
            grid%face(face4)%vertex(3) = vert1
         case (3)
            grid%face(face4)%vertex(1) = grid%face(parent_face(2))%vertex(1)
            grid%face(face4)%vertex(2) = grid%face(parent_face(2))%vertex(2)
            grid%face(face4)%vertex(3) = vert1
         end select
         exit
      endif
   end do
   grid%face(face4)%edge = (/ child_edge(4), edge2, edge4 /)
   grid%face(face4)%bmark = grid%face(refinement_face)%bmark
   if (associated(grid%face(refinement_face)%alt_bmark)) then
      allocate(grid%face(face4)%alt_bmark( &
                  size(grid%face(refinement_face)%alt_bmark)),stat=astat)
      if (astat /= 0) then
         ierr = ALLOC_FAILED
         call fatal("memory allocation failed in h_refine_element")
         stop
      endif
      grid%face(face4)%alt_bmark = grid%face(refinement_face)%alt_bmark
   else
      nullify(grid%face(face4)%alt_bmark)
   endif
   grid%face(face4)%tags = grid%face(refinement_face)%tags
   grid%face(face4)%gmsh_element = grid%face(refinement_face)%gmsh_element
   grid%face(face4)%degree = grid%face(refinement_face)%degree
   grid%face_type(face4,:) = grid%face_type(refinement_face,:)
   grid%face(face4)%assoc_elem = child2
   grid%face(face4)%marked_edge = grid%face(face4)%edge(1)
   grid%face(face4)%child = NO_CHILD
   grid%face(face4)%next = 0
   call nullify_face_oldsoln(grid,face4)
   grid%face(face4)%next = face4

   call hash_insert(grid%face(face4)%gid,face4,grid%face_hash)

endif ! children of parent face 2 didn't exist

! define the new face interior to the parent

face5 = get_next_free_face(grid,errcode)
if (errcode /= 0) return

if (new_face2) then
   grid%face(face5)%gid = 4*grid%face(parent_face(2))%gid + 2
elseif (new_face1) then
   grid%face(face5)%gid = 4*grid%face(parent_face(1))%gid + 2
else
   grid%face(face5)%gid = 4*grid%face(parent_face(2))%gid + 3
endif
grid%face(face5)%vertex = (/ common_vert(1), common_vert(2), vert1 /)
grid%face(face5)%edge = (/ common_edge, edge3, edge4 /)
if (grid%element_type(parent,1) == INTERIOR_DIRICHLET) then
   grid%face(face5)%bmark = grid%element(parent)%tags(1)
else
   grid%face(face5)%bmark = 0
endif
nullify(grid%face(face5)%alt_bmark)
grid%face(face5)%tags = grid%element(parent)%tags
grid%face(face5)%gmsh_element = .false.
grid%face(face5)%degree = grid%element(parent)%degree
if (grid%element_type(parent,1) == FLOATING) then
   grid%face_type(face5,:) = INTERIOR
else
   grid%face_type(face5,:) = grid%element_type(parent,:)
endif
grid%face(face5)%assoc_elem = child1
grid%face(face5)%child = NO_CHILD
grid%face(face5)%next = 0
! "The new face ... If the parent is type P_f, the marked edge is the edge
! connecting the new vertex to the new refinement edge.  Otherwise it is the
! edge opposite the new vertex."
select case (grid%element(parent)%type)
case ("Pf")
   if (child_edge(1) == grid%face(child_face(1))%marked_edge) then
      grid%face(face5)%marked_edge = edge3
   else
      grid%face(face5)%marked_edge = edge4
   endif
case default
   grid%face(face5)%marked_edge = grid%face(face5)%edge(1)
end select
call nullify_face_oldsoln(grid,face5)
grid%face(face5)%next = face5

call hash_insert(grid%face(face5)%gid,face5,grid%face_hash)

! define the two child elements

! first child

grid%element(child1)%gid = MAX_CHILD*grid%element(parent)%gid
grid%element(child1)%edge = (/child_edge(1), child_edge(2), common_edge, &
                              edge1, edge3, edge4/)
grid%element(child1)%face = 0
do i=1,VERTICES_PER_ELEMENT-1
   grid%element(child1)%vertex(i) = &
      grid%element(parent)%vertex(mod(pos(2)+i-1,VERTICES_PER_ELEMENT)+1)
   if (grid%element(child1)%vertex(i) == child_vert(1)) &
      grid%element(child1)%face(i) = face5
   if (grid%element(child1)%vertex(i) == common_vert(1)) &
      grid%element(child1)%face(i) = face3
   if (grid%element(child1)%vertex(i) == common_vert(2)) &
      grid%element(child1)%face(i) = face1
   if (grid%element(child1)%vertex(i) == child_vert(1)) then
      grid%element(child1)%neighbor_hint(i) = &
         MAX_CHILD*grid%element(parent)%gid+1
   else
      grid%element(child1)%neighbor_hint(i) = &
         parent_neighbors(mod(pos(2)+i-1,VERTICES_PER_ELEMENT)+1)
   endif
end do
grid%element(child1)%vertex(VERTICES_PER_ELEMENT) = vert1
grid%element(child1)%face(VERTICES_PER_ELEMENT) = child_face(1)
if (any(grid%element(child1)%face == 0)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("h_refine_element: failed to assign a face to child1")
   stop
endif
grid%element(child1)%neighbor_hint(VERTICES_PER_ELEMENT) = &
   parent_neighbors(mod(pos(2)+VERTICES_PER_ELEMENT-1,VERTICES_PER_ELEMENT)+1)
if (.not. clockwise_element(grid,child1)) then
   itemp = grid%element(child1)%vertex(1)
   grid%element(child1)%vertex(1) = grid%element(child1)%vertex(2)
   grid%element(child1)%vertex(2) = itemp
   itemp = grid%element(child1)%face(1)
   grid%element(child1)%face(1) = grid%element(child1)%face(2)
   grid%element(child1)%face(2) = itemp
   htemp = grid%element(child1)%neighbor_hint(1)
   grid%element(child1)%neighbor_hint(1) = grid%element(child1)%neighbor_hint(2)
   grid%element(child1)%neighbor_hint(2) = htemp
endif
call reorder_edges(grid%element(child1)%edge,grid%element(child1)%vertex,grid)
grid%element(child1)%level = grid%element(parent)%level+1
! set mate to BOUNDARY so that processing of the mate is always skipped
grid%element(child1)%mate = BOUNDARY
grid%element(child1)%iown = grid%element(parent)%iown
grid%element(child1)%hrefined_unowned = .false.
grid%element(child1)%prefined_unowned = .false.
grid%element(child1)%in  = grid%element(child1)%vertex(1)
grid%element(child1)%out = grid%element(child1)%vertex(2)
grid%element(child1)%next = grid%head_level_elem(grid%element(child1)%level)
grid%head_level_elem(grid%element(child1)%level) = child1
if (grid%element(child1)%next /= END_OF_LIST) then
   grid%element(grid%element(child1)%next)%previous = child1
endif
grid%element(child1)%previous = END_OF_LIST
grid%element(child1)%isleaf = .true.
grid%element(child1)%oldleaf = .false.
grid%element(child1)%tags = grid%element(parent)%tags
if (sem_simulation) then ! if odd, give extra one to child 1
   if (grid%element(child1)%tags(chargetag) < 0) then
      grid%element(child1)%tags(chargetag) = (grid%element(child1)%tags(chargetag)-1)/2
   else
      grid%element(child1)%tags(chargetag) = (grid%element(child1)%tags(chargetag)+1)/2
   endif
endif
call linked_list_create(grid%element(child1)%alt_tags)
call linked_list_copy(grid%element(parent)%alt_tags, &
                      grid%element(child1)%alt_tags)
grid%element(child1)%gmsh_element = grid%element(parent)%gmsh_element
grid%element(child1)%degree = grid%element(parent)%degree
grid%element_type(child1,:) = grid%element_type(parent,:)
grid%element(child1)%sp_eta_pred = refcont%sp_gamma_h * &
                  maxval(grid%element_errind(parent,1:max(1,grid%num_eval))) / &
                  2.0_my_real**(grid%element(parent)%degree/3.0_my_real)
grid%element(child1)%diameter = element_volume(child1,grid)**(1.0_my_real/3)
if (associated_elem_soln(grid,parent)) then
   call nullify_elem_soln(grid,child1)
   call allocate_elem_soln(grid,child1, &
                           size_basis=size_elem_soln(grid,parent,dim=BASIS_DIM))
   call set_elem_soln(grid,child1,colon,colon,colon,0.0_my_real)
   if (grid%have_true) then
      call nullify_elem_exact(grid,child1)
      call allocate_elem_exact(grid,child1, &
                          size_basis=size_elem_exact(grid,parent,dim=BASIS_DIM))
! Can't set elem_exact until exact is set for the edges and faces
   else
      call nullify_elem_exact(grid,child1)
   endif
else
   call nullify_elem_soln(grid,child1)
   call nullify_elem_exact(grid,child1)
endif
call nullify_elem_oldsoln(grid,child1)
if (grid%element(child1)%degree >= 4) then
      grid%dof = grid%dof + grid%system_size * &
                            element_dof(grid%element(child1)%degree)
endif
if (grid%element(child1)%iown .and. grid%element(child1)%degree >= 4) then
      grid%dof_own = grid%dof_own + grid%system_size * &
                                    element_dof(grid%element(child1)%degree)
endif
call hash_insert(grid%element(child1)%gid,child1,grid%elem_hash)
select case (grid%element(parent)%type)
case ("Pu")
   grid%element(child1)%type = "Pf"
case ("A")
   grid%element(child1)%type = "Pu"
case ("M")
   grid%element(child1)%type = "Pu"
case ("O")
   grid%element(child1)%type = "Pu"
case ("Pf")
   grid%element(child1)%type = "A"
end select
grid%element(child1)%refinement_edge = grid%face(child_face(1))%marked_edge
grid%element_errind(child1,:) = -1.0_my_real

! second child

grid%element(child2)%gid = MAX_CHILD*grid%element(parent)%gid + 1
grid%element(child2)%edge = (/child_edge(3), child_edge(4), common_edge, &
                              edge2, edge3, edge4/)
grid%element(child2)%face = 0
do i=1,VERTICES_PER_ELEMENT-1
   grid%element(child2)%vertex(i) = &
      grid%element(parent)%vertex(mod(pos(1)+i-1,VERTICES_PER_ELEMENT)+1)
   if (grid%element(child2)%vertex(i) == child_vert(2)) &
      grid%element(child2)%face(i) = face5
   if (grid%element(child2)%vertex(i) == common_vert(1)) &
      grid%element(child2)%face(i) = face4
   if (grid%element(child2)%vertex(i) == common_vert(2)) &
      grid%element(child2)%face(i) = face2
   if (grid%element(child2)%vertex(i) == child_vert(2)) then
      grid%element(child2)%neighbor_hint(i) = &
         MAX_CHILD*grid%element(parent)%gid
   else
      grid%element(child2)%neighbor_hint(i) = &
         parent_neighbors(mod(pos(1)+i-1,VERTICES_PER_ELEMENT)+1)
   endif
end do
grid%element(child2)%vertex(VERTICES_PER_ELEMENT) = vert1
grid%element(child2)%face(VERTICES_PER_ELEMENT) = child_face(2)
if (any(grid%element(child2)%face == 0)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("h_refine_element: failed to assign a face to child2")
   stop
endif
grid%element(child2)%neighbor_hint(VERTICES_PER_ELEMENT) = &
   parent_neighbors(mod(pos(1)+VERTICES_PER_ELEMENT-1,VERTICES_PER_ELEMENT)+1)
if (.not. clockwise_element(grid,child2)) then
   itemp = grid%element(child2)%vertex(1)
   grid%element(child2)%vertex(1) = grid%element(child2)%vertex(2)
   grid%element(child2)%vertex(2) = itemp
   itemp = grid%element(child2)%face(1)
   grid%element(child2)%face(1) = grid%element(child2)%face(2)
   grid%element(child2)%face(2) = itemp
   htemp = grid%element(child2)%neighbor_hint(1)
   grid%element(child2)%neighbor_hint(1) = grid%element(child2)%neighbor_hint(2)
   grid%element(child2)%neighbor_hint(2) = htemp
endif
call reorder_edges(grid%element(child2)%edge,grid%element(child2)%vertex,grid)
grid%element(child2)%level = grid%element(parent)%level+1
grid%element(child2)%mate = BOUNDARY
grid%element(child2)%iown = grid%element(parent)%iown
grid%element(child2)%hrefined_unowned = .false.
grid%element(child2)%prefined_unowned = .false.
grid%element(child2)%in  = grid%element(child2)%vertex(1)
grid%element(child2)%out = grid%element(child2)%vertex(2)
grid%element(child2)%next = grid%head_level_elem(grid%element(child2)%level)
grid%head_level_elem(grid%element(child2)%level) = child2
if (grid%element(child2)%next /= END_OF_LIST) then
   grid%element(grid%element(child2)%next)%previous = child2
endif
grid%element(child2)%previous = END_OF_LIST
grid%element(child2)%isleaf = .true.
grid%element(child2)%oldleaf = .false.
grid%element(child2)%tags = grid%element(parent)%tags
if (sem_simulation) then
   grid%element(child2)%tags(chargetag) = grid%element(child2)%tags(chargetag)/2
endif
call linked_list_create(grid%element(child2)%alt_tags)
call linked_list_copy(grid%element(parent)%alt_tags, &
                      grid%element(child2)%alt_tags)
grid%element(child2)%gmsh_element = grid%element(parent)%gmsh_element
grid%element(child2)%degree = grid%element(parent)%degree
grid%element_type(child2,:) = grid%element_type(parent,:)
grid%element(child2)%sp_eta_pred = grid%element(child1)%sp_eta_pred
grid%element(child2)%diameter = element_volume(child2,grid)**(1.0_my_real/3)
if (associated_elem_soln(grid,parent)) then
   call nullify_elem_soln(grid,child2)
   call allocate_elem_soln(grid,child2, &
                           size_basis=size_elem_soln(grid,parent,dim=BASIS_DIM))
   call set_elem_soln(grid,child2,colon,colon,colon,0.0_my_real)
   if (grid%have_true) then
      call nullify_elem_exact(grid,child2)
      call allocate_elem_exact(grid,child2, &
                          size_basis=size_elem_exact(grid,parent,dim=BASIS_DIM))
! Can't set element exact until exact has been set for the edges and faces
   else
      call nullify_elem_exact(grid,child2)
   endif
else
   call nullify_elem_soln(grid,child2)
   call nullify_elem_exact(grid,child2)
endif
call nullify_elem_oldsoln(grid,child2)
if (grid%element(child2)%degree >= 4) then
      grid%dof = grid%dof + grid%system_size * &
                            element_dof(grid%element(child2)%degree)
endif
if (grid%element(child2)%iown .and. grid%element(child2)%degree >= 4) then
      grid%dof_own = grid%dof_own + grid%system_size * &
                                    element_dof(grid%element(child2)%degree)
endif
call hash_insert(grid%element(child2)%gid,child2,grid%elem_hash)
select case (grid%element(parent)%type)
case ("Pu")
   grid%element(child2)%type = "Pf"
case ("A")
   grid%element(child2)%type = "Pu"
case ("M")
   grid%element(child2)%type = "Pu"
case ("O")
   grid%element(child2)%type = "Pu"
case ("Pf")
   grid%element(child2)%type = "A"
end select
grid%element(child2)%refinement_edge = grid%face(child_face(2))%marked_edge
grid%element_errind(child2,:) = -1.0_my_real

! for SEM with multiple point charges, redo the charge tag by splitting the
! charges according to which child contains each one

if (sem_simulation .and. sem_point_charges) then
   if (grid%element(parent)%tags(chargetag) /= 0) then
      grid%element(child1)%tags(chargetag) = 0
      grid%element(child2)%tags(chargetag) = 0
      i = grid%element(parent)%tags(chargetag)
      do while (i /= -1)
         call find_containing_element(sem_point_charge_x(i), &
                              sem_point_charge_y(i),have_it,elem,grid,parent, &
                              sem_point_charge_z(i))
         if (have_it == 0) then
            ierr = PHAML_INTERNAL_ERROR
            call fatal("h_refine_element: didn't find a child containing a point charge")
            stop
         endif
         if (grid%element(elem)%tags(chargetag) == 0) then
            grid%element(elem)%tags(chargetag) = i
         else
            j = grid%element(elem)%tags(chargetag)
            do while (sem_next_charge(j) /= -1)
              j = sem_next_charge(j)
            end do
            sem_next_charge(j) = i
         endif
         itemp = i
         i = sem_next_charge(i)
         sem_next_charge(itemp) = -1
      end do
   endif
endif

! the parent is no longer a leaf

grid%element(parent)%isleaf = .false.

if (grid%element(parent)%degree >= 4) then
      grid%dof = grid%dof - grid%system_size * &
                            element_dof(grid%element(parent)%degree)
endif
if (grid%element(parent)%iown .and. grid%element(parent)%degree >= 4) then
      grid%dof_own = grid%dof_own - grid%system_size * &
                                    element_dof(grid%element(parent)%degree)
endif

grid%element(parent)%order = (/1,2/)

! set Dirichlet boundary conditions and exact solution on new edges and
! faces.  This can't be done until after the child elements are defined
! because edge_exact and face_exact need an element containing the edge/face.

if (new_ref_edge) then

   if (grid%edge(edge1)%degree > 1) then
      call nullify_edge_soln(grid,edge1)
      call allocate_edge_soln(grid,edge1,size_basis=grid%edge(edge1)%degree-1)
      if (grid%have_true) then
         call nullify_edge_exact(grid,edge1)
         call allocate_edge_exact(grid,edge1, &
                                  size_basis=grid%edge(edge1)%degree-1)
      else
         call nullify_edge_exact(grid,edge1)
      endif
      do i=1,grid%system_size
         if (grid%edge_type(edge1,i) == DIRICHLET .or. &
             grid%edge_type(edge1,i) == INTERIOR_DIRICHLET .or. &
             grid%edge_type(edge1,i) == PERIODIC_MASTER_DIR .or. &
             grid%edge_type(edge1,i) == PERIODIC_SLAVE_DIR) then
            call edge_exact(grid,edge1,i,"d")
         else
            call set_edge_soln(grid,edge1,colon,i,colon,0.0_my_real)
         endif
         if (grid%have_true) call edge_exact(grid,edge1,i,"t")
      end do
   else
      call nullify_edge_soln(grid,edge1)
      call nullify_edge_exact(grid,edge1)
   endif

   if (grid%edge(edge2)%degree > 1) then
      call nullify_edge_soln(grid,edge2)
      call allocate_edge_soln(grid,edge2,size_basis=grid%edge(edge2)%degree-1)
      if (grid%have_true) then
         call nullify_edge_exact(grid,edge2)
         call allocate_edge_exact(grid,edge2, &
                                  size_basis=grid%edge(edge2)%degree-1)
      else
         call nullify_edge_exact(grid,edge2)
      endif
      do i=1,grid%system_size
         if (grid%edge_type(edge2,i) == DIRICHLET .or. &
             grid%edge_type(edge2,i) == INTERIOR_DIRICHLET .or. &
             grid%edge_type(edge2,i) == PERIODIC_MASTER_DIR .or. &
             grid%edge_type(edge2,i) == PERIODIC_SLAVE_DIR) then
            call edge_exact(grid,edge2,i,"d")
         else
            call set_edge_soln(grid,edge2,colon,i,colon,0.0_my_real)
         endif
         if (grid%have_true) call edge_exact(grid,edge2,i,"t")
      end do
   else
      call nullify_edge_soln(grid,edge2)
      call nullify_edge_exact(grid,edge2)
   endif

endif

if (new_face1) then

   if (grid%edge(edge3)%degree > 1) then
      call nullify_edge_soln(grid,edge3)
      call allocate_edge_soln(grid,edge3,size_basis=grid%edge(edge3)%degree-1)
      if (grid%have_true) then
         call nullify_edge_exact(grid,edge3)
         call allocate_edge_exact(grid,edge3, &
                                  size_basis=grid%edge(edge3)%degree-1)
      else
         call nullify_edge_exact(grid,edge3)
      endif
      do i=1,grid%system_size
         if (grid%edge_type(edge3,i) == DIRICHLET .or. &
             grid%edge_type(edge3,i) == INTERIOR_DIRICHLET) then
            call edge_exact(grid,edge3,i,"d")
         else
            call set_edge_soln(grid,edge3,colon,i,colon,0.0_my_real)
         endif
         if (grid%have_true) call edge_exact(grid,edge3,i,"t")
      end do
   else
      call nullify_edge_soln(grid,edge3)
      call nullify_edge_exact(grid,edge3)
   endif

   if (grid%face(face1)%degree > 2) then
      call nullify_face_soln(grid,face1)
      call allocate_face_soln(grid,face1, &
                              size_basis=face_dof(grid%face(face1)%degree))
      if (grid%have_true) then
         call nullify_face_exact(grid,face1)
         call allocate_face_exact(grid,face1, &
                                  size_basis=face_dof(grid%face(face1)%degree))
      else
         call nullify_face_exact(grid,face1)
      endif
      do i=1,grid%system_size
         if (grid%face_type(face1,i) == DIRICHLET .or. &
             grid%face_type(face1,i) == INTERIOR_DIRICHLET) then
            call face_exact(grid,face1,i,"d")
         else
            call set_face_soln(grid,face1,colon,i,colon,0.0_my_real)
         endif
         if (grid%have_true) call face_exact(grid,face1,i,"t")
      end do
   else
      call nullify_face_soln(grid,face1)
      call nullify_face_exact(grid,face1)
   endif

   if (grid%face(face2)%degree > 2) then
      call nullify_face_soln(grid,face2)
      call allocate_face_soln(grid,face2, &
                              size_basis=face_dof(grid%face(face2)%degree))
      if (grid%have_true) then
         call nullify_face_exact(grid,face2)
         call allocate_face_exact(grid,face2, &
                                  size_basis=face_dof(grid%face(face2)%degree))
      else
         call nullify_face_exact(grid,face2)
      endif
      do i=1,grid%system_size
         if (grid%face_type(face2,i) == DIRICHLET .or. &
             grid%face_type(face2,i) == INTERIOR_DIRICHLET) then
            call face_exact(grid,face2,i,"d")
         else
            call set_face_soln(grid,face2,colon,i,colon,0.0_my_real)
         endif
         if (grid%have_true) call face_exact(grid,face2,i,"t")
      end do
   else
      call nullify_face_soln(grid,face2)
      call nullify_face_exact(grid,face2)
   endif

endif

if (new_face2) then

   if (grid%edge(edge4)%degree > 1) then
      call nullify_edge_soln(grid,edge4)
      call allocate_edge_soln(grid,edge4,size_basis=grid%edge(edge4)%degree-1)
      if (grid%have_true) then
         call nullify_edge_exact(grid,edge4)
         call allocate_edge_exact(grid,edge4, &
                                  size_basis=grid%edge(edge4)%degree-1)
      else
         call nullify_edge_exact(grid,edge4)
      endif
      do i=1,grid%system_size
         if (grid%edge_type(edge4,i) == DIRICHLET .or. &
             grid%face_type(edge4,i) == INTERIOR_DIRICHLET) then
            call edge_exact(grid,edge4,i,"d")
         else
            call set_edge_soln(grid,edge4,colon,i,colon,0.0_my_real)
         endif
         if (grid%have_true) call edge_exact(grid,edge4,i,"t")
      end do
   else
      call nullify_edge_soln(grid,edge4)
      call nullify_edge_exact(grid,edge4)
   endif

   if (grid%face(face3)%degree > 2) then
      call nullify_face_soln(grid,face3)
      call allocate_face_soln(grid,face3, &
                              size_basis=face_dof(grid%face(face3)%degree))
      if (grid%have_true) then
         call nullify_face_exact(grid,face3)
         call allocate_face_exact(grid,face3, &
                                  size_basis=face_dof(grid%face(face3)%degree))
      else
         call nullify_face_exact(grid,face3)
      endif
      do i=1,grid%system_size
         if (grid%face_type(face3,i) == DIRICHLET .or. &
             grid%face_type(face3,i) == INTERIOR_DIRICHLET) then
            call face_exact(grid,face3,i,"d")
         else
            call set_face_soln(grid,face3,colon,i,colon,0.0_my_real)
         endif
         if (grid%have_true) call face_exact(grid,face3,i,"t")
      end do
   else
      call nullify_face_soln(grid,face3)
      call nullify_face_exact(grid,face3)
   endif

   if (grid%face(face4)%degree > 2) then
      call nullify_face_soln(grid,face4)
      call allocate_face_soln(grid,face4, &
                              size_basis=face_dof(grid%face(face4)%degree))
      if (grid%have_true) then
         call nullify_face_exact(grid,face4)
         call allocate_face_exact(grid,face4, &
                                  size_basis=face_dof(grid%face(face4)%degree))
      else
         call nullify_face_exact(grid,face4)
      endif
      do i=1,grid%system_size
         if (grid%face_type(face4,i) == DIRICHLET .or. &
             grid%face_type(face4,i) == INTERIOR_DIRICHLET) then
            call face_exact(grid,face4,i,"d")
         else
            call set_face_soln(grid,face4,colon,i,colon,0.0_my_real)
         endif
         if (grid%have_true) call face_exact(grid,face4,i,"t")
      end do
   else
      call nullify_face_soln(grid,face4)
      call nullify_face_exact(grid,face4)
   endif

endif

if (grid%face(face5)%degree > 2) then
   call nullify_face_soln(grid,face5)
   call allocate_face_soln(grid,face5, &
                           size_basis=face_dof(grid%face(face5)%degree))
   if (grid%have_true) then
      call nullify_face_exact(grid,face5)
      call allocate_face_exact(grid,face5, &
                               size_basis=face_dof(grid%face(face5)%degree))
   else
      call nullify_face_exact(grid,face5)
   endif
   do i=1,grid%system_size
      if (grid%face_type(face5,i) == DIRICHLET .or. &
          grid%face_type(face5,i) == INTERIOR_DIRICHLET) then
         call face_exact(grid,face5,i,"d")
      else
         call set_face_soln(grid,face5,colon,i,colon,0.0_my_real)
      endif
      if (grid%have_true) call face_exact(grid,face5,i,"t")
   end do
else
   call nullify_face_soln(grid,face5)
   call nullify_face_exact(grid,face5)
endif

! Couldn't set the element exact until after the edge and face exacts were set

if (associated_elem_exact(grid,child1)) then
   do i=1,grid%system_size
      call elem_exact(grid,child1,i,"t")
   end do
endif
if (associated_elem_exact(grid,child1)) then
   do i=1,grid%system_size
      call elem_exact(grid,child2,i,"t")
   end do
endif

! if parent face 1 is periodic, then set the "next" pointers if the matching
! face has already been refined, and likewise for parent face 2

do_face1 = is_periodic_face(parent_face(1),grid)
if (do_face1) then
   do_face1 = grid%face(grid%face(parent_face(1))%next)%child(1) /= NO_CHILD
endif
do_face2 = is_periodic_face(parent_face(2),grid)
if (do_face2) then
   do_face2 = grid%face(grid%face(parent_face(2))%next)%child(1) /= NO_CHILD
endif

if (do_face1 .or. do_face2) then

   if (grid%edge(edge3)%degree > 1) then
      call nullify_edge_soln(grid,edge3)
      call allocate_edge_soln(grid,edge3,size_basis=grid%edge(edge3)%degree-1)
      if (grid%have_true) then
         call nullify_edge_exact(grid,edge3)
         call allocate_edge_exact(grid,edge3, &
                                  size_basis=grid%edge(edge3)%degree-1)
         call set_edge_exact(grid,edge3,colon,colon,colon,0.0_my_real)
      else
         call nullify_edge_exact(grid,edge3)
      endif
      do i=1,grid%system_size
         if (grid%edge_type(edge3,i) == DIRICHLET .or. &
             grid%face_type(edge3,i) == INTERIOR_DIRICHLET) then
            call edge_exact(grid,edge3,i,"d")
         else
            call set_edge_soln(grid,edge3,colon,i,colon,0.0_my_real)
         endif
         if (grid%have_true) call edge_exact(grid,edge3,i,"t")
      end do
   else
      call nullify_edge_soln(grid,edge3)
      call nullify_edge_exact(grid,edge3)
   endif

! The matching faces are the children of the periodic mate of the face that was
! refined.  The child that matches face1 is the one that contains the matching
! edge for face1's first edge, etc.  Note that the first edges of face1 and
! face2 can't both be doubly periodic because they share a common point, so
! one of them should have the correct mate as the first "next".

   if (do_face1) then
      mate_face_1 = grid%face(grid%face(parent_face(1))%next)%child(1)
      mate_face_2 = grid%face(grid%face(parent_face(1))%next)%child(2)
      if (any(grid%face(mate_face_2)%edge == &
              grid%edge(grid%face(face1)%edge(1))%next) .or. &
          any(grid%face(mate_face_1)%edge == &
              grid%edge(grid%face(face2)%edge(1))%next)) then
         itemp = mate_face_1
         mate_face_1 = mate_face_2
         mate_face_2 = itemp
      elseif (all(grid%face(mate_face_1)%edge /= &
                  grid%edge(grid%face(face1)%edge(1))%next) .and. &
              all(grid%face(mate_face_2)%edge /= &
                  grid%edge(grid%face(face2)%edge(1))%next)) then
         ierr = PHAML_INTERNAL_ERROR
         call fatal("couldn't find periodic matching edge in matching face")
         stop
      endif
   endif
   if (do_face2) then
      mate_face_3 = grid%face(grid%face(parent_face(2))%next)%child(1)
      mate_face_4 = grid%face(grid%face(parent_face(2))%next)%child(2)
      if (any(grid%face(mate_face_4)%edge == &
              grid%edge(grid%face(face3)%edge(1))%next) .or. &
          any(grid%face(mate_face_3)%edge == &
              grid%edge(grid%face(face4)%edge(1))%next)) then
         itemp = mate_face_3
         mate_face_3 = mate_face_4
         mate_face_4 = itemp
      elseif (all(grid%face(mate_face_3)%edge /= &
                  grid%edge(grid%face(face3)%edge(1))%next) .and. &
              all(grid%face(mate_face_4)%edge /= &
                  grid%edge(grid%face(face4)%edge(1))%next)) then
         ierr = PHAML_INTERNAL_ERROR
         call fatal("couldn't find periodic matching edge in matching face")
         stop
      endif
   endif

! The matching vertex of the new vertex is the third vertex of any matching
! face.

   if (do_face1) then
      mate_vert_1 = grid%face(mate_face_1)%vertex(3)
   else
      mate_vert_1 = grid%face(mate_face_3)%vertex(3)
   endif

! The matching edge of edge1 and edge2 is the second edge of the matching face.

   if (do_face1) then
      mate_edge_1 = grid%face(mate_face_1)%edge(2)
      mate_edge_2 = grid%face(mate_face_2)%edge(2)
   else
      mate_edge_1 = grid%face(mate_face_3)%edge(2)
      mate_edge_2 = grid%face(mate_face_4)%edge(2)
   endif

! For the edge that cuts the face the matching edge is the 3rd edge
! of the matching face.

   if (do_face1) then
      mate_edge_3 = grid%face(mate_face_1)%edge(3)
   endif
   if (do_face2) then
      mate_edge_4 = grid%face(mate_face_3)%edge(3)
   endif

! The faces simply point at each other; it doesn't matter which is master.
! The edges that split faces cannot be doubly periodic, so they just point at 
! at each other.

   if (do_face1) then
      grid%face(face1)%next = mate_face_1
      grid%face(mate_face_1)%next = face1
      grid%face(face2)%next = mate_face_2
      grid%face(mate_face_2)%next = face2
      grid%edge(edge3)%next = mate_edge_3
      grid%edge(mate_edge_3)%next = edge3
   endif
   if (do_face2) then
      grid%face(face3)%next = mate_face_3
      grid%face(mate_face_3)%next = face3
      grid%face(face4)%next = mate_face_4
      grid%face(mate_face_4)%next = face4
      grid%edge(edge4)%next = mate_edge_4
      grid%edge(mate_edge_4)%next = edge4
   endif

! three possibilities for the edges:
! 1) mate_edge_1 is already in a cycle of periodic edges which does not
!    contain edge1 (indicating it is doubly periodic)
! 2) mate_edge_1 and edge1 are already in a cycle together (previously assigned)
! 3) mate_edge_1 is not in a cycle; if we start at mate_edge_1 and don't come
!    back to it within 5 nexts, it is not in a cycle
! mate_edge_2 must be in the same state as mate_edge_1

   mate_edge_1_in_cycle = .false.
   edge1_in_cycle = .false.
   itemp = mate_edge_1
   do i=1,5
      itemp = grid%edge(itemp)%next
      if (itemp == END_OF_LIST) exit
      if (itemp == edge1) edge1_in_cycle = .true.
      if (itemp == mate_edge_1) then
         mate_edge_1_in_cycle = .true.
         exit
      endif
   end do
   edge1_in_cycle = edge1_in_cycle .and. mate_edge_1_in_cycle

! it doesn't matter where the master is in the cycle.  If mate_edge_1
! is already in a cycle and edge1 is not, insert edge1 after mate_edge_1.  If
! it is not, make them point to each other.

   if (.not. edge1_in_cycle) then
      if (mate_edge_1_in_cycle) then
         grid%edge(edge1)%next = grid%edge(mate_edge_1)%next
         grid%edge(mate_edge_1)%next = edge1
         grid%edge(edge2)%next = grid%edge(mate_edge_2)%next
         grid%edge(mate_edge_2)%next = edge2
      else
         grid%edge(edge1)%next = mate_edge_1
         grid%edge(mate_edge_1)%next = edge1
         grid%edge(edge2)%next = mate_edge_2
         grid%edge(mate_edge_2)%next = edge2
      endif
   endif

! if both face1 and face2 are periodic, we have only done the edges for the
! first one; do it again with the matching edge of the second face, but now
! the insertion is a little different

   if (do_face1 .and. do_face2) then

      mate_edge_1 = grid%face(mate_face_3)%edge(2)
      mate_edge_2 = grid%face(mate_face_4)%edge(2)

      mate_edge_1_in_cycle = .false.
      edge1_in_cycle = .false.
      itemp = mate_edge_1
      do i=1,5
         itemp = grid%edge(itemp)%next
         if (itemp == END_OF_LIST) exit
         if (itemp == edge1) edge1_in_cycle = .true.
         if (itemp == mate_edge_1) then
            mate_edge_1_in_cycle = .true.
            exit
         endif
      end do
      edge1_in_cycle = edge1_in_cycle .and. mate_edge_1_in_cycle

      if (.not. edge1_in_cycle) then
         if (mate_edge_1_in_cycle) then
            itemp = grid%edge(edge1)%next
            grid%edge(edge1)%next = grid%edge(mate_edge_1)%next
            grid%edge(mate_edge_1)%next = itemp
            itemp = grid%edge(edge2)%next
            grid%edge(edge2)%next = grid%edge(mate_edge_2)%next
            grid%edge(mate_edge_2)%next = itemp
         else
            grid%edge(mate_edge_1)%next = grid%edge(mate_edge_1)%next
            grid%edge(edge1)%next = mate_edge_1
            grid%edge(mate_edge_2)%next = grid%edge(mate_edge_2)%next
            grid%edge(edge2)%next = mate_edge_2
         endif
      endif

   endif

! same for vertices, except check 9 for a cycle

   mate_vert_1_in_cycle = .false.
   vert1_in_cycle = .false.
   itemp = mate_vert_1
   do i=1,9
      itemp = grid%vertex(itemp)%next
      if (itemp == END_OF_LIST) exit
      if (itemp == vert1) vert1_in_cycle = .true.
      if (itemp == mate_vert_1) then
         mate_vert_1_in_cycle = .true.
         exit
      endif
   end do
   vert1_in_cycle = vert1_in_cycle .and. mate_vert_1_in_cycle

   if (.not. vert1_in_cycle) then
      if (mate_vert_1_in_cycle) then
         grid%vertex(vert1)%next = grid%vertex(mate_vert_1)%next
         grid%vertex(mate_vert_1)%next = vert1
      else
         grid%vertex(vert1)%next = mate_vert_1
         grid%vertex(mate_vert_1)%next = vert1
      endif
   endif

   if (do_face1 .and. do_face2) then

      mate_vert_1 = grid%face(mate_face_3)%vertex(3)

      mate_vert_1_in_cycle = .false.
      vert1_in_cycle = .false.
      itemp = mate_vert_1
      do i=1,9
         itemp = grid%vertex(itemp)%next
         if (itemp == END_OF_LIST) exit
         if (itemp == vert1) vert1_in_cycle = .true.
         if (itemp == mate_vert_1) then
            mate_vert_1_in_cycle = .true.
            exit
         endif
      end do
      vert1_in_cycle = vert1_in_cycle .and. mate_vert_1_in_cycle

      if (.not. vert1_in_cycle) then
         if (mate_vert_1_in_cycle) then
            itemp = grid%vertex(vert1)%next
            grid%vertex(vert1)%next = grid%vertex(mate_vert_1)%next
            grid%vertex(mate_vert_1)%next = itemp
         else
            grid%vertex(mate_vert_1)%next = grid%vertex(mate_vert_1)%next
            grid%vertex(vert1)%next = mate_vert_1
         endif
      endif

   endif

endif ! do_face1 or do_face2

! associated element for the vertices of the parent

! make this OpenMP critical just to be sure another thread doesn't try to
! read it while this thread is writing it

!$omp critical (reset_assoc_elem_critical)

if (grid%vertex(child_vert(1))%assoc_elem == parent) then
   grid%vertex(child_vert(1))%assoc_elem = child1
endif
if (grid%vertex(child_vert(2))%assoc_elem == parent) then
   grid%vertex(child_vert(2))%assoc_elem = child2
endif
if (grid%vertex(common_vert(1))%assoc_elem == parent) then
   grid%vertex(common_vert(1))%assoc_elem = child1
endif
if (grid%vertex(common_vert(2))%assoc_elem == parent) then
   grid%vertex(common_vert(2))%assoc_elem = child1
endif

! associated elements of the edges of the parent

if (grid%edge(child_edge(1))%assoc_elem == parent) then
   grid%edge(child_edge(1))%assoc_elem = child1
endif
if (grid%edge(child_edge(2))%assoc_elem == parent) then
   grid%edge(child_edge(2))%assoc_elem = child1
endif
if (grid%edge(child_edge(3))%assoc_elem == parent) then
   grid%edge(child_edge(3))%assoc_elem = child2
endif
if (grid%edge(child_edge(4))%assoc_elem == parent) then
   grid%edge(child_edge(4))%assoc_elem = child2
endif
if (grid%edge(common_edge)%assoc_elem == parent) then
   grid%edge(common_edge)%assoc_elem = child1
endif

! associated elements of the faces of the parent

if (grid%face(child_face(1))%assoc_elem == parent) then
   grid%face(child_face(1))%assoc_elem = child1
endif
if (grid%face(child_face(2))%assoc_elem == parent) then
   grid%face(child_face(2))%assoc_elem = child2
endif

!$omp end critical (reset_assoc_elem_critical)

! children of the bisected edge and faces

if (new_ref_edge) then
   grid%edge(refinement_edge)%child = (/edge1,edge2/)
endif
if (new_face1) then
   grid%face(parent_face(1))%child = (/face1,face2/)
endif
if (new_face2) then
   grid%face(parent_face(2))%child = (/face3,face4/)
endif

! grid scalars

if (new_ref_edge) then

   grid%nvert = grid%nvert + 1
   if (grid%element(grid%vertex(vert1)%assoc_elem)%iown) then
      grid%nvert_own = grid%nvert_own + 1
   endif
   grid%nedge = grid%nedge + 2
   grid%nedge_leaf = grid%nedge_leaf + 1
   if (grid%element(grid%edge(edge1)%assoc_elem)%iown) then
      grid%nedge_leaf_own = grid%nedge_leaf_own + 1
   endif

   if (grid%any_periodic) then
      do comp=1,grid%system_size
         if (.not. is_periodic_vert_slave(vert1,grid,comp)) then
            grid%dof = grid%dof + 1
            if (grid%element(grid%vertex(vert1)%assoc_elem)%iown) &
               grid%dof_own = grid%dof_own + 1
         endif
      end do
   else
      grid%dof = grid%dof + grid%system_size
      if (grid%element(grid%vertex(vert1)%assoc_elem)%iown) &
         grid%dof_own = grid%dof_own + grid%system_size
   endif

   if (grid%edge(edge1)%degree >= 2) then
      if (grid%any_periodic) then
         inc = grid%edge(edge1)%degree - 1
         if (inc /= 0) then
            do comp=1,grid%system_size
               if (.not. is_periodic_edge_slave(edge1,grid,comp)) then
                  grid%dof = grid%dof + inc
                  if (grid%element(grid%edge(edge1)%assoc_elem)%iown) &
                     grid%dof_own = grid%dof_own + inc
               endif
            end do
         endif
      else
         grid%dof = grid%dof + grid%system_size*(grid%edge(edge1)%degree - 1)
         if (grid%element(grid%edge(edge1)%assoc_elem)%iown) &
            grid%dof_own = grid%dof_own + &
                           grid%system_size*(grid%edge(edge1)%degree - 1)
      endif
   endif

   if (grid%edge(edge2)%degree >= 2) then
      if (grid%any_periodic) then
         inc = grid%edge(edge2)%degree - 1
         if (inc /= 0) then
            do comp=1,grid%system_size
               if (.not. is_periodic_edge_slave(edge2,grid,comp)) then
                  grid%dof = grid%dof + inc
                  if (grid%element(grid%edge(edge2)%assoc_elem)%iown) &
                     grid%dof_own = grid%dof_own + inc
               endif
            end do
         endif
      else
         grid%dof = grid%dof + grid%system_size*(grid%edge(edge2)%degree - 1)
         if (grid%element(grid%edge(edge2)%assoc_elem)%iown) &
            grid%dof_own = grid%dof_own + &
                           grid%system_size*(grid%edge(edge2)%degree - 1)
      endif
   endif

   if (grid%edge(refinement_edge)%degree >= 2) then
      if (grid%any_periodic) then
         inc = grid%edge(refinement_edge)%degree - 1
         if (inc /= 0) then
            do comp=1,grid%system_size
               if (.not. is_periodic_edge_slave(refinement_edge,grid,comp)) then
                  grid%dof = grid%dof - inc
                  if (grid%element(grid%edge(refinement_edge)%assoc_elem)%iown)&
                     grid%dof_own = grid%dof_own - inc
               endif
            end do
         endif
      else
         grid%dof = grid%dof - grid%system_size*(grid%edge(refinement_edge)%degree-1)
         if (grid%element(grid%edge(refinement_edge)%assoc_elem)%iown) &
            grid%dof_own = grid%dof_own - &
                          grid%system_size*(grid%edge(refinement_edge)%degree-1)
      endif
   endif
endif

if (new_face1) then

   grid%nedge = grid%nedge + 1
   grid%nedge_leaf = grid%nedge_leaf + 1
   if (grid%element(grid%edge(edge3)%assoc_elem)%iown) then
      grid%nedge_leaf_own = grid%nedge_leaf_own + 1
   endif
   grid%nface = grid%nface + 2
   grid%nface_leaf = grid%nface_leaf + 1
   if (grid%element(grid%face(face1)%assoc_elem)%iown) then
      grid%nface_leaf_own = grid%nface_leaf_own + 1
   endif

   if (grid%edge(edge3)%degree >= 2) then
      if (grid%any_periodic) then
         inc = grid%edge(edge3)%degree - 1
         if (inc /= 0) then
            do comp=1,grid%system_size
               if (.not. is_periodic_edge_slave(edge3,grid,comp)) then
                  grid%dof = grid%dof + inc
                  if (grid%element(grid%edge(edge3)%assoc_elem)%iown) &
                     grid%dof_own = grid%dof_own + inc
               endif
            end do
         endif
      else
         grid%dof = grid%dof + grid%system_size*(grid%edge(edge3)%degree - 1)
         if (grid%element(grid%edge(edge3)%assoc_elem)%iown) &
            grid%dof_own = grid%dof_own + &
                           grid%system_size*(grid%edge(edge3)%degree - 1)
      endif
   endif

   if (grid%face(face1)%degree >= 3) then
      if (grid%any_periodic) then
         inc = face_dof(grid%face(face1)%degree)
         if (inc /= 0) then
            do comp=1,grid%system_size
               if (.not. is_periodic_face_slave(face1,grid,comp)) then
                  grid%dof = grid%dof + inc
                  if (grid%element(grid%face(face1)%assoc_elem)%iown) &
                     grid%dof_own = grid%dof_own + inc
               endif
            end do
         endif
      else
         grid%dof = grid%dof+grid%system_size*face_dof(grid%face(face1)%degree)
         if (grid%element(grid%face(face1)%assoc_elem)%iown) &
            grid%dof_own = grid%dof_own + &
                           grid%system_size*face_dof(grid%face(face1)%degree)
      endif
   endif

   if (grid%face(face2)%degree >= 3) then
      if (grid%any_periodic) then
         inc = face_dof(grid%face(face2)%degree)
         if (inc /= 0) then
            do comp=1,grid%system_size
               if (.not. is_periodic_face_slave(face2,grid,comp)) then
                  grid%dof = grid%dof + inc
                  if (grid%element(grid%face(face2)%assoc_elem)%iown) &
                     grid%dof_own = grid%dof_own + inc
               endif
            end do
         endif
      else
         grid%dof = grid%dof+grid%system_size*face_dof(grid%face(face2)%degree)
         if (grid%element(grid%face(face2)%assoc_elem)%iown) &
            grid%dof_own = grid%dof_own + &
                           grid%system_size*face_dof(grid%face(face2)%degree)
      endif
   endif

   if (grid%face(parent_face(1))%degree >= 3) then
      if (grid%any_periodic) then
         inc = face_dof(grid%face(parent_face(1))%degree)
         if (inc /= 0) then
            do comp=1,grid%system_size
               if (.not. is_periodic_face_slave(parent_face(1),grid,comp)) then
                  grid%dof = grid%dof - inc
                  if (grid%element(grid%face(parent_face(1))%assoc_elem)%iown) &
                     grid%dof_own = grid%dof_own - inc
               endif
            end do
         endif
      else
         grid%dof = grid%dof - grid%system_size*face_dof(grid%face(parent_face(1))%degree)
         if (grid%element(grid%face(parent_face(1))%assoc_elem)%iown) &
            grid%dof_own = grid%dof_own - &
                           grid%system_size*face_dof(grid%face(parent_face(1))%degree)
      endif
   endif

endif

if (new_face2) then

   grid%nface = grid%nface + 2
   grid%nface_leaf = grid%nface_leaf + 1
   if (grid%element(grid%face(face2)%assoc_elem)%iown) then
      grid%nface_leaf_own = grid%nface_leaf_own + 1
   endif
   grid%nedge = grid%nedge + 1
   grid%nedge_leaf = grid%nedge_leaf + 1
   if (grid%element(grid%edge(edge4)%assoc_elem)%iown) then
      grid%nedge_leaf_own = grid%nedge_leaf_own + 1
   endif

   if (grid%edge(edge4)%degree >= 2) then
      if (grid%any_periodic) then
         inc = grid%edge(edge4)%degree - 1
         if (inc /= 0) then
            do comp=1,grid%system_size
               if (.not. is_periodic_edge_slave(edge4,grid,comp)) then
                  grid%dof = grid%dof + inc
                  if (grid%element(grid%edge(edge4)%assoc_elem)%iown) &
                     grid%dof_own = grid%dof_own + inc
               endif
            end do
         endif
      else
         grid%dof = grid%dof + grid%system_size*(grid%edge(edge4)%degree - 1)
         if (grid%element(grid%edge(edge4)%assoc_elem)%iown) &
            grid%dof_own = grid%dof_own + &
                           grid%system_size*(grid%edge(edge4)%degree - 1)
      endif
   endif

   if (grid%face(face3)%degree >= 3) then
      if (grid%any_periodic) then
         inc = face_dof(grid%face(face3)%degree)
         if (inc /= 0) then
            do comp=1,grid%system_size
               if (.not. is_periodic_face_slave(face3,grid,comp)) then
                  grid%dof = grid%dof + inc
                  if (grid%element(grid%face(face3)%assoc_elem)%iown) &
                     grid%dof_own = grid%dof_own + inc
               endif
            end do
         endif
      else
         grid%dof = grid%dof+grid%system_size*face_dof(grid%face(face3)%degree)
         if (grid%element(grid%face(face3)%assoc_elem)%iown) &
            grid%dof_own = grid%dof_own + &
                           grid%system_size*face_dof(grid%face(face3)%degree)
      endif
   endif

   if (grid%face(face4)%degree >= 3) then
      if (grid%any_periodic) then
         inc = face_dof(grid%face(face4)%degree)
         if (inc /= 0) then
            do comp=1,grid%system_size
               if (.not. is_periodic_face_slave(face4,grid,comp)) then
                  grid%dof = grid%dof + inc
                  if (grid%element(grid%face(face4)%assoc_elem)%iown) &
                     grid%dof_own = grid%dof_own + inc
               endif
            end do
         endif
      else
         grid%dof = grid%dof+grid%system_size*face_dof(grid%face(face4)%degree)
         if (grid%element(grid%face(face4)%assoc_elem)%iown) &
            grid%dof_own = grid%dof_own + &
                           grid%system_size*face_dof(grid%face(face4)%degree)
      endif
   endif

   if (grid%face(parent_face(2))%degree >= 3) then
      if (grid%any_periodic) then
         inc = face_dof(grid%face(parent_face(2))%degree)
         if (inc /= 0) then
            do comp=1,grid%system_size
               if (.not. is_periodic_face_slave(parent_face(2),grid,comp)) then
                  grid%dof = grid%dof - inc
                  if (grid%element(grid%face(parent_face(2))%assoc_elem)%iown) &
                     grid%dof_own = grid%dof_own - inc
               endif
            end do
         endif
      else
         grid%dof = grid%dof - grid%system_size*face_dof(grid%face(parent_face(2))%degree)
         if (grid%element(grid%face(parent_face(2))%assoc_elem)%iown) &
            grid%dof_own = grid%dof_own - &
                           grid%system_size*face_dof(grid%face(parent_face(2))%degree)
      endif
   endif

endif

if (grid%face(face5)%degree >= 3) then
   if (grid%any_periodic) then
      inc = face_dof(grid%face(face5)%degree)
      if (inc /= 0) then
         do comp=1,grid%system_size
            if (.not. is_periodic_face_slave(face5,grid,comp)) then
               grid%dof = grid%dof + inc
               if (grid%element(grid%face(face5)%assoc_elem)%iown) &
                  grid%dof_own = grid%dof_own + inc
            endif
         end do
      endif
   else
      grid%dof = grid%dof + grid%system_size*face_dof(grid%face(face5)%degree)
      if (grid%element(grid%face(face5)%assoc_elem)%iown) &
         grid%dof_own = grid%dof_own + &
                        grid%system_size*face_dof(grid%face(face5)%degree)
   endif
endif

grid%nelem = grid%nelem + 2
grid%nelem_leaf = grid%nelem_leaf + 1
if (grid%element(parent)%iown) then
   grid%nelem_leaf_own = grid%nelem_leaf_own + 1
endif
grid%nface = grid%nface + 1
grid%nface_leaf = grid%nface_leaf + 1
if (grid%element(grid%face(face5)%assoc_elem)%iown) then
   grid%nface_leaf_own = grid%nface_leaf_own + 1
endif
grid%nlev = max(grid%nlev,grid%element(child1)%level)

grid%grid_modified = .true.

! If we are using the compatibility chain, refine all the elements that share
! the refinement edge.

if (loc_use_compatibility_chain) then
   refedge = refinement_edge
   do
      do i=1,nshare_refedge
         if (share_refedge(i)==parent) cycle
         if (.not. grid%element(share_refedge(i))%isleaf) cycle
         call h_refine_element(grid,share_refedge(i),errcode,refcont, &
                               solver_control,reftype,reduce_p,reduce_p_max, &
                               vert_child_lid,edge_child_lid,elem_child_lid, &
                               delta_nelem, delta_nelem_leaf, &
                               delta_nelem_leaf_own,delta_nedge, &
                               delta_nedge_leaf,delta_nedge_leaf_own, &
                               delta_nvert, delta_nvert_own, delta_dof, &
                               delta_dof_own, max_nlev,desired_level, &
                               desired_degree,elem_list, was_hcoarsened, &
                               override_unrefinable=.true., &
                               been_there=been_there)
      end do
      deallocate(share_refedge)
      if (is_periodic_edge(refedge,grid)) then
         refedge = grid%edge(refedge)%next
         if (refedge == refinement_edge) exit
         call get_edge_elements(grid,refedge,share_refedge)
         nshare_refedge = size(share_refedge)
      else
         exit
      endif
   end do

! If we need closure then there was a cycle in the compatibility chain and we
! now have hanging nodes.  Remove the hanging nodes.

   if (my_need_closure) then
      call remove_hanging_nodes(grid,refcont,solver_control, &
                                desired_level,desired_degree,elem_list)
      my_need_closure = .false.
   endif

! Having used the compatibility chain allows for a better initial guess,

   if (refcont%error_estimator == INITIAL_CONDITION) then
      call init_guess_ic(grid,refinement_edge=refinement_edge)
   else
      call h_refine_init_guess(grid,refinement_edge,solver_control)
   endif

! and then we can compute new error indicators.

   refedge = refinement_edge
   call get_edge_elements(grid,refedge,share_refedge)
   nshare_refedge = size(share_refedge)
   do
      do i=1,nshare_refedge
         children = get_child_lid(grid%element(share_refedge(i))%gid, &
                                  ALL_CHILDREN,grid%elem_hash)
         call error_indicator(grid,children(1),refcont%error_estimator, &
                              solver_control, &
                              grid%element_errind(children(1),:), &
                              grid%element(children(1))%work)
         call error_indicator(grid,children(2),refcont%error_estimator, &
                              solver_control, &
                              grid%element_errind(children(2),:), &
                              grid%element(children(2))%work)
      end do
      deallocate(share_refedge)
      if (is_periodic_edge(refedge,grid)) then
         refedge = grid%edge(refedge)%next
         if (refedge == refinement_edge) exit
         call get_edge_elements(grid,refedge,share_refedge)
         nshare_refedge = size(share_refedge)
      else
         exit
      endif
   end do

! If we did not use the compatibility chain and been_there is not present, the
! error indicators are old

else
   if (.not. present(been_there)) then
      grid%errind_up2date = .false.
   endif

endif

if (present(need_closure)) need_closure = my_need_closure

return

contains

function get_ordered_neighbors(elem,jerr)
!        ---------------------

!----------------------------------------------------
! This routine returns the neighbors of element elem in the same order as
! the face they share.  If that neighbor has already been refined, it is
! the parent that is returned.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

integer, intent(in) :: elem
integer, intent(out) :: jerr
type(hash_key) :: get_ordered_neighbors(NEIGHBORS_PER_ELEMENT)
!----------------------------------------------------
! Local variables:

integer :: neigh(NEIGHBORS_PER_ELEMENT), num_boundary, f, n, &
           neigh_parent(NEIGHBORS_PER_ELEMENT)
!----------------------------------------------------
! Begin executable code

jerr = 0

! get the neighbors

neigh = get_neighbors(elem,grid,missing_OK=.true.)

! get the parents of the neighbors

do n=1,NEIGHBORS_PER_ELEMENT
   if (neigh(n) == BOUNDARY) then
      neigh_parent(n) = BOUNDARY
   elseif (grid%element(neigh(n))%level == 1) then
      neigh_parent(n) = BOUNDARY ! not really, but it keeps it from looking
   elseif (grid%element(neigh(n))%gid/MAX_CHILD == &
           grid%element(elem)%gid/MAX_CHILD) then
      neigh_parent(n) = BOUNDARY ! avoid looking at my own parent
   else
      neigh_parent(n) = hash_decode_key(grid%element(neigh(n))%gid/MAX_CHILD, &
                                        grid%elem_hash)
      if (neigh_parent(n) == HASH_NOT_FOUND) then
         ierr = PHAML_INTERNAL_ERROR
         call fatal("get_ordered_neighbors: parent of neighbor not found")
         stop
      endif
   endif
end do

! for each face, find the neighbor that shares that face

num_boundary = 0
do f=1,FACES_PER_ELEMENT
   get_ordered_neighbors(f) = BOUNDARY
   do n=1,NEIGHBORS_PER_ELEMENT
      if (neigh(n) /= BOUNDARY) then
         if (is_periodic_face(grid%element(elem)%face(f),grid)) then
            if (any(grid%element(neigh(n))%face == &
                grid%face(grid%element(elem)%face(f))%next)) then
               get_ordered_neighbors(f) = grid%element(neigh(n))%gid
               exit
            endif
         else
            if (any(grid%element(neigh(n))%face == &
                grid%element(elem)%face(f))) then
               get_ordered_neighbors(f) = grid%element(neigh(n))%gid
               exit
            endif
         endif
      endif
      if (neigh_parent(n) /= BOUNDARY) then
         if (is_periodic_face(grid%element(elem)%face(f),grid)) then
            if (any(grid%element(neigh_parent(n))%face == &
                grid%face(grid%element(elem)%face(f))%next)) then
               get_ordered_neighbors(f) = grid%element(neigh_parent(n))%gid
               exit
            endif
         else
            if (any(grid%element(neigh_parent(n))%face == &
                grid%element(elem)%face(f))) then
               get_ordered_neighbors(f) = grid%element(neigh_parent(n))%gid
               exit
            endif
         endif
      endif
   end do
   if (get_ordered_neighbors(f) == BOUNDARY) then
      num_boundary = num_boundary + 1
   endif
end do

if (count(neigh == BOUNDARY) /= num_boundary) then
   jerr = 1
endif

end function get_ordered_neighbors

end subroutine h_refine_element

!          --------------
subroutine after_h_refine(grid,element_list,nelem,vert_lid,edge_lid,elem_lid, &
                          refine_control,solver_control)
!          --------------

!----------------------------------------------------
! This routine performs parts of h refinement that must be done by a single
! OpenMP thread after the OpenMP-parallel refinement.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
integer, intent(in) :: element_list(:), nelem, vert_lid(:,:), edge_lid(:,:), &
                       elem_lid(:,:)
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

! Not supporting OpenMP in 3D

end subroutine after_h_refine

!                    ------------------------
recursive subroutine h_refine_refinement_edge(grid,refedge,refine_control, &
                                              solver_control,refedge_locks, &
                                              lock_initialized,errcode)
!                    ------------------------

!----------------------------------------------------
! This routine refines (all the elements around) refinement edge refedge.
! If any element that contains refedge has a different refinement edge, that
! element is refined first.  If compiled for OpenMP, each instance of this
! routine is a parallel task.
!
! Error codes:
!  0 success
!  1 refedge already refined
!  2 out of memory
!  3 an element would exceed number of levels or be too small
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
integer, intent(in) :: refedge
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
integer(omp_lock_kind) :: refedge_locks(:)
logical(small_logical), intent(inout) :: lock_initialized(:)
integer, intent(out) :: errcode

!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

ierr = PHAML_INTERNAL_ERROR
call fatal("h_refine_refinement_edge not supported in 3D")
stop

end subroutine h_refine_refinement_edge

!                    --------------
recursive subroutine create_element(grid,elemgid,refine_control, &
                                    solver_control,err,desired_level, &
                                    desired_degree,was_hcoarsened,inhalo, &
                                    disown)
!                    --------------

!----------------------------------------------------
! This routine performs the refinements needed to create the element
! with global id elemgid
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(hash_key), intent(in) :: elemgid
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
integer, intent(inout) :: err
integer, pointer, optional :: desired_level(:), desired_degree(:,:)
logical(small_logical), pointer, optional :: was_hcoarsened(:)
logical, intent(inout), optional :: inhalo
logical, intent(in), optional :: disown
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

call fatal("create_element not written for 3D")
stop
end subroutine create_element

!          -------------
subroutine init_guess_ic(grid,element,refinement_edge)
!          -------------

!----------------------------------------------------
! This routine sets the solution in either the given element or the children of
! the elements around the given refinement edge from the function in iconds.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
integer, optional, intent(in) :: element, refinement_edge
!----------------------------------------------------
! Local variables:

integer :: refedge, nshare_refedge, i, ichild, child, ivert, vert, comp, &
           eigen, iedge, edge, iface, face
integer :: children(MAX_CHILD)
integer, pointer :: share_refedge(:)

!----------------------------------------------------
! Begin executable code

! Exactly one of element or refinement_edge should be present

if (present(element) .eqv. present(refinement_edge)) then
   ierr = PHAML_INTERNAL_ERROR
   call fatal("init_guess_ic: exactly one of element and refinement_edge should be present")
   stop
endif

! Get the elements around the refinement edge

if (present(refinement_edge)) then
   refedge = refinement_edge
   call get_edge_elements(grid,refedge,share_refedge)
   nshare_refedge = size(share_refedge)
else
   nshare_refedge = 1
endif

! For each element around the refinement edge ...

outer: do
   do i=1,nshare_refedge

! Get the children of the element, or place the input element in children

      if (present(refinement_edge)) then
         children = get_child_lid(grid%element(share_refedge(i))%gid, &
                                  ALL_CHILDREN,grid%elem_hash)
      else
         children(1) = element
      endif

! For each child ...

      do ichild=1,MAX_CHILD
         child = children(ichild)

! Set the vertex basis function coefficients

         do ivert=1,VERTICES_PER_ELEMENT
            vert = grid%element(child)%vertex(ivert)
            do comp=1,grid%system_size
               if (grid%vertex_type(vert,comp) /= DIRICHLET .and. &
                   grid%vertex_type(vert,comp) /= INTERIOR_DIRICHLET .and. &
                   grid%vertex_type(vert,comp) /= PERIODIC_MASTER_DIR .and. &
                   grid%vertex_type(vert,comp) /= PERIODIC_SLAVE_DIR) then
                  do eigen=1,max(1,grid%num_eval)
                     call set_vert_soln(grid,vert,comp,eigen, &
                                        my_iconds(grid%vertex(vert)%coord%x, &
                                                  grid%vertex(vert)%coord%y, &
                                                  grid%vertex(vert)%coord%z, &
                                                  comp,eigen))
                  end do
               endif
            end do
         end do

! set the edge basis function coefficients

         do iedge=1,EDGES_PER_ELEMENT
            edge = grid%element(child)%edge(iedge)
            do comp=1,grid%system_size
               if (grid%edge_type(edge,comp) /= DIRICHLET .and. &
                   grid%edge_type(edge,comp) /= INTERIOR_DIRICHLET .and. &
                   grid%edge_type(edge,comp) /= PERIODIC_MASTER_DIR .and. &
                   grid%edge_type(edge,comp) /= PERIODIC_SLAVE_DIR) then
                  call edge_exact(grid,edge,comp,"i")
               endif
            end do
         end do

! set the face basis function coefficients

         do iface=1,FACES_PER_ELEMENT
            face = grid%element(child)%face(iface)
            do comp=1,grid%system_size
               if (grid%face_type(face,comp) /= DIRICHLET .and. &
                   grid%face_type(face,comp) /= INTERIOR_DIRICHLET .and. &
                   grid%face_type(face,comp) /= PERIODIC_MASTER_DIR .and. &
                   grid%face_type(face,comp) /= PERIODIC_SLAVE_DIR) then
                  call face_exact(grid,face,comp,"i")
               endif
            end do
         end do

! Set the element basis function coefficients

         do comp=1,grid%system_size
            call elem_exact(grid,child,comp,"i")
         end do

! If the input was an element, get out of here

         if (present(element)) then
            exit outer
         endif

      end do ! next child

   end do ! next element that shares the refinement edge

! If the refinement edge is periodic, process the periodic mate of the edge

   deallocate(share_refedge)
   if (is_periodic_edge(refedge,grid)) then
      refedge = grid%edge(refedge)%next
      if (refedge == refinement_edge) exit
      call get_edge_elements(grid,refedge,share_refedge)
      nshare_refedge = size(share_refedge)
   else
      exit
   endif
end do outer

end subroutine init_guess_ic

!          ------------
subroutine init_guess_p(grid,elem)
!          ------------

!----------------------------------------------------
! This routine sets the initial guess for the solution components associated
! with bases that have been added by the p refinement of element elem.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout), target :: grid
integer, intent(in) :: elem
!----------------------------------------------------
! Local variables:

!----------------------------------------------------
! Begin executable code

! just leave the new hierarchical coefficients at 0

end subroutine init_guess_p

!          -------------------
subroutine hcoarsen_init_guess(grid,elements,refinement_edge)
!          -------------------

!----------------------------------------------------
! This routine sets the initial guess of the solution for the parents
! when a group of elements sharing a refinement edge are unbisected.  The
! size of refinement edge is 1 for a regular edge, 2 for a periodic edge,
! and 4 for a doubly periodic edge.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout), target :: grid
integer, intent(in) :: elements(:), refinement_edge(:)
!----------------------------------------------------
! Local variables:

integer :: i, j, k, face
!----------------------------------------------------
! Begin executable code

! We just want to make sure an old solution from before the elements were
! refined doesn't give an error indicator out in left field.  For example, the
! old solution might be from a much earlier time in a time dependent problem.
! The vertices and faces that do not contain the refinement edge are current.
! We only need to do the refinement edge, faces that contain the refinement
! edge, and elements.  Setting them to 0 still gives a reasonable estimate from
! all the things that are current.

if (grid%edge(refinement_edge(1))%degree >= 2) then
   do i=1,size(refinement_edge)
      call set_edge_soln(grid,refinement_edge(i),colon,colon,colon,0.0_my_real)
   end do
endif

do i=1,size(elements)

   if (grid%element(elements(i))%degree >= 4) then
      call set_elem_soln(grid,elements(i),colon,colon,colon,0.0_my_real)
   endif

   do j=1,FACES_PER_ELEMENT
      face = grid%element(elements(i))%face(j)
      if (grid%face(face)%degree >= 3) then
         do k=1,size(refinement_edge)
            if (any(grid%face(face)%edge == refinement_edge(k))) then
               call set_face_soln(grid,face,colon,colon,colon,0.0_my_real)
            endif
         end do
      endif
   end do

end do

end subroutine hcoarsen_init_guess

!          -------------------
subroutine h_refine_init_guess(grid,refinement_edge,solver_control)
!          -------------------

!----------------------------------------------------
! This routine computes an initial guess of the solution for elements created
! by the refinement of refinement edge.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
integer, intent(in) :: refinement_edge
type(solver_options), intent(in) :: solver_control
!----------------------------------------------------
! Local variables:

integer :: new_vert, nparent, nneigh_vert, astat, ncoef, ss, inc_quad_order, &
           iparent, ichild, child, i, new_vert_row, elem_col, vert, col
real(my_real) :: diag, rhs, initial_guess
logical :: resid_prob
character(len=1) :: which_basis, which_set
integer, pointer :: parents(:)
integer :: children(MAX_CHILD), degree(EDGES_PER_ELEMENT+FACES_PER_ELEMENT+1)
integer :: edge_type(EDGES_PER_ELEMENT,1), face_type(FACES_PER_ELEMENT,1), &
           bmark(EDGES_PER_ELEMENT+FACES_PER_ELEMENT)
real(my_real) :: lmat(VERTICES_PER_ELEMENT,VERTICES_PER_ELEMENT), &
                 lrhs(VERTICES_PER_ELEMENT)
real(my_real) :: xvert(VERTICES_PER_ELEMENT), yvert(VERTICES_PER_ELEMENT), &
                 zvert(VERTICES_PER_ELEMENT)
real(my_real), allocatable :: coef(:)
integer, allocatable :: idcoef(:)
!----------------------------------------------------
! Begin executable code

! When an element is refined, the solution over the new elements is unchanged,
! i.e. the solution at the new vertex is the linear interpolation of the
! solution at the endpoints of the refinement edge, and coefficients to all
! high order (p-hierarchical) bases is set to 0.  Compute a better
! approximation to the solution at the new vertex, and leave the high order
! (hierarchical) d.o.f. at 0.

! Assuming linear elements, solve a local Dirichlet problem for the solution at
! the new vertex, i.e. compute the row of the matrix associated with the new
! vertex and perform Gauss-Siedel relaxation on just that row assuming the
! existing values at the neighboring vertices.

! Only doing scalar PDEs.

if (grid%system_size > 1) return

! Not doing the initial guess where there are periodic boundary conditions.

if (is_periodic_edge(refinement_edge,grid)) return

! Identify the new vertex.  It is the vertex of a child of the refinement edge
! that is not a vertex of the refinement edge.

if ((grid%edge(grid%edge(refinement_edge)%child(1))%vertex(1) == &
     grid%edge(refinement_edge)%vertex(1)) .or. &
    (grid%edge(grid%edge(refinement_edge)%child(1))%vertex(1) == &
     grid%edge(refinement_edge)%vertex(2))) then
   new_vert = grid%edge(grid%edge(refinement_edge)%child(1))%vertex(2)
else
   new_vert = grid%edge(grid%edge(refinement_edge)%child(1))%vertex(1)
endif

! If the new vertex is a Dirichlet point, the solution was set from the
! boundary conditions and should not be changed.

if (grid%vertex_type(new_vert,1) == DIRICHLET .or. &
    grid%vertex_type(new_vert,1) == INTERIOR_DIRICHLET) return

! Get the parent elements around the refinement edge

call get_edge_elements(grid,refinement_edge,parents)
nparent = size(parents)

! We do not need to build the whole linear system, just the row for the new
! vertex.  Put the diagonal in diag, the other entries in the row in coef(:),
! the corresponding vertex lids in idcoef(:), and the right hand side in rhs.

! Each parent element contains the two endpoints of the refinement edge and two
! other vertices.  If the new vertex is interior, then one of those vertices is
! shared with another parent, so the number of neighboring vertices is
! nparent + 2.  If the new vertex has a natural or mixed boundary condition,
! then two of the parents have a vertex that is not shared with another parent,
! so there is one more neighboring vertex.  All other types for the new vertex
! have been removed by exits earlier in this routine.

select case(grid%vertex_type(new_vert,1))
case(INTERIOR)
   nneigh_vert = nparent + 2
case(NATURAL, MIXED)
   nneigh_vert = nparent + 3
case default
   ierr = PHAML_INTERNAL_ERROR
   call fatal("h_refine_init_guess: unexpected vertex type")
   stop
end select

allocate(coef(nneigh_vert),idcoef(nneigh_vert),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("h_refine_init_guess: memory allocation failed")
   stop
endif

! idcoef will be built up by searching for the vertex lid in idcoef and adding
! it to the end if it is not there.  Each child element will contain the new
! vertex, one of the end points of the refinement edge, and two other
! neighboring vertices, one of which is the one that was added by the previous
! child element at least half the time (the second child of a parent is done
! right after the first child).  So to minimize the search, place the two end
! points of the refinement edge first in idcoef and search those two, then the
! last one, then the rest.

idcoef(1) = grid%edge(refinement_edge)%vertex(1)
idcoef(2) = grid%edge(refinement_edge)%vertex(2)
ncoef = 2

coef = 0.0_my_real
diag = 0.0_my_real
rhs  = 0.0_my_real

! Arguments to elemental_matrix that are constant

degree = 1
ss = 1
inc_quad_order = 0
resid_prob = .false.
which_basis = "p"
which_set = "a"

! For each child element ...

do iparent=1,nparent
   children = get_child_lid(grid%element(parents(iparent))%gid,ALL_CHILDREN, &
                            grid%elem_hash)
   do ichild=1,MAX_CHILD
      child = children(ichild)

! Arguments to elemental matrix for this element

      xvert = grid%vertex(grid%element(child)%vertex)%coord%x
      yvert = grid%vertex(grid%element(child)%vertex)%coord%y
      zvert= zcoord(grid%vertex(grid%element(child)%vertex)%coord)
      do i=1,EDGES_PER_ELEMENT
         edge_type(i,1) = grid%edge_type(grid%element(child)%edge(i),1)
         bmark(i) = grid%edge(grid%element(child)%edge(i))%bmark
      end do
      do i=1,FACES_PER_ELEMENT
         face_type(i,1) = grid%face_type(grid%element(child)%face(i),1)
         bmark(EDGES_PER_ELEMENT+i)=grid%face(grid%element(child)%face(i))%bmark
      end do

! Compute the elemental matrix

      call elemental_matrix(grid, xvert, yvert, degree, edge_type, face_type, &
                            bmark, ss, inc_quad_order, resid_prob, child, &
                            which_basis, which_set, lmat, lrhs, &
                            loc_bconds_s = my_bconds, zvert=zvert, &
                            solver_control=solver_control)

! Identify the row for new vertex

      do new_vert_row=1,VERTICES_PER_ELEMENT
         if (grid%element(child)%vertex(new_vert_row) == new_vert) exit
      end do

      if (new_vert_row > VERTICES_PER_ELEMENT) then
         ierr = PHAML_INTERNAL_ERROR
         call fatal("h_refine_init_guess: failed to find new vertex in element")
         stop
      endif

! Assemble

! For each column of the elemental matrix

      do elem_col=1,VERTICES_PER_ELEMENT
         vert = grid%element(child)%vertex(elem_col)

! If this is the new vertex, add to diag

         if (vert == new_vert) then
            diag = diag + lmat(new_vert_row,elem_col)

! Otherwise, identify the corresponding column in the assembled matrix ...

         else
            col = 0
            if (idcoef(1) == vert) then
               col = 1
            elseif (idcoef(2) == vert) then
               col = 2
            elseif (idcoef(ncoef) == vert) then
               col = ncoef
            else
               do i=3,ncoef-1
                  if (idcoef(i) == vert) then
                     col = i
                     exit
                  endif
               end do
               if (col == 0) then
                  ncoef = ncoef + 1
                  if (ncoef > nneigh_vert) then
                     ierr = PHAML_INTERNAL_ERROR
                     call fatal("h_refine_init_guess: ncoef > nneigh_vert")
                     stop
                  endif
                  idcoef(ncoef) = vert
                  col = ncoef
               endif
            endif

! ... and add to coef.

            coef(col) = coef(col) + lmat(new_vert_row,elem_col)

         endif

      end do ! next column of elemental matrix

! Add to rhs

      rhs = rhs + lrhs(new_vert_row)

   end do ! next child
end do ! next parent

! Relaxation to get the initial guess.

initial_guess = rhs
do col=1,ncoef
   initial_guess = initial_guess - coef(col)*get_vert_soln(grid,idcoef(col),1,1)
end do
initial_guess = initial_guess/diag

! Set the initial guess.

call set_vert_soln(grid,new_vert,1,1,initial_guess)

! Clean up memory

deallocate(parents,coef,idcoef)

end subroutine h_refine_init_guess

!          --------------------
subroutine remove_hanging_nodes(grid,refine_control,solver_control, &
                                desired_level,desired_degree,elem_list, &
                                was_hcoarsened)
!          --------------------

!----------------------------------------------------
! This routine removes any hanging nodes in the grid
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
integer, optional, pointer :: desired_level(:), desired_degree(:,:), &
                              elem_list(:)
logical(small_logical), optional, pointer :: was_hcoarsened(:)
!----------------------------------------------------
! Local variables:

integer :: lev, elem, iedge, edge, errcode, numperiodic, idum, mates(4), &
           jedge
logical :: changed
!----------------------------------------------------
! Begin executable code

changed = .true.
do while(changed)

! Check each edge of every leaf element to see if the edge has been refined,
! i.e. contains a hanging node.  If so, refine the element.  If not, see if
! the edge is a periodic edge, and if any of the periodic mates have been
! refined, refine the element.

   changed = .false.
   do lev=1,grid%nlev
      elem = grid%head_level_elem(lev)
      do while(elem /= END_OF_LIST)
         if (grid%element(elem)%isleaf) then
            do iedge=1,EDGES_PER_ELEMENT
               edge = grid%element(elem)%edge(iedge)
               if (grid%edge(edge)%child(1) /= NO_CHILD) then
                  changed = .true.
                  call h_refine_element(grid,elem,errcode, &
                                        refine_control,solver_control, &
                                        desired_level=desired_level, &
                                        desired_degree=desired_degree, &
                                        elem_list=elem_list, &
                                        was_hcoarsened=was_hcoarsened, &
                                        override_unrefinable=.true.)
               else
                  if (grid%any_periodic) then
                     numperiodic = num_matching_periodic_edge(grid,edge)
                     if (numperiodic /= 0) then
                        idum = matching_periodic_edge(grid,edge,mates)
                        do jedge=1,numperiodic
                           if (grid%edge(mates(jedge))%child(1) /= NO_CHILD) then
                              changed = .true.
                              call h_refine_element(grid,elem,errcode, &
                                                 refine_control,solver_control,&
                                                 desired_level=desired_level, &
                                                 desired_degree=desired_degree,&
                                                 elem_list=elem_list, &
                                                 was_hcoarsened=was_hcoarsened,&
                                                 override_unrefinable=.true.)
                           endif
                        end do
                     endif
                  endif
               endif
            end do
         endif
         elem = grid%element(elem)%next
      end do
   end do
end do

end subroutine remove_hanging_nodes

end module refine_elements
